/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "config.h"

#if ENABLE(WEB_AUDIO)

//#if USE(WEBAUDIO_WIIU)

#include "FFTFrame.h"

#include "VectorMath.h"

#if PLATFORM(WIIU)
#include <libffmpeg-fft/ffmpeg_FFT_WiiU.h>
#include <shim_memory.h>
#endif

#include <wtf/MathExtras.h>

namespace WebCore {


class FFTFrame::Impl
    {
public:
    Impl(unsigned fftSize, unsigned fftSizeLog2);
    Impl(const Impl &rhs);
    ~Impl();

    void doFFT(float* data);
    void doInverseFFT(float* data);

private:
    static void *_fftAlloc(unsigned int numBytes);
    static void _fftFree(void *pAddress);
    static void *_fftRealloc(void *pAddress, unsigned int newByteSize);

    static void Initialize();
    static bool s_bInitialized;

    void        *m_pHandle;

    unsigned    m_FFTSize;
    unsigned    m_log2FFTSize;

public:
    float* getUpToDateComplexData();
    AudioFloatArray m_complexData;
    AudioFloatArray m_realData;
    AudioFloatArray m_imagData;
    };

bool FFTFrame::Impl::s_bInitialized = false;

//static
void *FFTFrame::Impl::_fftAlloc(unsigned int numBytes)
{
    void *pA = malloc(size_t(numBytes));
    ASSERT(pA);
    return(pA);
}

//static
void FFTFrame::Impl::_fftFree(void *pAddress)
{
    free(pAddress);
}

//static
void *FFTFrame::Impl::_fftRealloc(void *pAddress, unsigned int newByteSize)
{
	#if PLATFORM(WIIU)
    return(shim_Realloc(pAddress, size_t(newByteSize)));
	#else
	return realloc(pAddress, newByteSize);
	#endif
}

//static
void FFTFrame::Impl::Initialize()
{
    if (!s_bInitialized)
    {
		#if PLATFORM(WIIU)
        ffmpeg_fft_init_config_wiiu cfg;
        cfg.mpMemCallbacks.Allocate        = _fftAlloc;
        cfg.mpMemCallbacks.AllocateAligned = NULL;
        cfg.mpMemCallbacks.Deallocate      = _fftFree;
        cfg.mpMemCallbacks.Reallocate      = _fftRealloc;
        _ffmpeg_FFT_Init_WiiU(&cfg);
		#endif

        s_bInitialized = true;
    }
}

FFTFrame::Impl::Impl(unsigned fftSize, unsigned fftSizeLog2)
    : m_FFTSize(fftSize)
    , m_log2FFTSize(fftSizeLog2)
    , m_complexData(fftSize)
    , m_realData(fftSize / 2)
    , m_imagData(fftSize / 2)
{
    Initialize();
	#if PLATFORM(WIIU)
    _ffmpeg_FFT_Create_WiiU(&m_pHandle, m_log2FFTSize);
	#endif
}

FFTFrame::Impl::Impl(const FFTFrame::Impl &rhs)
    : m_FFTSize(rhs.m_FFTSize)
    , m_log2FFTSize(rhs.m_log2FFTSize)
    , m_complexData(rhs.m_FFTSize)
    , m_realData(rhs.m_FFTSize / 2)
    , m_imagData(rhs.m_FFTSize / 2)
{
    Initialize();
	#if PLATFORM(WIIU)
    _ffmpeg_FFT_Create_WiiU(&m_pHandle, m_log2FFTSize);
	#endif
}

FFTFrame::Impl::~Impl()
{
	#if PLATFORM(WIIU)
    _ffmpeg_FFT_Destroy_WiiU(m_pHandle);
	#endif
}


float* FFTFrame::Impl::getUpToDateComplexData()
{
    // FIXME: if we can't completely get rid of this method, SSE
    // optimization could be considered if it shows up hot on profiles.
    int len = m_FFTSize / 2;
    for (int i = 0; i < len; ++i) {
        int baseComplexIndex = 2 * i;
        m_complexData[baseComplexIndex] = m_realData[i];
        m_complexData[baseComplexIndex + 1] = m_imagData[i];
    }
    return const_cast<float*>(m_complexData.data());
}

void FFTFrame::Impl::doFFT(float* data)
{
    // Copy since processing is in-place.
    float* p = m_complexData.data();
    memcpy(p, data, sizeof(float) * m_FFTSize);

	#if PLATFORM(WIIU)
    // Compute Forward transform.
    _ffmpeg_FFT_CalcForward_WiiU(m_pHandle, p);
	#endif

    // De-interleave to separate real and complex arrays.
    int len = m_FFTSize / 2;

    // FIXME: see above comment in multiply() about scaling.
    const float scale = 2.0f;

    for (int i = 0; i < len; ++i) {
        int baseComplexIndex = 2 * i;
        // m_realData[0] is the DC component and m_imagData[0] is the nyquist component
        // since the interleaved complex data is packed.
        m_realData[i] = scale * p[baseComplexIndex];
        m_imagData[i] = scale * p[baseComplexIndex + 1];
    }
}

void FFTFrame::Impl::doInverseFFT(float* data)
{
    // Prepare interleaved data.
    float* interleavedData = getUpToDateComplexData();

	#if PLATFORM(WIIU)
    // Compute inverse transform.
    _ffmpeg_FFT_CalcInverse_WiiU(m_pHandle, interleavedData);
	#endif

    // Scale so that a forward then inverse FFT yields exactly the original data.
    const float scale = 1.0 / m_FFTSize;
    VectorMath::vsmul(interleavedData, 1, &scale, data, 1, m_FFTSize);
}

// Normal constructor: allocates for a given fftSize.
FFTFrame::FFTFrame(unsigned fftSize)
    : m_FFTSize(fftSize)
    , m_log2FFTSize(static_cast<unsigned>(log2(fftSize)))
{
    // We only allow power of two.
    ASSERT(1UL << m_log2FFTSize == m_FFTSize);
    m_Impl = new Impl(m_FFTSize, m_log2FFTSize);
}

// Creates a blank/empty frame (interpolate() must later be called).
FFTFrame::FFTFrame()
    : m_FFTSize(0)
    , m_log2FFTSize(0)
    , m_Impl(0)
{
}

// Copy constructor.
FFTFrame::FFTFrame(const FFTFrame& frame)
    : m_FFTSize(frame.m_FFTSize)
    , m_log2FFTSize(frame.m_log2FFTSize)
    , m_Impl(0)
{
    if (frame.m_Impl)
    {
        m_Impl = new Impl(*frame.m_Impl);
        // Copy/setup frame data.
        unsigned nbytes = sizeof(float) * (m_FFTSize / 2);
        memcpy(realData(), frame.realData(), nbytes);
        memcpy(imagData(), frame.imagData(), nbytes);
    }
}

void FFTFrame::initialize()
{
}

void FFTFrame::cleanup()
{
}

FFTFrame::~FFTFrame()
{
    delete m_Impl;
}

void FFTFrame::multiply(const FFTFrame& frame)
{
    FFTFrame& frame1 = *this;
    FFTFrame& frame2 = const_cast<FFTFrame&>(frame);

    float* realP1 = frame1.realData();
    float* imagP1 = frame1.imagData();
    const float* realP2 = frame2.realData();
    const float* imagP2 = frame2.imagData();

    // Scale accounts the peculiar scaling of vecLib on the Mac.
    // This ensures the right scaling all the way back to inverse FFT.
    // FIXME: if we change the scaling on the Mac then this scale
    // factor will need to change too.
    float scale = 0.5f;

    // Multiply the packed DC/nyquist component
    realP1[0] *= scale * realP2[0];
    imagP1[0] *= scale * imagP2[0];

    // Complex multiplication. If this loop turns out to be hot then
    // we should use SSE or other intrinsics to accelerate it.
    unsigned halfSize = fftSize() / 2;

    for (unsigned i = 1; i < halfSize; ++i) {
        float realResult = realP1[i] * realP2[i] - imagP1[i] * imagP2[i];
        float imagResult = realP1[i] * imagP2[i] + imagP1[i] * realP2[i];

        realP1[i] = scale * realResult;
        imagP1[i] = scale * imagResult;
    }
}

void FFTFrame::doFFT(float* data)
{
    m_Impl->doFFT(data);
}

void FFTFrame::doInverseFFT(float* data)
{
    m_Impl->doInverseFFT(data);
}

float* FFTFrame::realData() const
{
    return const_cast<float*>(m_Impl->m_realData.data());
}

float* FFTFrame::imagData() const
{
    return const_cast<float*>(m_Impl->m_imagData.data());
}

} // namespace WebCore

//#endif // USE(WEBAUDIO_WIIU)

#endif // ENABLE(WEB_AUDIO)
