/*
 * Copyright (C) 2006, 2007, 2008 Apple Inc.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2007 Pioneer Research Center USA, Inc.
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (c) 2012 NINTENDO.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef FontPlatformDataFreeType_h
#define FontPlatformDataFreeType_h

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "NotImplemented.h"
#define GLYPH_CACHE_SIZE 1024
#endif
#include "FontDescription.h"
#include "FontOrientation.h"
#include "GlyphBuffer.h"
#include "HashFunctions.h"
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include "RefPtrCairo.h"
#endif
#include <wtf/Forward.h>

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "ImageSource.h"
#include "efontconfig/efontconfig.h"

#include <vector>
#else
#include "renderer/WiiURenderer.h"
#endif


typedef struct _FcFontSet FcFontSet;

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
typedef struct glyph_cache_entry {
	short			bitmap_left;
	short			bitmap_top;
	short			bitmap_width;
	short			bitmap_rows;
	float			advance_x;
	float			s0, t0, s1, t1;
} glyph_cache_entry_t;

typedef struct font_cache_info {
	int				width;
	int				height;
	int				xpos;
	int				ypos;
	int				maxGlyphHeight;
} font_cache_info_t;

struct font_cache_pair_t {
	font_cache_pair_t() {}
	font_cache_pair_t(WebCore::Glyph g, int i) : mGlyph(g), mIndex(i) {}
	font_cache_pair_t(const font_cache_pair_t & src) : mGlyph(src.mGlyph), mIndex(src.mIndex) {}
	
	bool operator < (const font_cache_pair_t & rhs) const { return(mGlyph < rhs.mGlyph); }
	bool operator > (const font_cache_pair_t & rhs) const { return(mGlyph > rhs.mGlyph); }
	bool operator == (const font_cache_pair_t & rhs) const { return(mGlyph == rhs.mGlyph); }
	bool operator != (const font_cache_pair_t & rhs) const { return(mGlyph != rhs.mGlyph); }
	
	WebCore::Glyph	mGlyph;
	int				mIndex;
};

typedef std::vector<font_cache_pair_t>	font_cache_map_t;

struct font_face_t {
	static font_face_t* create();
	void destroy();
	void use();
	
    void*				face;
    bool				face_valid;
	
    font_cache_info_t	cache_info;
    glyph_cache_entry_t	glyph_cache[GLYPH_CACHE_SIZE];
	font_cache_map_t	glyph_map;
#if PLATFORM(WIIU)
	WebCore::NativeImagePtr	m_FontSheet;
#else
	unsigned int m_FontTextureID;
	unsigned int m_FontTextureWidth;		
	unsigned int m_FontTextureHeight;		
#endif

private:
	font_face_t();
	~font_face_t();
	unsigned int		ref_cnt;
};

inline font_face_t *font_face_t::create() {	return new font_face_t(); }
inline void font_face_t::destroy()		  {	if (--ref_cnt == 0) delete this; }
inline void font_face_t::use()			  {	++ref_cnt; }

#endif

namespace WebCore {

class FontPlatformData {
public:
    FontPlatformData(WTF::HashTableDeletedValueType)
        : m_size(0)
        , m_syntheticBold(false)
        , m_syntheticOblique(false)
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
        , m_fallbacks(0)
        , m_scaledFont(hashTableDeletedFontValue())
#else
        , m_handle(NULL)
#endif
        {
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
    		memset(&m_pattern, 0, sizeof(m_pattern));
    		m_pattern.family_name = hashTableDeletedFontValue();
#endif
        }
 
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    FontPlatformData()
        : m_fallbacks(0)
        , m_size(0)
        , m_syntheticBold(false)
        , m_syntheticOblique(false)
        , m_scaledFont(0)
        { }
#else
    FontPlatformData()
        : m_size(0)
        , m_syntheticBold(false)
        , m_syntheticOblique(false)
        , m_scaledFont(0)
        , m_handle(NULL)
        {
            memset(&m_pattern, 0, sizeof(m_pattern));
            m_pattern.family_name = hashTableDeletedFontValue();
        }
#endif

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    FontPlatformData(FcPattern*, const FontDescription&);
    FontPlatformData(cairo_font_face_t* fontFace, float size, bool bold, bool italic);
#else
    FontPlatformData(struct font_config_handle *handle, struct font_param *pattern, const FontDescription& fontDescription);
    FontPlatformData(font_face_t *fontFace, float size, bool bold, bool italic);
#endif
    FontPlatformData(float size, bool bold, bool italic);
    FontPlatformData(const FontPlatformData&);
    FontPlatformData(const FontPlatformData&, float size);

    ~FontPlatformData();

    bool isFixedPitch();
    float size() const { return m_size; }
    void setSize(float size) { m_size = size; }
    bool syntheticBold() const { return m_syntheticBold; }
    bool syntheticOblique() const { return m_syntheticOblique; }
    bool hasCompatibleCharmap();

    FontOrientation orientation() const { return Horizontal; } // FIXME: Implement.
    void setOrientation(FontOrientation) { } // FIXME: Implement.

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    cairo_scaled_font_t* scaledFont() const { return m_scaledFont; }
#endif

    unsigned hash() const
    {
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
        return PtrHash<cairo_scaled_font_t*>::hash(m_scaledFont);
#else
        return PtrHash<char *>::hash(m_pattern.family_name);
#endif
    }

    bool operator==(const FontPlatformData&) const;
    FontPlatformData& operator=(const FontPlatformData&);
    bool isHashTableDeletedValue() const
    {
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
        return m_scaledFont == hashTableDeletedFontValue();
#else
        return m_pattern.family_name ==  hashTableDeletedFontValue();
#endif
    }

#ifndef NDEBUG
    String description() const;
#endif

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    RefPtr<FcPattern> m_pattern;
    mutable FcFontSet* m_fallbacks; // Initialized lazily.
#endif
    float m_size;
    bool m_syntheticBold;
    bool m_syntheticOblique;
    bool m_fixedWidth;
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    cairo_scaled_font_t* m_scaledFont;
#else
    font_face_t *				m_scaledFont;
    struct font_config_handle *	m_handle;
    struct font_param			m_pattern;
	
    void createFontCache();
    int addGlyphToCache(Glyph glyph_idx) const;
	int getGlyphCacheIndex(Glyph glyph_idx) const;
#endif

private:
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    void initializeWithFontFace(cairo_font_face_t*);
    static cairo_scaled_font_t* hashTableDeletedFontValue() { return reinterpret_cast<cairo_scaled_font_t*>(-1); }
#else
    static char *hashTableDeletedFontValue() { return reinterpret_cast<char *>(-1); }
#endif
};

}

#endif // FontPlatformDataFreeType_h
