/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if USE(ACCELERATED_COMPOSITING)

#include "LayerWiiU.h"

#include "TextStream.h"
#include "wtf/text/CString.h"

namespace WebCore {

using namespace std;

static int s_nextLayerId = 1;

PassRefPtr<LayerWiiU> LayerWiiU::create(GraphicsLayerWiiU* owner)
{
    return adoptRef(new LayerWiiU(owner));
}

LayerWiiU::LayerWiiU(GraphicsLayerWiiU* owner)
    : m_owner(owner)
    , m_contentsDirty(false)
    , m_layerId(s_nextLayerId++)
    , m_parent(0)
    , m_anchorPoint(0.5, 0.5)
    , m_backgroundColor(0, 0, 0, 0)
    , m_debugBorderWidth(0)
    , m_opacity(1.0)
    , m_zPosition(0.0)
    , m_anchorPointZ(0)
    , m_clearsContext(false)
    , m_hidden(false)
    , m_masksToBounds(false)
    , m_opaque(true)
    , m_geometryFlipped(false)
    , m_needsDisplayOnBoundsChange(false)
    , m_doubleSided(true)
    , m_usesLayerScissor(false)
    , m_isRootLayer(false)
    , m_isTransformLayer(false)
    , m_replicaLayer(0)
    , m_preserves3D(false)
    , m_drawOpacity(0)
{
}

LayerWiiU::~LayerWiiU()
{
    // Our parent should be holding a reference to us so there should be no
    // way for us to be destroyed while we still have a parent.
    ASSERT(!parent());

    // Remove the parent reference from all children.
    removeAllChildren();
}

void LayerWiiU::cleanupResources()
{
}

void LayerWiiU::cleanupResourcesRecursive()
{
    for (size_t i = 0; i < children().size(); ++i)
        children()[i]->cleanupResourcesRecursive();

    if (maskLayer())
        maskLayer()->cleanupResourcesRecursive();
    if (replicaLayer())
        replicaLayer()->cleanupResourcesRecursive();

    cleanupResources();
}

void LayerWiiU::setNeedsCommit()
{
    // Call notifySyncRequired(), which for non-root layers plumbs through to
    // call setRootLayerNeedsDisplay() on the WebView, which will cause LayerRendererWiiU
    // to render a frame.
    // This function has no effect on root layers.
    if (m_owner)
        m_owner->notifySyncRequired();
}

void LayerWiiU::setParent(LayerWiiU* layer)
{
    ASSERT(!layer || !layer->hasAncestor(this));
    m_parent = layer;
}

bool LayerWiiU::hasAncestor(LayerWiiU* ancestor) const
{
    for (LayerWiiU* layer = parent(); layer; layer = layer->parent()) {
        if (layer == ancestor)
            return true;
    }
    return false;
}

void LayerWiiU::addChild(PassRefPtr<LayerWiiU> child)
{
    insertChild(child, numChildren());
}

void LayerWiiU::insertChild(PassRefPtr<LayerWiiU> child, size_t index)
{
    index = min(index, m_children.size());
    child->removeFromParent();
    child->setParent(this);
    m_children.insert(index, child);
    setNeedsCommit();
}

void LayerWiiU::removeFromParent()
{
    if (m_parent)
        m_parent->removeChild(this);
}

void LayerWiiU::removeChild(LayerWiiU* child)
{
    int foundIndex = indexOfChild(child);
    if (foundIndex == -1)
        return;

    child->setParent(0);
    m_children.remove(foundIndex);
    setNeedsCommit();
}

void LayerWiiU::replaceChild(LayerWiiU* reference, PassRefPtr<LayerWiiU> newLayer)
{
    ASSERT_ARG(reference, reference);
    ASSERT_ARG(reference, reference->parent() == this);

    if (reference == newLayer)
        return;

    int referenceIndex = indexOfChild(reference);
    if (referenceIndex == -1) {
        ASSERT_NOT_REACHED();
        return;
    }

    reference->removeFromParent();

    if (newLayer) {
        newLayer->removeFromParent();
        insertChild(newLayer, referenceIndex);
    }
}

int LayerWiiU::indexOfChild(const LayerWiiU* reference)
{
    for (size_t i = 0; i < m_children.size(); i++) {
        if (m_children[i] == reference)
            return i;
    }
    return -1;
}

void LayerWiiU::setBounds(const IntSize& size)
{
    if (bounds() == size)
        return;

    bool firstResize = !bounds().width() && !bounds().height() && size.width() && size.height();

    m_bounds = size;

    if (firstResize)
        setNeedsDisplay(FloatRect(0, 0, bounds().width(), bounds().height()));
    else
        setNeedsCommit();
}

void LayerWiiU::setFrame(const FloatRect& rect)
{
    if (rect == m_frame)
      return;

    m_frame = rect;
    setNeedsDisplay(FloatRect(0, 0, bounds().width(), bounds().height()));
}

const LayerWiiU* LayerWiiU::rootLayer() const
{
    const LayerWiiU* layer = this;
    for (LayerWiiU* parent = layer->parent(); parent; layer = parent, parent = parent->parent()) { }
    return layer;
}

void LayerWiiU::removeAllChildren()
{
    while (m_children.size()) {
        LayerWiiU* layer = m_children[0].get();
        ASSERT(layer->parent());
        layer->removeFromParent();
    }
}

void LayerWiiU::setChildren(const Vector<RefPtr<LayerWiiU> >& children)
{
    if (children == m_children)
        return;

    removeAllChildren();
    size_t listSize = children.size();
    for (size_t i = 0; i < listSize; i++)
        addChild(children[i]);
}

LayerWiiU* LayerWiiU::parent() const
{
    return m_parent;
}

void LayerWiiU::setName(const String& name)
{
    m_name = name;
}

void LayerWiiU::setNeedsDisplay(const FloatRect& dirtyRect)
{
    // Simply mark the contents as dirty. For non-root layers, the call to
    // setNeedsCommit will schedule a fresh compositing pass.
    // For the root layer, setNeedsCommit has no effect.
    m_contentsDirty = true;

    m_dirtyRect.unite(dirtyRect);
    setNeedsCommit();
}

void LayerWiiU::setNeedsDisplay()
{
    m_dirtyRect.setLocation(FloatPoint());
    m_dirtyRect.setSize(bounds());
    m_contentsDirty = true;
    setNeedsCommit();
}

void LayerWiiU::resetNeedsDisplay()
{
    m_dirtyRect = FloatRect();
    m_contentsDirty = false;
}

String LayerWiiU::layerTreeAsText() const
{
    TextStream ts;
    dumpLayer(ts, 0);
    return ts.release();
}

static void writeIndent(TextStream& ts, int indent)
{
    for (int i = 0; i != indent; ++i)
        ts << "  ";
}

void LayerWiiU::dumpLayer(TextStream& ts, int indent) const
{
    writeIndent(ts, indent);
    ts << layerTypeAsString() << "(" << m_name << ")\n";
    dumpLayerProperties(ts, indent+2);
    if (m_replicaLayer) {
        writeIndent(ts, indent+2);
        ts << "Replica:\n";
        m_replicaLayer->dumpLayer(ts, indent+3);
    }
    if (m_maskLayer) {
        writeIndent(ts, indent+2);
        ts << "Mask:\n";
        m_maskLayer->dumpLayer(ts, indent+3);
    }
    for (size_t i = 0; i < m_children.size(); ++i)
        m_children[i]->dumpLayer(ts, indent+1);
}

void LayerWiiU::dumpLayerProperties(TextStream& ts, int indent) const
{
    writeIndent(ts, indent);
    ts << "id: " << id() << " drawsContent: " << drawsContent() << " bounds " << m_bounds.width() << "x" << m_bounds.height() << " usesLayerScissor: " << usesLayerScissor()
        << " scissorRect: (" << m_scissorRect.x() << ", " << m_scissorRect.y() << ", " << m_scissorRect.width() << ", " << m_scissorRect.height() << ")"
        << " visibleLayerRect: (" << m_visibleLayerRect.x() << ", " << m_visibleLayerRect.y() << ", " << m_visibleLayerRect.width() << ", " << m_visibleLayerRect.height() << ")"
        << "\n";

}

void LayerWiiU::setBorderColor(const Color& color)
{
    m_debugBorderColor = color;
    setNeedsCommit();
}

void LayerWiiU::setBorderWidth(float width)
{
    m_debugBorderWidth = width;
    setNeedsCommit();
}

void LayerWiiU::drawDebugBorder(GraphicsContext *ctx, float width, float height)
{
    setBorderColor(Color(255, 0, 255));
    setBorderWidth(5);

    ctx->save();
    ctx->setStrokeColor(m_debugBorderColor, ColorSpaceDeviceRGB);
    ctx->strokeRect(FloatRect(0, 0, width, height), m_debugBorderWidth);

    if (m_owner->showRepaintCounter()) {

    } 

    ctx->restore();
}

void LayerWiiU::paintContentsIfDirty()
{
	paintContentLayerIfDirty(); 	    
	
	//recursively paint any children
    for (int i = 0; i < children().size(); ++i) {
        children()[i]->paintContentsIfDirty();
    }
}

void LayerWiiU::paint(GraphicsContext *ctx, bool preserves3D, float opacity)
{
    ASSERT(ctx);

    //Each layer has an anchor point which defines the point about which the transforms are
    //applied.  For example, (0.5, 0.5) means that the object should be transformed with an
    //origin at the middle of the layer.
    float anchor_x = m_owner->anchorPoint().x();
    float anchor_y = m_owner->anchorPoint().y();
    float width  = (float)contentBounds().width();
    float height = (float)contentBounds().height();

    //OS::Report("Layer %s Pos (%fx%f)  Size (%fx%f)\n", name().utf8().data(), m_owner->position().x(), m_owner->position().y(), width, height);
    ctx->save();

    //sometimes webkit will want us to clip a set of bounds, sometimes not
    //note, it is also possible to mask to a layer, but we do not support this at the moment
    if (masksToBounds()) {
        ctx->clip(FloatRect(m_owner->position().x(), m_owner->position().y(), width, height));
    }

    if (m_owner->preserves3D())
        preserves3D = true;

    TransformationMatrix t = transform();

    if (preserves3D) {
        //Only do the Translation to the position on the Transform layer.
        if (m_owner->hasTransformLayer()  && !isTransformLayer()) {
            ctx->translate(anchor_x*width, anchor_y*height);
        } else {
            //translate so that anchor point is the origin and apply owner's position offset
            ctx->translate(anchor_x*width + m_owner->position().x(), anchor_y*height + m_owner->position().y());
        }
    } else {
        //Flatten Transform
        t.setM31(0);
        t.setM32(0);

        ctx->translate(anchor_x*width + m_owner->position().x(), anchor_y*height + m_owner->position().y());
    }

    //concatenate our transform so that is is applied to any children
    ctx->concatCTM(t);

    // Apply perspective Transform
    if (!sublayerTransform().isIdentity()) {
        ctx->concatCTM(sublayerTransform());
    }
    //translate back from anchor
    ctx->translate(-anchor_x*width, -anchor_y*height);
    
    opacity = opacity * m_opacity;

    if (drawsContent())  {

        //we handle opacity by just setting the global alpha prior to a render
        //possibly something more sophisticated is require
        if (!m_doubleSided && ctx->isBacksideFacingViewer())
            ctx->setAlpha(0); 
        else 
            ctx->setAlpha(opacity);
          
        paintContents(ctx);    
     }

    //recursively paint any children
    for (int i = 0; i < children().size(); ++i) {
        children()[i]->paint(ctx, preserves3D, opacity);
    }
    ctx->restore();
}


bool LayerWiiU::descendantDrawsContent()
{
    for (size_t i = 0; i < m_children.size(); ++i) {
        if (m_children[i]->drawsContent() || m_children[i]->descendantDrawsContent())
            return true;
    }
    return false;
}

void sortLayers(Vector<RefPtr<LayerWiiU> >::iterator, Vector<RefPtr<LayerWiiU> >::iterator, void*)
{
    // Currently we don't use z-order to decide what to paint, so there's no need to actually sort LayerWiiUs.
}

}
#endif // USE(ACCELERATED_COMPOSITING)
