/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*

	Video Playback for WiiU with Mobiclip's MVPlayer
	(high level interface with Webkit)
	
	make webcore_wiiu/fast
*/
#include "config.h"

#if 1//ENABLE(VIDEO) && PLATFORM(WIIU)

// Import wiiu's curl to define symbols like CURLcode, CURLoption ...
#include "SocketStreamHandlePrivate.h"
#include <curl/curl.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
// Import WKgfx::Matrix
#include "GraphicsContextPlatformPrivate.h"
#include <wiiu/MVPlayer_wiiu.h>
#include <pthread.h>
#include <wiiu/thread_ctrl.h>


//--------------------------------------------------------------------
#include <PlatformSSL.h>
#include <PlatformDefines.h>
#include <PlatformMemory.h>
using namespace LibPlatform;

#define CURLOPT_NSSL_CONTEXT (CURLoption)210
#define CURLOPT_NSSL_PEER_VERIFY_OPT (CURLoption)211
#define NSSL_VERIFY_ALL 7

#include "wksup/wksupFS.h" // FIXME: refactor SSLRootCA read so this isn't necessary!


// --------------------------------------------------------------------


#include "MediaPlayerPrivateMobiclip.h"
#include "TimeRanges.h"

#include "Logging.h"
#include "HTMLMediaElement.h"
#include "FrameLoaderClient.h"
#include "FrameLoader.h"
#include "Frame.h"
#include "WebViewWiiU.h"
#include "FrameLoaderClientWiiU.h"

#include <wtf/MainThread.h>

#define LOG_MediaPlayerPrivateMobiclip 0

// --------------------------------------------------------------------
// --------------------------------------------------------------------

namespace WebCore {

// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Create an instance of this media engine
*/
PassOwnPtr<MediaPlayerPrivateInterface> MediaPlayerPrivateMobiclip::create(MediaPlayer* player)
{
	MediaPlayerPrivateMobiclip* o = new MediaPlayerPrivateMobiclip(player);
	// If memory allocation fails (e.g. the default heap is all taken) bail out now
	if( !o->m_v )
	{
		delete o;
		return PassOwnPtr<MediaPlayerPrivateInterface>();
	}
    return adoptPtr(o);
}

// --------------------------------------------------------------------
/**
	Return the supported MIME types of this media engine
*/
void MediaPlayerPrivateMobiclip::getSupportedTypes(HashSet<String>& types)
{
	int n = MVPlayer::GetSupportedTypesCount();
	while( n-- )
	{
		const char* stype = MVPlayer::GetSupportedType(n);
		String s(stype);
		//s = s.lower();
		if( !types.contains(s) )
			types.add(s);
	}
}

// --------------------------------------------------------------------
/**
	Check if we support a given type
*/
MediaPlayer::SupportsType MediaPlayerPrivateMobiclip::supportsType(const String& type, const String& codecs)
{
	CString ctype = type.utf8();
	const char* stype = ctype.data();
	CString ccodecs = codecs.utf8();
	const char* scodecs = ccodecs.data();

	int confidence = MVPlayer::CanPlayType( stype, scodecs );
	MediaPlayer::SupportsType supported = MediaPlayer::IsNotSupported;
	if( confidence > 0 )
	{
		supported = ( confidence > 50 ) ? MediaPlayer::IsSupported : MediaPlayer::MayBeSupported;
	}

	return supported;
}

// --------------------------------------------------------------------
/**
	Check if we support a given type by Extension
*/
const char* MediaPlayerPrivateMobiclip::GetTypeFromExtension( const char* fileExtension )
{
	return MVPlayer::GetTypeFromExtension(fileExtension);
}


// --------------------------------------------------------------------
/**
	Register the media engine with Webkit
*/
void MediaPlayerPrivateMobiclip::registerMediaEngine(MediaEngineRegistrar registrar)
{
    registrar(create, getSupportedTypes, supportsType, 0, 0, 0);
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

class WiiUMediaPlayerObserver : public MVPlayerObserver
{
	unsigned int				m_refCount;
	MediaPlayerPrivateMobiclip* m_mediaPlayer;

public:

	WiiUMediaPlayerObserver( MediaPlayerPrivateMobiclip* mediaPlayer ):
		m_refCount(1),
		m_mediaPlayer(mediaPlayer)
	{
	}

private:

	~WiiUMediaPlayerObserver()
	{
		m_mediaPlayer = 0;
	}

public:

	unsigned int AddRef()
	{
		m_refCount ++;
		return m_refCount;
	}

	unsigned int Release()
	{
		m_refCount --;
		if( m_refCount == 0 )
		{
			delete this;
		}

		return m_refCount;
	}

public:

	void SetMediaPlayer( MediaPlayerPrivateMobiclip* mediaPlayer )
	{
		m_mediaPlayer = mediaPlayer;
	}

private:

	static void staticStateChange(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnStateChange();
        }

		This->Release();
	}

	static void staticBufferingUpdate(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnBufferingUpdate();
        }

		This->Release();
	}

	static void staticFrameUpdate(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnFrameUpdate();
        }

		This->Release();
	}

public:

	virtual void OnStateChange( MVPlayer* from, MVPlayerState newState )
	{
		AddRef();
		callOnMainThread(WiiUMediaPlayerObserver::staticStateChange, this);
	}

	virtual void OnBufferingUpdate( MVPlayer* from, int percent )
	{
		AddRef();
		m_mediaPlayer->setBufferingPercentage( percent );
		callOnMainThread(WiiUMediaPlayerObserver::staticBufferingUpdate, this);
	}

	virtual void OnFrameUpdate( MVPlayer* from )
	{
		AddRef();
		callOnMainThread(WiiUMediaPlayerObserver::staticFrameUpdate, this);

	}
};

class WiiUMVHook : public MVHttpHook//, public MVThreadHook
{
	unsigned int	m_refCount;

	u8* SSLRootCA;
	u32 SSLRootCASize;

	LibPlatform::SSL::SSLContextHandle SSLCtx;

	CString			m_userAgent;

public:

	WiiUMVHook():
		m_refCount(1),
		m_userAgent()
	{
		SSLCtx = -1;
	}

	~WiiUMVHook()
	{
		if(SSLCtx != -1)
		{
			LibPlatform::SSL::DestroyContext(SSLCtx);
			SSLCtx = -1;
		}
	}

private:

	void registerRootCA(const char* in_pem, int in_len)
	{
		if (SSLCtx < 0)
			return;

		// ToDo:
		// In SDK 2.06.02 calling NSSLAddServerPKIExternal() after some NSSLAddServerPKIExternal() return failure
		// seems to cause a program to freeze.
		// So, once NSSLAddServerPKIExternal() returns faulure, 
		// NSSLAddServerPKIExternal() is no longer called in this function.
		static bool failed = false;
		if (failed)
			return;

		X509 *cert = 0;
		BIO *bio = 0;
		unsigned char *buf = 0;
		unsigned char *buf_start = 0;
		int len;
		int rval; 

		bio = BIO_new_mem_buf((void *)in_pem, in_len);
		if (!bio)
			return;

		cert = PEM_read_bio_X509(bio, NULL, 0, NULL);
		if (!cert)
			goto error_end;

		len = i2d_X509(cert, NULL);

		if (len < 0)
			goto error_end;

		buf_start = (unsigned char*)LibPlatform::Memory::AllocFromDefaultHeapEx(len, LibPlatform::Memory::GetPPCIOBufferAlign());
		if (!buf_start)
			goto error_end;

		buf = buf_start;
		len = i2d_X509(cert, &buf);

		rval = LibPlatform::SSL::AddServerPKIExternal(SSLCtx, (const u8*)buf_start, len, LibPlatform::SSL::SSL_CERT_TYPE_DER);
		if (rval < 0)
		{
			LOG( Media,"ERROR: Could not add server PKI\n", rval );
		}

		LibPlatform::Memory::FreeToDefaultHeap(buf_start);
		X509_free(cert);
		BIO_free(bio);

		return;

	error_end:
		if (buf_start) {
			LibPlatform::Memory::FreeToDefaultHeap(buf_start);
		}
		if (cert) {
			X509_free(cert);
		}
		if (bio) {
			BIO_free(bio);
		}
	}

public:

	void SetUserAgent( const CString& userAgent )
	{
		m_userAgent = userAgent;
	}

	int Init()
	{
		int e = LibPlatform::SSL::Init();
		if( e )
		{
			LOG( Media,"NSSL init failed %x. Check that cos.xml grants permission NSEC\n", e );
			return e;
		}

		SSLCtx = LibPlatform::SSL::CreateContext(LibPlatform::SSL::SSL_SSLVERSION_SSLv3);
		if (SSLCtx < 0)
		{
			LOG( Media,"ERROR: NSSL context creation failed\n" );
			return SSL_RVAL_INVALID_NSSL_CTX;
		}

		SSLRootCA = (u8*)wksupFSSimpleRead("/vol/content/app/certificate/ca-bundle.pem", &SSLRootCASize);
		if( SSLRootCASize )
		{
			int sum = 0;
			char* pos = (char *)(SSLRootCA);
			while (sum + 52 < SSLRootCASize)
			{
				char *outerBegin = strstr(pos, "-----BEGIN CERTIFICATE-----");
				if (!outerBegin)
				  break;
				char *innerBegin = outerBegin + 27;

				char *innerEnd = strstr(innerBegin, "-----END CERTIFICATE-----");
				if (!innerEnd)
				  break;
				char *outerEnd = innerEnd + 25;

				int innerLength = innerEnd - innerBegin;
				int outerLength = outerEnd - outerBegin;

				registerRootCA(outerBegin, outerLength);

				sum += outerLength;
				pos += outerLength;
			}

			LibPlatform::Memory::FreeToDefaultHeap( SSLRootCA );
		}
		return e;
	}

public:

	unsigned int AddRef()
	{
		m_refCount ++;
		return m_refCount;
	}

	unsigned int Release()
	{
		m_refCount --;
		if( m_refCount == 0 )
		{
			delete this;
			MV_HttpHook = 0;
		}

		return m_refCount;
	}

public:

	int OnHttpRequest( CURL* easyRequest, const char* uri )
	{
		const char* userAgent = m_userAgent.data();
		if( userAgent )
		{
			Call_curl_easy_setopt( easyRequest, CURLOPT_USERAGENT, userAgent );
		}

		if ( SSLCtx != -1)
		{
			Call_curl_easy_setopt(easyRequest, CURLOPT_NSSL_CONTEXT, (long)SSLCtx);
			Call_curl_easy_setopt(easyRequest, CURLOPT_NSSL_PEER_VERIFY_OPT, (long)NSSL_VERIFY_ALL);
			//Call_curl_easy_setopt(easyRequest, CURLOPT_NSSL_PEER_VERIFY_OPT, (long)NSSL_VERIFY_NONE);
		}

		return 0;
	}

	void OnHttpError( CURLM* multi, CURL* easyRequest, const char* uri, curl_socket_t fd )
	{
	}
};

// --------------------------------------------------------------------------
/**
 * Returns the webview holding this document.
 * We use this to know if we are displaying on GamePad or TV.
 */
static WebView* GetWebView( HTMLMediaElement* mediaElement )
{
	Document* document = mediaElement->document();
	Frame* frame = document ? document->frame() : 0;
	FrameLoader* frameLoader = frame ? frame->loader() : 0;
	FrameLoaderClient* frameLoaderClient = frameLoader ? frameLoader->client() : 0;
	return frameLoaderClient ? static_cast<FrameLoaderClientWiiU*>(frameLoaderClient)->webView() : 0;
}

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateMobiclip::MediaPlayerPrivateMobiclip(MediaPlayer* player)
	: m_iSeekPerformed(false)
	, m_fgCurrentTimeSecs(0.0f)
	, m_bufferingPercentage(0)
	, m_isRemote(false)
	, m_isLocal(false)
	, m_isInBackground(false)
	, m_player(player)
	, m_preLoadMode(MediaPlayer::Auto)
	, m_networkState(MediaPlayer::Empty)
	, m_readyState(MediaPlayer::HaveNothing)
	, m_size(IntSize())
{
	m_v = MVPlayer::Create();
	m_vObserver = new WiiUMediaPlayerObserver(this);
	ProcessSwitchStart();

	if( MV_HttpHook == 0 )
	{
		MV_HttpHook = new WiiUMVHook;
		static_cast<WiiUMVHook*>(MV_HttpHook)->Init();
	}
	else
	{
		static_cast<WiiUMVHook*>(MV_HttpHook)->AddRef();
	}
}

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateMobiclip::~MediaPlayerPrivateMobiclip()
{
	ProcessSwitchEnd();
	m_v->SetObserver( 0 );
	static_cast<WiiUMediaPlayerObserver*>(m_vObserver)->SetMediaPlayer( 0 );
	static_cast<WiiUMediaPlayerObserver*>(m_vObserver)->Release();

	static_cast<WiiUMVHook*>(MV_HttpHook)->Release();

	MVPlayer::Destroy( m_v );
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setPreload(MediaPlayer::Preload preLoad)
{
	m_preLoadMode = preLoad;
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::load(const String& url)
{
	m_mediaURL = url;

	CString curl = url.utf8();
	const char* surl = curl.data();

	m_networkState	= MediaPlayer::Empty;
	m_readyState	= MediaPlayer::HaveNothing;

	m_isRemote		= false;
	m_isLocal		= false;
	m_loadType		= MediaPlayer::Unknown;

	if( strncmp(surl, "http", 4) == 0 )
	{
		HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( m_player->mediaPlayerClient() );
		Document* document = mediaElement->document();
		Frame* frame = document ? document->frame() : 0;
		FrameLoader* frameLoader = frame ? frame->loader() : 0;

		if( frameLoader )
		{
			const KURL& src = mediaElement->currentSrc();

			CString userAgent = frameLoader->userAgent(src).utf8();

			static_cast<WiiUMVHook*>(MV_HttpHook)->SetUserAgent( userAgent );
		}
		m_isRemote = true;
	}
	else
	{
		m_isLocal = (strncmp(surl, "file", 4) == 0);
		if( m_isLocal )
		{
			m_loadType = MediaPlayer::StoredStream;
		}
	}

	m_bufferingPercentage = 0;

	m_v->SetSource( surl );
	m_v->SetObserver( m_vObserver );

	m_v->viewTV.volumePercent = 0;
	m_v->viewGamePad.volumePercent = 0;

	if( m_preLoadMode != MediaPlayer::None )
	{
#if LOG_MediaPlayerPrivateMobiclip==1
		OSReport( "[MediaPlayerPrivateMobiclip] AutoStart\n" );
#endif
		m_v->Prepare();
	}
}


// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::cancelLoad()
{
	m_v->Abort();
	newNetworkState(MediaPlayer::Empty);
	newReadyState(MediaPlayer::HaveNothing);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::prepareToPlay()
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] prepareToPlay\n" );
#endif
	m_v->Prepare();
}

#if USE(ACCELERATED_COMPOSITING)
// --------------------------------------------------------------------------
/**
*/
PlatformLayer* MediaPlayerPrivateMobiclip::platformLayer() const
{
	return m_videoLayer.get();
}
#endif

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::play()
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] play\n" );
#endif
	m_v->Play();
}

// --------------------------------------------------------------------------
/**
	Pause playback
*/
void MediaPlayerPrivateMobiclip::pause()
{
	bool isPlaying = (m_v->state == MVPLAYER_STATE_PLAYING);
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] pause (forwarded:%d)\n", isPlaying );
#endif
	if( isPlaying )
		m_v->Pause();
}

// --------------------------------------------------------------------------
/**
*/
IntSize MediaPlayerPrivateMobiclip::naturalSize() const
{
    if (!hasVideo() || m_readyState < MediaPlayer::HaveMetadata)
        return IntSize();
    return IntSize(m_v->videoWidth, m_v->videoHeight);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasVideo() const
{
	return m_v && (m_v->videoWidth > 0);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasAudio() const
{
	return m_v && (m_v->audioFrequency > 0);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setVisible(bool bVisible)
{
	WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

	MVView* playerView = &m_v->viewTV; // By default render to TV
	if( webView )
	{
		// If we can query information on display device, make a better choice
		playerView = webView->GetDisplayType() == WebView::TV_DISPLAY_TYPE ? &m_v->viewTV : &m_v->viewGamePad;
	}

	playerView->visibility = (bVisible ? MVPLAYER_VISIBILITY_VIDEO_ONLY : MVPLAYER_VISIBILITY_HIDDEN);
#if USE(ACCELERATED_COMPOSITING)    
    if (bVisible)
        setUpVideoRendering();
#endif
}

//---------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateMobiclip::duration() const
{
	return (float)m_v->mediaDurationSecs;
}



//---------------------------------------------------------------------------
/**
	Get current time of playback in seconds
	
	@note	This will return 0.0 if no data is known
*/
float MediaPlayerPrivateMobiclip::currentTime() const
{
	float currentTime = (float)m_v->GetCurrentTimeSecs();
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] currentTime %f\n", currentTime );
#endif
	return currentTime;
}



//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::seek(float time)
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] seek %f\n", time );
#endif
	m_v->SeekTo( time );

	if( m_v->state != MVPLAYER_STATE_SEEKING )
	{
		// We didn't seek because data were already available
		newReadyState(MediaPlayer::HaveCurrentData);
	}
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::seeking() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] seeking\n" );
#endif
    return m_v->state == MVPLAYER_STATE_SEEKING;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setRate(float)
{
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::paused() const
{
	bool isPaused = (m_v->state != MVPLAYER_STATE_PLAYING);
	//bool isPaused = (m_v->state == MVPLAYER_STATE_PAUSED);
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] paused %d\n", isPaused );
#endif
    return isPaused;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setVolume(float pcent)
{
	WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

	MVView* playerView = &m_v->viewTV; // By default render to TV
	if( webView )
	{
		// If we can query information on display device, make a better choice
		playerView = webView->GetDisplayType() == WebView::TV_DISPLAY_TYPE ? &m_v->viewTV : &m_v->viewGamePad;
	}

	playerView->volumePercent = pcent*100;
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::NetworkState MediaPlayerPrivateMobiclip::networkState() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] networkState %d\n", m_networkState );
#endif
    return(m_networkState);
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::ReadyState MediaPlayerPrivateMobiclip::readyState() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] readyState %d\n", m_readyState );
#endif
    return(m_readyState);
}

//-----------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateMobiclip::maxTimeSeekable() const
{
	return duration();
}

//---------------------------------------------------------------------------
/**
*/
PassRefPtr<TimeRanges> MediaPlayerPrivateMobiclip::buffered() const
{
    RefPtr<TimeRanges> timeRanges = TimeRanges::create();
	double start, end;
	m_v->GetBufferedRange(start,end);
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] buffered Range %lf %lf\n", start, end );
#endif
    timeRanges->add( start, end );
    return timeRanges.release();
}

//---------------------------------------------------------------------------
/**
*/
unsigned MediaPlayerPrivateMobiclip::bytesLoaded() const
{
	return 0;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setSize(const IntSize& size)
{
    m_size = size;
#if USE(ACCELERATED_COMPOSITING)
		if( m_videoLayer.get() && supportsAcceleratedRendering() )
			m_videoLayer->setNeedsDisplay( IntRect( 0, 0, size.width(), size.height()));
#endif
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::paint(GraphicsContext* context, const IntRect& rect)
{
	// Please see comment at method: MediaPlayerPrivateMobiclip::naturalSize()
	if (context->paintingDisabled() || !hasVideo())
        return;

	if( !rect.isEmpty() )
	{
		GraphicsContextPlatformPrivate* g = context->platformContext();

		g->makeCurrent();
		g->beginCustomDraw();

		WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

		MVView* playerView = &m_v->viewTV; // By default render to TV
		if( webView )
		{
			// If we can query information on display device, make a better choice
			playerView = webView->GetDisplayType() == WebView::TV_DISPLAY_TYPE ? &m_v->viewTV : &m_v->viewGamePad;
		}

		float x = rect.x();
		float y = rect.y();
		float w = rect.width();
		float h = rect.height();
		const WKgfx::Matrix& viewM = g->getViewMatrix();

		playerView->InitTransform( x,y, w,h, viewM.m );

		playerView->projection = g->getProjMatrix();

		playerView->Draw();

		g->endCustomDraw(10);
	}
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasSingleSecurityOrigin() const
{
    // We tell to disallow resources that come from different origins so we are sure all media is single origin
    return true;
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::newReadyState(MediaPlayer::ReadyState newState)
{
	bool bRet = (m_readyState != newState);

	if( bRet )
	{
		m_readyState = newState;
		m_player->readyStateChanged();
	}

	return(bRet);
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::newNetworkState(MediaPlayer::NetworkState newState)
{
	bool bRet = (m_networkState != newState);

	if( bRet )
	{
		m_networkState = newState;
		m_player->networkStateChanged();
	}

	return(bRet);
}

//-----------------------------------------------------------------------------
/**
*/
MediaPlayer::MovieLoadType MediaPlayerPrivateMobiclip::movieLoadType() const
{
	//Unknown, Download, StoredStream, LiveStream 
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] movieLoadType %d\n", m_loadType );
#endif

	return m_loadType;
}


#if USE(ACCELERATED_COMPOSITING)

//-----------------------------------------------------------------------------
/**
*/ 
// whether accelerated rendering is supported by the media engine for the current media.
bool MediaPlayerPrivateMobiclip::supportsAcceleratedRendering() const
{
    return true;
}

//-----------------------------------------------------------------------------
/**
*/ 
// called when the rendering system flips the into or out of accelerated rendering mode.
void MediaPlayerPrivateMobiclip::acceleratedRenderingStateChanged()
{ 
    setUpVideoRendering(); 
} 

bool MediaPlayerPrivateMobiclip::isReadyForVideoSetup() const
{
    return m_readyState >= MediaPlayer::HaveMetadata && m_player->visible();
}

MediaPlayerPrivateMobiclip::MediaRenderingMode MediaPlayerPrivateMobiclip::currentRenderingMode() const
{
    if (platformLayer())
       return MediaRenderingToLayer;

    if (m_player->mediaPlayerClient())
        return MediaRenderingToContext;

    return MediaRenderingNone;
}

MediaPlayerPrivateMobiclip::MediaRenderingMode MediaPlayerPrivateMobiclip::preferredRenderingMode() const
{
    if (!m_player->visible() || !m_player->frameView())
        return MediaRenderingNone;

    if (supportsAcceleratedRendering() && m_player->mediaPlayerClient()->mediaPlayerRenderingCanBeAccelerated(m_player))
         return MediaRenderingToLayer;

    return MediaRenderingToContext;
}

void MediaPlayerPrivateMobiclip::setUpVideoRendering()
{
    if (!isReadyForVideoSetup())
        return;

    MediaRenderingMode currentMode = currentRenderingMode();
    MediaRenderingMode preferredMode = preferredRenderingMode();
    if (currentMode == preferredMode && currentMode != MediaRenderingNone)
       return;

    if (currentMode != MediaRenderingNone) {  
        tearDownVideoRendering();
    }
    if (preferredMode == MediaRenderingToLayer) {
        createVideoLayer();   
    }

    // If using a movie layer, inform the client so the compositing tree is updated.
    if (currentMode == MediaRenderingToLayer || preferredMode == MediaRenderingToLayer) {
          m_player->mediaPlayerClient()->mediaPlayerRenderingModeChanged(m_player);
   }

}

void MediaPlayerPrivateMobiclip::createVideoLayer()
{
    if (m_videoLayer) {
        return;
    }
     m_videoLayer = VideoLayerWiiU::create(0); 
     m_videoLayer->setPlayer(m_player); 
}

void MediaPlayerPrivateMobiclip::tearDownVideoRendering()
{
    if (platformLayer()) {
        m_videoLayer.release(); 
    }
}
#endif
//---------------------------------------------
//---------------------------------------------
//---------------------------------------------

void MediaPlayerPrivateMobiclip::OnFrameUpdate()
{
	m_player->repaint();

#if USE(ACCELERATED_COMPOSITING)
		if( m_videoLayer.get() && supportsAcceleratedRendering() )
			m_videoLayer->setNeedsDisplay( IntRect( 0, 0, m_videoLayer->bounds().width(), m_videoLayer->bounds().height()) );
#endif

	double startSecs, endSecs;
	m_v->GetBufferedRange( startSecs, endSecs );

	if( (m_v->mediaDurationSecs > 0) && (endSecs == m_v->mediaDurationSecs) )
	{
		m_bufferingPercentage = 100;
		OnBufferingUpdate();
	}
}

void MediaPlayerPrivateMobiclip::OnBufferingUpdate()
{
	double startSecs=0, endSecs=0;
	m_v->GetBufferedRange( startSecs, endSecs );

	if( (m_bufferingPercentage == 100 || endSecs-startSecs > 3) && (m_v->state == MVPLAYER_STATE_PAUSED || m_v->state == MVPLAYER_STATE_PLAYING) )
	{
		newNetworkState(MediaPlayer::Loaded);
		newReadyState(MediaPlayer::HaveEnoughData); // Enough to reach the end without stalling
	}
}

void MediaPlayerPrivateMobiclip::OnStateChange()
{
	MVPlayerState state = m_v->state;
#if LOG_MediaPlayerPrivateMobiclip==1
	OSReport( "[MediaPlayerPrivateMobiclip] OnStateChange %d\n", state );
#endif

	if( m_isInBackground )
	{
		// We are recovering from background mode
		// The video is loading and it will switch to paused state once it's done
		// Then we have to seek to the saved playback position and wait again for paused state (i.e. end of seek)
		// see ProcessSwitchResumeState
		if( state == MVPLAYER_STATE_PAUSED )
		{
			if( m_iSeekPerformed )
			{
				// We are at the right playback position
				// Mobiclip's media player is ready to resume
				// Tell webkit's HTMLMediaElement to quit freeze state
				HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( m_player->mediaPlayerClient() );
				Document* document = mediaElement->document();
				document->resumeActiveDOMObjects();
				m_isInBackground = false;
			}
			else
			{
				if( m_fgCurrentTimeSecs > 0.0f )
				{
					seek(m_fgCurrentTimeSecs);
				}
				m_iSeekPerformed = true;
			}
		}
	}
	else
	{
		switch( state )
		{
		case MVPLAYER_STATE_SEEKING:
		case MVPLAYER_STATE_BUFFERING:
			newReadyState(MediaPlayer::HaveCurrentData);	// Data at playback position is available
			break;

		case MVPLAYER_STATE_PLAYING:
		case MVPLAYER_STATE_PAUSED:
			if( m_isRemote )
			{
				m_loadType = (duration() > 0.0) ? MediaPlayer::Download : MediaPlayer::LiveStream;
			}

			if( m_loadType == MediaPlayer::LiveStream || m_loadType == MediaPlayer::StoredStream )
			{
				newNetworkState(MediaPlayer::Loaded);
				newReadyState(MediaPlayer::HaveEnoughData); // Enough to reach the end without stalling
			}
			else
			{
				newReadyState( MediaPlayer::HaveFutureData);	// Have data to play for a while
			}
			break;

		case MVPLAYER_STATE_PREPARING:
			newNetworkState(MediaPlayer::Loading);
			newReadyState(MediaPlayer::HaveMetadata);	// Metadata loaded
			break;
		}
	}

	if( state == MVPLAYER_STATE_ERROR )
	{
		newReadyState(MediaPlayer::HaveNothing);
		switch(m_v->error)
		{
			case MVPLAYER_ERROR_ABORTED:
			case MVPLAYER_ERROR_NETWORK:
				newNetworkState(MediaPlayer::NetworkError);
				break;
			case MVPLAYER_ERROR_DECODE:
				newNetworkState(MediaPlayer::DecodeError);
				break;
			case MVPLAYER_ERROR_SRC_NOT_SUPPORTED:
				newNetworkState(MediaPlayer::FormatError);
				break;
		}
	}

}

int MediaPlayerPrivateMobiclip::ProcessSwitchStart()
{
	pthread_mutex_init( &m_ProcessSwitchMutex, NULL );
	pthread_cond_init( &m_ProcessSwitchCond, NULL );

	pthread_create( &m_ProcessSwitchThread, NULL, ProcessSwitchLoopStatic, this );
}

void MediaPlayerPrivateMobiclip::ProcessSwitchEnd()
{
	pthread_cond_signal( &m_ProcessSwitchCond );
	void* ignore;
	pthread_join( m_ProcessSwitchThread, &ignore );

	pthread_mutex_destroy( &m_ProcessSwitchMutex );
	pthread_cond_destroy( &m_ProcessSwitchCond );
}

void* MediaPlayerPrivateMobiclip::ProcessSwitchLoopStatic( void* arg )
{
	((MediaPlayerPrivateMobiclip*)arg)->ProcessSwitchLoop();
	return 0;
}


void MediaPlayerPrivateMobiclip::ProcessSwitchAbortStatic(void* context)
{
	MediaPlayerPrivateMobiclip* This = static_cast<MediaPlayerPrivateMobiclip*>(context);
    pthread_mutex_lock(&This->m_ProcessSwitchMutex);

    // Save the playback position to restore it if we come back in foreground
    // see OnStateChange
    This->m_fgCurrentTimeSecs = This->currentTime();
    This->m_iSeekPerformed = false;

    //
    // Use available method from cache management to freeze Webkit's HTLMMediaElement
    //
    HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( This->m_player->mediaPlayerClient() );
    Document* document = mediaElement->document();
    document->suspendActiveDOMObjects(ActiveDOMObject::DocumentWillBecomeInactive);

    //
    // Free all resources used by Mobiclip's mediaplayer (end all threads)
    //
    This->m_v->Abort();

    This->m_isInBackground = true;

	pthread_cond_signal( &This->m_ProcessSwitchCond );
    pthread_mutex_unlock(&This->m_ProcessSwitchMutex);
}

void MediaPlayerPrivateMobiclip::ProcessSwitchResumeState(void* context)
{
	MediaPlayerPrivateMobiclip* This = static_cast<MediaPlayerPrivateMobiclip*>(context);

    pthread_mutex_lock(&This->m_ProcessSwitchMutex);

    CString curl = This->m_mediaURL.utf8();
    const char* surl = curl.data();

    //
    // We have to restart Webkit's HTLMMediaElement
    // Before that Mobiclip's mediaplayer must be restored
    // See ProcessSwitchAbortStatic
    // See OnStateChange
    //
    This->m_v->SetSource( surl );
    This->m_v->Prepare(); // This call will load the video and put the mediaplayer in pause state

	pthread_cond_signal( &This->m_ProcessSwitchCond );
    pthread_mutex_unlock(&This->m_ProcessSwitchMutex);
}


void MediaPlayerPrivateMobiclip::ProcessSwitchLoop()
{
	// Register us for process switch notification
	//bool bBackground = shim_IsBackground();
	shim_RegisterSyncThread(NULL);

	pthread_mutex_lock(&m_ProcessSwitchMutex);
	for(;;)
	{
		struct timeval now;
		gettimeofday(&now, NULL);
		struct timespec until;
		until.tv_sec  = now.tv_sec;
		until.tv_nsec = now.tv_usec * 1000;
		until.tv_nsec += 15 * 1e6;

		int e = pthread_cond_timedwait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex, &until );
		if( e!=ETIMEDOUT )
			break;

		int switching = shim_PollForSwitch();
		switch(switching)
		{
			case SHIM_SWITCH_TO_BG:
			case SHIM_SWITCH_SHUTDOWN:
				callOnMainThread(ProcessSwitchAbortStatic, this);
				pthread_cond_wait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex );
				shim_AcknowledgeSwitch();
				break;
			case SHIM_SWITCH_TO_FG:
				callOnMainThread(ProcessSwitchResumeState, this);
				pthread_cond_wait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex );
				shim_AcknowledgeSwitch();
				break;
			default:
				break;
		}
	}
	pthread_mutex_unlock(&m_ProcessSwitchMutex);

	shim_UnregisterSyncThread(NULL);
}



}; // namespace WebCore

#endif
