/*
 * Copyright (C) 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Collabora, Ltd. All rights reserved.
 * Copyright (C) 2008 Kenneth Rohde Christiansen.
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (C) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FileSystem.h"

#include "NotImplemented.h"

#include <dirent.h>
#if PLATFORM(WIIU)
#include <dlfcn.h>
#include <shim.h>
#else
#include <stdio.h> // snprintf
#endif
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <wtf/text/CString.h>
#include <wtf/CurrentTime.h>

namespace WebCore {

bool fileExists(const String& path)
{
    #if PLATFORM(WIIU)
    if (path.isNull())
        return false;

    if (!path.utf8().data() || path.utf8().data()[0] == '\0')
        return false;

    struct stat fileInfo;

    // stat(...) returns 0 on successful stat'ing of the file, and non-zero in any case where the file doesn't exist or cannot be accessed
    String nullTerminated(String(path).charactersWithNullTermination());
    return !stat(nullTerminated.utf8().data(), &fileInfo);
    #else
    notImplemented();
    return false;
    #endif
}

bool deleteFile(const String& path)
{
    #if PLATFORM(WIIU)
    String nullTerminated(String(path).charactersWithNullTermination());
    // unlink(...) returns 0 on successful deletion of the path and non-zero in any other case (including invalid permissions or non-existent file)
    return !unlink(nullTerminated.utf8().data());
    #else
    notImplemented();
    return false;
    #endif
}

PlatformFileHandle openFile(const String& path, FileOpenMode mode)
{
    PlatformFileHandle handle = invalidPlatformFileHandle;

    #if PLATFORM(WIIU)
    if (path.isNull())
        return invalidPlatformFileHandle;

    int platformFlag = 0;
    if (mode == OpenForRead)
        platformFlag |= O_RDONLY;
    else if (mode == OpenForWrite)
        platformFlag |= (O_WRONLY | O_CREAT | O_TRUNC);

    String nullTerminated(String(path).charactersWithNullTermination());
    handle = open(nullTerminated.utf8().data(), platformFlag, 0x666);
    return handle;
    #else
    notImplemented();
    return handle;
    #endif
}

void closeFile(PlatformFileHandle& handle)
{
    #if PLATFORM(WIIU)
    if (isHandleValid(handle)) {
        close(handle);
        handle = invalidPlatformFileHandle;
    }
    #else
    notImplemented();
    #endif
}

long long seekFile(PlatformFileHandle handle, long long offset, FileSeekOrigin origin)
{
    #if PLATFORM(WIIU)
    int whence = SEEK_SET;
    switch (origin) {
    case SeekFromBeginning:
        whence = SEEK_SET;
        break;
    case SeekFromCurrent:
        whence = SEEK_CUR;
        break;
    case SeekFromEnd:
        whence = SEEK_END;
        break;
    default:
        ASSERT_NOT_REACHED();
    }
    return static_cast<long long>(lseek(handle, offset, whence));
    #else
    notImplemented();
    return -1;
    #endif
}

bool truncateFile(PlatformFileHandle handle, long long offset)
{
    #if PLATFORM(WIIU)
    // ftruncate returns 0 to indicate the success.
    return !ftruncate(handle, offset);
    #else
    notImplemented();
    return false;
    #endif
}

int writeToFile(PlatformFileHandle handle, const char* data, int length)
{
    #if PLATFORM(WIIU)
    do {
        int bytesWritten = write(handle, (void *) data, static_cast<size_t>(length));
        if (bytesWritten >= 0)
            return bytesWritten;
    } while (errno == EINTR);
    #else
    notImplemented();
    return -1;
    #endif
}

int readFromFile(PlatformFileHandle handle, char* data, int length)
{
    #if PLATFORM(WIIU)
    do {
        int bytesRead = read(handle, data, static_cast<size_t>(length));
        if (bytesRead >= 0)
            return bytesRead;
    } while (errno == EINTR);
    #else
    notImplemented();
    return -1;
    #endif
}

bool deleteEmptyDirectory(const String& path)
{
    #if PLATFORM(WIIU)
    String nullTerminated(String(path).charactersWithNullTermination());
    // unlink(...) returns 0 on successful deletion of the path and non-zero in any other case (including invalid permissions or non-existent file)
    return !unlink(nullTerminated.utf8().data());
    #else
    notImplemented();
    return false;
    #endif
}

bool getFileSize(const String& path, long long& result)
{
    #if PLATFORM(WIIU)
    if (!path.utf8().data() || path.utf8().data()[0] == '\0')
        return false;

    struct stat fileInfo;

    String nullTerminated(String(path).charactersWithNullTermination());
    if (stat(nullTerminated.utf8().data(), &fileInfo))
        return false;

    result = fileInfo.st_size;
    return true;
    #else
    notImplemented();
    return false;
    #endif
}

bool getFileModificationTime(const String& path, time_t& result)
{
    #if PLATFORM(WIIU)
    if (!path.utf8().data() || path.utf8().data()[0] == '\0')
        return false;

    struct stat fileInfo;

    String nullTerminated(String(path).charactersWithNullTermination());
    if (stat(nullTerminated.utf8().data(), &fileInfo))
        return false;

    result = fileInfo.st_mtime;
    return true;
    #else
    notImplemented();
    return false;
    #endif
}

String pathByAppendingComponent(const String& path, const String& component)
{
    #if PLATFORM(WIIU)
    if (path.isNull() && !component.isNull() ) {
        return component;
    }

    if (!path.isNull() && component.isNull()) {
        return path;
    }

    if (path.isNull() && component.isNull()) {
        return "";
    }


    if (path.endsWith("/"))
        return path + component;
    else
        return path + "/" + component;
    #else
    notImplemented();
    return String();
    #endif
}

bool makeAllDirectories(const String& path)
{
    #if PLATFORM(WIIU)
	unsigned int start = UINT_MAX;
	String pathStr(path);

	// Does the path end with a separator?
	unsigned endPos = pathStr.length();
	if (pathStr[endPos - 1] == '/')
		pathStr.truncate(endPos - 1);

	// Is this the base "save" volume path? We don't want to create that! If it's all empty: also escape!
	if (pathStr == "/vol/save/account" || pathStr == "/vol/save/common" || pathStr.length() == 0)
		return true;

	// Where is the last path separator?
	size_t lastDivPos = pathStr.reverseFind('/', start);

	// Any separator still found?
	if (lastDivPos != notFound)
	{
		// Yes. Pass the front part recursively back into this function...
		if (!makeAllDirectories(pathStr.substring(0, lastDivPos)))
			return false;
	}

	// Once here, we know all directories "above" exist. Make this one...
	int ret = mkdir(pathStr.substring(0, endPos).utf8().data(), 0x644);
	return (ret == 0 || errno == EEXIST);
    #else
    notImplemented();
    return false;
    #endif
}

String pathGetFileName(const String& path)
{
    #if PLATFORM(WIIU)
    String nullTerminated(String(path).charactersWithNullTermination());
    return nullTerminated.substring(nullTerminated.reverseFind('/') + 1);
    #else
    notImplemented();
    return String();
    #endif
}

String directoryName(const String& path)
{
    #if PLATFORM(WIIU)
    unsigned int start = UINT_MAX;
    size_t lastDivPos = path.reverseFind('/', start);
    if (lastDivPos == notFound) {
        return String();
    }
    String root = path.substring(0, lastDivPos);
    String nullTerminated(String(root).charactersWithNullTermination());
    return nullTerminated;
    #else
    notImplemented();
    return String();
    #endif
}

CString fileSystemRepresentation(const String& path)
{
    #if PLATFORM(WIIU)
    return "";
    #else
    notImplemented();
    return "";
    #endif
}

String openTemporaryFile(const String& prefix, PlatformFileHandle& handle)
{
    #if PLATFORM(WIIU)
    const int kMaxArgpathSize = shim_GetFSMaxArgPathSize();
    const char* tmpDir = "/vol/save/account/webkit/tmp";
    #else
    // FIXME?: maxiumum filename lengths for each platform? -TS-NSA 2012-04-12
    const int kMaxArgpathSize = 4096;
    const char *tmpDir = "";
    #endif
    char* buffer = new char[kMaxArgpathSize];

    String nullTerminated(String(prefix).charactersWithNullTermination());
    if (snprintf(buffer, kMaxArgpathSize, "%s/%s%d", tmpDir, nullTerminated.utf8().data(), (int)currentTime()) >= kMaxArgpathSize)
    {
        goto end;
    }

    handle = openFile(String(buffer),  OpenForWrite);
    if (handle == invalidPlatformFileHandle)
        goto end;

    delete[] buffer;
    return String::fromUTF8(buffer);

end:
    handle = invalidPlatformFileHandle;
    delete[] buffer;
    return String();
}

bool unloadModule(PlatformModule module)
{
    // caution, closing handle will make memory vanish and any remaining
    // timer, idler, threads or any other left-over will crash,
    // maybe just ignore this is a safer solution?
    notImplemented();
    return false;
    //return !dlclose(module);
}

static String storageDirectory()
{
    String path("/vol/save/account/webkit");
    String nullTerminated(String(path).charactersWithNullTermination());
    return nullTerminated;
}

static String cachedStorageDirectory()
{
    String path("/vol/save/account/webkit");
    String nullTerminated(String(path).charactersWithNullTermination());
    return nullTerminated;
}

String homeDirectoryPath()
{
    String path("/vol/save/account/webkit");
    String nullTerminated(String(path).charactersWithNullTermination());
    return nullTerminated;
}

String localUserSpecificStorageDirectory()
{
    return cachedStorageDirectory();
}

Vector<String> listDirectory(const String& path, const String& filter)
{
    #if PLATFORM(WIIU)
//
// NOTE: the code below currently does NOT adhere to the filter at all!
//
    Vector<String> entries;
    CString cpath = path.utf8();
    CString cfilter = filter.utf8();
    const int kMaxEntNameSize = shim_GetFSMaxEntNameSize();
    char* filePath = new char[kMaxEntNameSize];
    char* fileName;
    size_t fileNameSpace;
    DIR* dir;

    if (cpath.length() >= sizeof(filePath)) {
        delete [] filePath;
        return entries;
    }
    // loop invariant: directory part + '/'
    memcpy(filePath, cpath.data(), cpath.length());
    fileName = filePath + cpath.length();
    if (cpath.length() > 0 && filePath[cpath.length() - 1] != '/') {
        fileName[0] = '/';
        fileName++;
    }
    fileNameSpace = sizeof(filePath) - (fileName - filePath) - 1;

    dir = opendir(cpath.data());
    if (!dir)
    {
        delete[] filePath;
        return entries;
    }

    dirent* de;
    while (de = readdir(dir)) {
        size_t nameLen;
        nameLen = strlen(de->d_name);

		// Too long?
        if (nameLen >= fileNameSpace)
            continue;	// Yes...

        memcpy(fileName, de->d_name, nameLen + 1);
        entries.append(filePath);
    }
    closedir(dir);
    delete[] filePath;
    return entries;
    #else
    notImplemented();
    return Vector<String>();
    #endif
}

}
