/*
 * Copyright (C) 2006, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "PopupMenuWiiU.h"

#include "Chrome.h"
#include "ChromeClientWiiU.h"
#include "Frame.h"
#include "FrameView.h"
#include "NotImplemented.h"
#include "Page.h"
#include "PopupMenuClient.h"
#include "TextRun.h"

static const int maxPopupHeight = 360;

namespace WebCore {

PopupMenuWiiU::PopupMenuWiiU(PopupMenuClient* client)
    : m_popupClient(client)
    , m_view(0)
{
}

PopupMenuWiiU::~PopupMenuWiiU()
{
    // Tell client to destroy data related to this popup since this object is
    // going away.
    if (m_view)
        hide();
}

const int separatorPadding = 4;
const int separatorHeight = 1;
void PopupMenuWiiU::paint(GraphicsContext *ctx) 
{

    ASSERT(m_popupClient);
 
    int itemCount = m_popupClient->listSize();
    PopupMenuStyle itemStyle = m_popupClient->itemStyle(0); 
    const Font &itemFont = itemStyle.font();

    ctx->save();
    ctx->translate(-m_view->scrollXForFixedPosition(), -m_view->scrollYForFixedPosition());

    // listRect is the damageRect translated into the coordinates of the entire menu list (which is itemCount * m_itemHeight pixels tall)

    for (int index = 0; index < itemCount; index++) {
        Color optionBackgroundColor, optionTextColor;
        itemStyle = m_popupClient->itemStyle(index);
        /*if (index == focusedIndex()) {
            optionBackgroundColor = RenderTheme::defaultTheme()->activeListBoxSelectionBackgroundColor();
            optionTextColor = RenderTheme::defaultTheme()->activeListBoxSelectionForegroundColor();
        } else*/ {
            optionBackgroundColor = itemStyle.backgroundColor();
            optionTextColor = itemStyle.foregroundColor();
        }

        // itemRect is in client coordinates
        // 0 == m_scrollOffset
        IntRect itemRect(m_drawRect.x(), m_drawRect.y() + (index - 0) * m_itemHeight, m_drawRect.width(), m_itemHeight);
        // Draw the background for this menu item
        if (itemStyle.isVisible())
            ctx->fillRect(itemRect, optionBackgroundColor, ColorSpaceDeviceRGB);

        if (m_popupClient->itemIsSeparator(index)) {
            IntRect separatorRect(itemRect.x() + separatorPadding, itemRect.y() + (itemRect.height() - separatorHeight) / 2, itemRect.width() - 2 * separatorPadding, separatorHeight);
            ctx->fillRect(separatorRect, optionTextColor, ColorSpaceDeviceRGB);
            continue;
        }

        String itemText = m_popupClient->itemText(index);
            
        unsigned length = itemText.length();
        const UChar* string = itemText.characters();
        TextDirection direction = (itemText.defaultWritingDirection() == WTF::Unicode::RightToLeft) ? RTL : LTR;
        TextRun textRun(string, length, false, 0, 0, TextRun::AllowTrailingExpansion, direction);

        ctx->setFillColor(optionTextColor, ColorSpaceDeviceRGB);
/*
        if (m_popupClient->itemIsLabel(index)) {
            FontDescription d = itemFont.fontDescription();
            d.setWeight(d.bolderWeight());
            itemFont = Font(d, itemFont.letterSpacing(), itemFont.wordSpacing());
            itemFont.update(m_popupClient->fontSelector());
        }
*/
        // Draw the item text
        if (itemStyle.isVisible()) {
            int textX = itemRect.x() + m_popupClient->clientPaddingLeft() - m_popupClient->clientInsetLeft();
            if (textX < itemRect.x()) {
                textX = itemRect.x();
            }
            //if (RenderTheme::defaultTheme()->popupOptionSupportsTextIndent() && itemStyle.textDirection() == LTR)
//                textX += itemStyle.textIndent().calcMinValue(itemRect.width());
            int textY = itemRect.y() + itemFont.fontMetrics().ascent() + (itemRect.height() - itemFont.fontMetrics().height()) / 2;
            ctx->drawBidiText(itemFont, textRun, IntPoint(textX, textY));
        }
    }

    //Draw border lines
    ctx->setStrokeColor(itemStyle.foregroundColor(), ColorSpaceDeviceRGB);
    ctx->drawLine(IntPoint(m_drawRect.x(), m_drawRect.y()), IntPoint(m_drawRect.x(), m_drawRect.y() + m_drawRect.height()-1));
    ctx->drawLine(IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y()), IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y() + m_drawRect.height()-1));
    ctx->drawLine(IntPoint(m_drawRect.x(), m_drawRect.y() + m_drawRect.height()-1), IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y() + m_drawRect.height()-1));

    ctx->restore();
}

void PopupMenuWiiU::calculatePositionAndSize(const IntRect& rect, FrameView* view) 
{
    int itemCount = m_popupClient->listSize();
    m_itemHeight = m_popupClient->menuStyle().font().fontMetrics().height() + 1;
    int popupHeight = m_itemHeight * (itemCount);
    if (popupHeight > maxPopupHeight) {
        popupHeight = maxPopupHeight;
    }
    // The popup should show an integral number of items (i.e. no partial items should be visible)
    popupHeight -= popupHeight % m_itemHeight;

    // For now, assume we are keepig it the same width as the text portion of the combobox
    m_drawRect.setX(rect.x());
    m_drawRect.setY(rect.y() + rect.height()); 
    m_drawRect.setWidth(rect.width()); 
    m_drawRect.setHeight(popupHeight);  
}

void PopupMenuWiiU::show(const IntRect& rect, FrameView* view, int index)
{
    ASSERT(m_popupClient);
    m_view = view;

    calculatePositionAndSize(rect, view);
    ChromeClientWiiU* chromeClient = static_cast<ChromeClientWiiU*>(view->frame()->page()->chrome()->client());
    ASSERT(chromeClient);
    chromeClient->showPopup(this, m_drawRect);
    m_view->hostWindow()->invalidateContentsAndWindow(m_drawRect, false);
}

void PopupMenuWiiU::hide()
{
    if (m_view && m_popupClient) { 
        m_popupClient->popupDidHide();

        ChromeClientWiiU* chromeClient = static_cast<ChromeClientWiiU*>(m_view->frame()->page()->chrome()->client());
        ASSERT(chromeClient);
        chromeClient->hidePopup();
        m_view->hostWindow()->invalidateContentsAndWindow(IntRect(m_drawRect.x() - 1, m_drawRect.y(), m_drawRect.width() + 2, m_drawRect.height()), false);
    }
}

void PopupMenuWiiU::processEvent(int x, int y) 
{
    //should always be true
    if (m_popupClient) {
        int new_index = m_popupClient->selectedIndex();
        if (m_drawRect.contains(x, y)) {
            int potential_new_index = (y - m_drawRect.y()) / m_itemHeight;

            //this should be fine due to the contains() call above, but why not be safe...
            if (potential_new_index >= 0 && potential_new_index < m_popupClient->listSize()) {
                new_index = potential_new_index;
            }
        }

        m_popupClient->valueChanged(new_index);

        hide();
    }
}

void PopupMenuWiiU::updateFromElement()
{
    m_popupClient->setTextFromItem(m_popupClient->selectedIndex());
}

void PopupMenuWiiU::disconnectClient()
{
    if (m_view)
        hide();
    m_popupClient = 0;
}

}

