/*
 * Copyright (C) 2007 Apple Inc.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2008 Collabora Ltd.
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"

#include "BitmapImage.h"
#include "RenderThemeWiiU.h"
#include "Path.h"
#include "PaintInfo.h"
#include "Color.h"
#include "FloatPoint.h"
#include "TransformationMatrix.h"
#include "GraphicsContext.h"
#include "NotImplemented.h"
#include "RenderObject.h"
#include "SharedBuffer.h"
#include "assets/RenderTheme_AssetData.h"

#if ENABLE(PROGRESS_TAG)
#include "RenderProgress.h"
#endif

#if ENABLE(VIDEO)
#include "MediaControlElements.h"
#include "HTMLMediaElement.h"
#include "TimeRanges.h"
#endif

#ifndef NDEBUG 
#include <iostream>
#endif

namespace WebCore {

#if !PLATFORM(WIIU)
char RenderThemeWiiU_assetsPath[1024] = {0};

void RenderThemeWiiU::setAssetsPath(const char *set)
{
	if (set)
		strcpy(RenderThemeWiiU_assetsPath, set);
	else
		RenderThemeWiiU_assetsPath[0] = 0;
}

std::string RenderThemeWiiU::prependAssetsPath(const char *path)
{
	std::string fullpath = RenderThemeWiiU_assetsPath;
	if (path) fullpath.append(path);
	return fullpath;
}
#endif

PassRefPtr<Image> RenderThemeWiiU::loadRenderThemeImage(const char *path)
{ 
    RefPtr<SharedBuffer> buffer = SharedBuffer::createWithContentsOfFile(path);    
    RefPtr<BitmapImage> img;
    if (buffer) {
        img = BitmapImage::create();
        img->setData(buffer.release(), true);
        return img.release();
    }
 
    return 0;
}

PassRefPtr<Image> RenderThemeWiiU::loadRenderThemeImageFromMemory(const unsigned char *image, int size )
{
    RefPtr<SharedBuffer> buffer = SharedBuffer::create(image, size);
    RefPtr<BitmapImage> img;
    if (buffer) {
        img = BitmapImage::create();
        img->setData(buffer.release(), true);
        return img.release();
    }

    return 0;
}

void RenderThemeWiiU::paintImagePreserveHorizontalEdges(GraphicsContext* const ctx, Image *image, FloatRect &frect)
{
    //draw the middle 
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x() + 2, frect.y(), frect.width() - 2, frect.height()), //dest
        FloatRect(2, 0, image->width() - 4, image->height())); //src

    //redraw the sides to firm things up (left side)
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x(), frect.y(), 2, frect.height()), //dest
        FloatRect(0, 0, 2, image->height())); //src

    // and right side
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x() + frect.width() - 2, frect.y(), 2, frect.height()), //dest
        FloatRect(image->width() - 2, 0, 2, image->height())); //src
}

void RenderThemeWiiU::paintImagePreserveVerticalEdges(GraphicsContext* const ctx, Image *image, FloatRect &frect)
{
    //draw the middle 
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x() + 2, frect.y(), frect.width() - 2, frect.height()), //dest
        FloatRect(2, 0, image->width() - 4, image->height())); //src

    //redraw the sides to firm things up (left side)
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x(), frect.y(), 2, frect.height()), //dest
        FloatRect(0, 0, 2, image->height())); //src

    // and right side
    ctx->drawImage(image, ColorSpaceDeviceRGB,
        FloatRect(frect.x() + frect.width() - 2, frect.y(), 2, frect.height()), //dest
        FloatRect(image->width() - 2, 0, 2, image->height())); //src
}

static const int trackWidth = 5;

PassRefPtr<RenderTheme> RenderTheme::themeForPage(Page* page)
{
    return adoptRef(new RenderThemeWiiU());
}


RenderThemeWiiU::RenderThemeWiiU()
:defaultBorderColor(Color(145, 145, 145)), /* match border of checkbox in render theme */
defaultButtonBGColor(Color(220, 220, 220)),
defaultButtonCorner(IntSize(5,5))
{
}


static float min(float x, float y){
	return (x<y ? x : y);
}

enum {
    topPadding,
    rightPadding,
    bottomPadding,
    leftPadding
};

enum NSControlSize{
	NSRegularControlSize = 0,
	NSSmallControlSize = 1,    
    NSMiniControlSize = 2
};


static NSControlSize controlSizeForFont(RenderStyle* style)
{
    int fontSize = style->fontSize();
    
    if (fontSize >= 16){
        return NSRegularControlSize;
    }
    else if (fontSize >= 11){
        return NSSmallControlSize;
    }
    else{
    	return NSMiniControlSize;
    }
}

/* CRANK: DLM
    These sizes were chosen to line up with the default wiiu theme images
*/
static const IntSize* buttonSizes()
{
    static const IntSize sizes[3] = { IntSize(0, 32), IntSize(0, 16), IntSize(0, 12) };
    return sizes;
}


static const IntSize* radioSizes()
{
    static const IntSize sizes[3] = { IntSize(16, 16), IntSize(12, 13), IntSize(10, 10) };
    return sizes;
}


static const IntSize* checkboxSizes()
{
    static const IntSize sizes[3] = { IntSize(16, 16), IntSize(12, 12), IntSize(10, 10) };
    return sizes;
}


static IntSize sizeForFont(RenderStyle* style, const IntSize* sizes)
{
    return sizes[controlSizeForFont(style)];
}

static void setSizeFromFont(RenderStyle* style, const IntSize* sizes)
{
    // FIXME: Check is flawed, since it doesn't take min-width/max-width into account.
    IntSize size = sizeForFont(style, sizes);
    
    if (style->width().isIntrinsicOrAuto() && size.width() > 0)
        style->setWidth(Length(size.width(), Fixed));
    
    if (style->height().isAuto() && size.height() > 0)
        style->setHeight(Length(size.height(), Fixed));
}


bool RenderThemeWiiU::supportsFocusRing(const RenderStyle* style) const
{
    return false;
}


int RenderThemeWiiU::baselinePosition(const RenderObject* o) const
{
    return RenderTheme::baselinePosition(o);
}


static void setButtonPaddingFromControlSize(RenderStyle* style, NSControlSize size)
{
    // Just use 8px.  AppKit wants to use 11px for mini buttons, but that padding is just too large
    // for real-world Web sites (creating a huge necessary minimum width for buttons whose space is
    // by definition constrained, since we select mini only for small cramped environments.
    // This also guarantees the HTML4 <button> will match our rendering by default, since we're using a consistent
    // padding.
    const int padding = 8;
    style->setPaddingLeft(Length(padding, Fixed));
    style->setPaddingRight(Length(padding, Fixed));
    style->setPaddingTop(Length(0, Fixed));
    style->setPaddingBottom(Length(0, Fixed));
}


void RenderThemeWiiU::setCheckboxSize(RenderStyle* style) const
{
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, checkboxSizes());
}

bool RenderThemeWiiU::paintCheckbox(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{

	GraphicsContext* const ctx = i.context;

    FloatRect frect(rect);

    if (RenderTheme::isEnabled(o)) {
        if(RenderTheme::isChecked(o)) {
            if (!imgCheckBoxChecked) {
                imgCheckBoxChecked = loadRenderThemeImageFromMemory(checkBoxCheckedBmp, sizeof(checkBoxCheckedBmp));
            }

            ctx->drawImage(imgCheckBoxChecked.get(), ColorSpaceDeviceRGB, frect); 
        } else {
            if (!imgCheckBoxUnchecked) {
                imgCheckBoxUnchecked = loadRenderThemeImageFromMemory(checkBoxUncheckedBmp, sizeof(checkBoxUncheckedBmp));
            }

            ctx->drawImage(imgCheckBoxUnchecked.get(), ColorSpaceDeviceRGB, frect); 
        }
    } else { 
        if(RenderTheme::isChecked(o)) {
            if (!imgCheckBoxCheckedDisabled) {
                imgCheckBoxCheckedDisabled = loadRenderThemeImageFromMemory(checkBoxCheckedDisabledBmp, sizeof(checkBoxCheckedDisabledBmp));
            }

            ctx->drawImage(imgCheckBoxCheckedDisabled.get(), ColorSpaceDeviceRGB, frect); 
        } else {
            if (!imgCheckBoxUncheckedDisabled) {
                imgCheckBoxUncheckedDisabled = loadRenderThemeImageFromMemory(checkBoxUncheckedDisabledBmp, sizeof(checkBoxUncheckedDisabledBmp));
            }

            ctx->drawImage(imgCheckBoxUncheckedDisabled.get(), ColorSpaceDeviceRGB, frect); 
        }

    }

	return false;
}

void RenderThemeWiiU::setRadioSize(RenderStyle* style) const
{
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, radioSizes());
}

bool RenderThemeWiiU::paintRadio(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{ 
	GraphicsContext* const ctx = i.context;

    FloatRect frect(rect);

    if (RenderTheme::isEnabled(o)) {
        if(RenderTheme::isChecked(o)) {
            if (!imgRadioButtonChecked) {
                imgRadioButtonChecked = loadRenderThemeImageFromMemory(radioButtonCheckedBmp, sizeof(radioButtonCheckedBmp));
            }

            ctx->drawImage(imgRadioButtonChecked.get(), ColorSpaceDeviceRGB, frect); 
        } else {
            if (!imgRadioButtonUnchecked) {
                imgRadioButtonUnchecked = loadRenderThemeImageFromMemory(radioButtonUncheckedBmp, sizeof(radioButtonUncheckedBmp));
            }

            ctx->drawImage(imgRadioButtonUnchecked.get(), ColorSpaceDeviceRGB, frect); 
        }
    } else { 
        if(RenderTheme::isChecked(o)) {
            if (!imgRadioButtonCheckedDisabled) {
                imgRadioButtonCheckedDisabled = loadRenderThemeImageFromMemory(radioButtonCheckedDisabledBmp, sizeof(radioButtonCheckedDisabledBmp));
            }

            ctx->drawImage(imgRadioButtonCheckedDisabled.get(), ColorSpaceDeviceRGB, frect); 
        } else {
            if (!imgRadioButtonUncheckedDisabled) {
                imgRadioButtonUncheckedDisabled = loadRenderThemeImageFromMemory(radioButtonUncheckedDisabledBmp, sizeof(radioButtonUncheckedDisabledBmp));
            }

            ctx->drawImage(imgRadioButtonUncheckedDisabled.get(), ColorSpaceDeviceRGB, frect); 
        }

    }

	return false;
}


void RenderThemeWiiU::setButtonSize(RenderStyle* style) const
{
    // If the width and height are both specified, then we have nothing to do.
    if (!style->width().isIntrinsicOrAuto() && !style->height().isAuto())
        return;

    // Use the font size to determine the intrinsic width of the control.
    setSizeFromFont(style, buttonSizes());
}

void RenderThemeWiiU::adjustButtonStyle(CSSStyleSelector* selector, RenderStyle* style, WebCore::Element* e) const
{
    // There are three appearance constants for buttons.
    // (1) Push-button is the constant for the default Aqua system button.  Push buttons will not scale vertically and will not allow
    // custom fonts or colors.  <input>s use this constant.  This button will allow custom colors and font weights/variants but won't
    // scale vertically.
    // (2) square-button is the constant for the square button.  This button will allow custom fonts and colors and will scale vertically.
    // (3) Button is the constant that means "pick the best button as appropriate."  <button>s use this constant.  This button will
    // also scale vertically and allow custom fonts and colors.  It will attempt to use Aqua if possible and will make this determination
    // solely on the rectangle of the control.

    // Determine our control size based off our font.
    NSControlSize controlSize = controlSizeForFont(style);

    if(style->appearance() == PushButtonPart) {
        // Ditch the border.
        style->resetBorder();

        // Height is locked to auto.
        style->setHeight(Length(Auto));

        // White-space is locked to pre
        style->setWhiteSpace(PRE);

        // Set the button's vertical size.
        setButtonSize(style);

        // Add in the padding that we'd like to use.
        setButtonPaddingFromControlSize(style, controlSize);
    } else {
        // Set a min-height so that we can't get smaller than the mini button.
        style->setMinHeight(Length(15, Fixed));

        // Reset the top and bottom borders.
        style->resetBorderTop();
        style->resetBorderBottom();
    }
}


bool RenderThemeWiiU::paintButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* const ctx = i.context;

    FloatRect frect(rect);

    if (RenderTheme::isPressed(o)) {
        if (!imgButtonPressed) {
            imgButtonPressed = loadRenderThemeImageFromMemory(buttonPressedBmp, sizeof(buttonPressedBmp));
        } 

        paintImagePreserveVerticalEdges(ctx, imgButtonPressed.get(), frect);

    } else if (RenderTheme::isEnabled(o)) {
        if (!imgButtonNormal) {
            imgButtonNormal = loadRenderThemeImageFromMemory(buttonNormalBmp, sizeof(buttonNormalBmp));
        }

        paintImagePreserveVerticalEdges(ctx, imgButtonNormal.get(), frect);

    } else {

        if (!imgButtonDisabled) {
            imgButtonDisabled = loadRenderThemeImageFromMemory(buttonDisabledBmp, sizeof(buttonDisabledBmp));
        }

        paintImagePreserveVerticalEdges(ctx, imgButtonDisabled.get(), frect);

    }

    return false;
}

static const float baseFontSize = 11.0f;
static const float baseArrowHeight = 16.0f;
static const float baseArrowWidth = 16.0f;
static const int arrowPaddingLeft = 5;
static const int arrowPaddingRight = 5;
static const int paddingBeforeSeparator = 4;
static const int baseBorderRadius = 5;
static const int styledPopupPaddingLeft = 8;
static const int styledPopupPaddingTop = 1;
static const int styledPopupPaddingBottom = 2;

static const int* popupButtonPadding(NSControlSize size)
{
    static const int padding[3][4] =
    {
        { 2, 26, 3, 8 },
        { 2, 23, 3, 8 },
        { 2, 22, 3, 10 }
    };
    return padding[size];
}


int RenderThemeWiiU::popupInternalPaddingLeft(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[leftPadding];
    
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingLeft;
    return 0;
}

int RenderThemeWiiU::popupInternalPaddingRight(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[rightPadding];
    
    if (style->appearance() == MenulistButtonPart) {
        float fontScale = style->fontSize() / baseFontSize;
        float arrowWidth = baseArrowWidth * fontScale;
        return static_cast<int>(ceilf(arrowWidth + arrowPaddingLeft + arrowPaddingRight + paddingBeforeSeparator));
    }
    return 0;
}

int RenderThemeWiiU::popupInternalPaddingTop(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[topPadding];
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingTop;
    return 0;
}

int RenderThemeWiiU::popupInternalPaddingBottom(RenderStyle* style) const
{
    if (style->appearance() == MenulistPart)
        return popupButtonPadding(controlSizeForFont(style))[bottomPadding];
    if (style->appearance() == MenulistButtonPart)
        return styledPopupPaddingBottom;
    return 0;
}


void RenderThemeWiiU::adjustMenuListStyle(CSSStyleSelector* selector, RenderStyle* style, WebCore::Element* e) const
{
    style->resetBorder();
    style->resetPadding();
    style->setHeight(Length(Auto));
    style->setWhiteSpace(PRE);
}


bool RenderThemeWiiU::paintMenuListButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* const ctx = i.context;

    FloatRect bounds(rect.x() + o->style()->borderLeftWidth(),
                    rect.y() + o->style()->borderTopWidth(),
                    rect.width() - o->style()->borderLeftWidth() - o->style()->borderRightWidth(),
                    rect.height() - o->style()->borderTopWidth() - o->style()->borderBottomWidth());

    float fontScale = min(o->style()->fontSize() / baseFontSize, bounds.height() / baseArrowHeight);
    float centerY = bounds.y() + bounds.height() / 2.0f;
    float arrowHeight = baseArrowHeight * fontScale;
    float arrowWidth = baseArrowWidth * fontScale;
    float leftEdge = bounds.x() + bounds.width() - arrowPaddingRight - arrowWidth;

    FloatRect frect(leftEdge + arrowPaddingLeft, centerY - 0.5*arrowWidth, arrowWidth, arrowHeight);

    if (RenderTheme::isPressed(o)) {
        if (!imgMenuListButtonPressed) {
			imgMenuListButtonPressed = loadRenderThemeImageFromMemory(menuListButtonFocusedBmp, sizeof(menuListButtonFocusedBmp));
        } 

        ctx->drawImage(imgMenuListButtonPressed.get(), ColorSpaceDeviceRGB, frect); 
    } else if (RenderTheme::isEnabled(o)) {
        if (!imgMenuListButtonNormal) {
            imgMenuListButtonNormal = loadRenderThemeImageFromMemory(menuListButtonNormalBmp, sizeof(menuListButtonNormalBmp));
        }

        ctx->drawImage(imgMenuListButtonNormal.get(), ColorSpaceDeviceRGB, frect); 
    } else {
        if (!imgMenuListButtonDisabled) {
            imgMenuListButtonDisabled = loadRenderThemeImageFromMemory(menuListButtonDisabledBmp, sizeof(menuListButtonDisabledBmp));
        }

        ctx->drawImage(imgMenuListButtonDisabled.get(), ColorSpaceDeviceRGB, frect); 
    }

    return false;

    /*

	GraphicsContext* const ctx = paintInfo.context;
	
    IntRect bounds = IntRect(r.x() + o->style()->borderLeftWidth(),
                             r.y() + o->style()->borderTopWidth(),
                             r.width() - o->style()->borderLeftWidth() - o->style()->borderRightWidth(),
                             r.height() - o->style()->borderTopWidth() - o->style()->borderBottomWidth());
        
    // Since we actually know the size of the control here, we restrict the font scale to make sure the arrow will fit vertically in the bounds
    float fontScale = min(o->style()->fontSize() / baseFontSize, bounds.height() / baseArrowHeight);
    float centerY = bounds.y() + bounds.height() / 2.0f;
    float arrowHeight = baseArrowHeight * fontScale;
    float arrowWidth = baseArrowWidth * fontScale;
    float leftEdge = bounds.x() +bounds.width() - arrowPaddingRight - arrowWidth;

    if (bounds.width() < arrowWidth + arrowPaddingLeft)
        return false;

    ctx->setFillColor(Color(0xff,0xff,0xff), ColorSpaceDeviceRGB);
    ctx->setStrokeColor(NoStroke, ColorSpaceDeviceRGB);

    FloatPoint arrow[3];
    arrow[0] = FloatPoint(leftEdge, centerY - arrowHeight / 2.0f);
    arrow[1] = FloatPoint(leftEdge + arrowWidth, centerY - arrowHeight / 2.0f);
    arrow[2] = FloatPoint(leftEdge + arrowWidth / 2.0f, centerY + arrowHeight / 2.0f);

    // Draw the arrow
    ctx->drawConvexPolygon(3, arrow, true);

    Color leftSeparatorColor(0, 0, 0, 40);
    Color rightSeparatorColor(255, 255, 255, 40);
    
    // FIXME: Should the separator thickness and space be scaled up by fontScale?
    int separatorSpace = 2;
    int leftEdgeOfSeparator = static_cast<int>(leftEdge - arrowPaddingLeft); // FIXME: Round?

    // Draw the separator to the left of the arrows
    ctx->setStrokeThickness(1.0f);
    ctx->setStrokeStyle(SolidStroke);
    ctx->setStrokeColor(leftSeparatorColor, ColorSpaceDeviceRGB);
    ctx->drawLine(IntPoint(leftEdgeOfSeparator, bounds.y()),
                                IntPoint(leftEdgeOfSeparator, bounds.y()+bounds.height()));

    ctx->setStrokeColor(rightSeparatorColor, ColorSpaceDeviceRGB);
    ctx->drawLine(IntPoint(leftEdgeOfSeparator + separatorSpace, bounds.y()),
                                IntPoint(leftEdgeOfSeparator + separatorSpace, bounds.y()+bounds.height()));

    return false;
*/
}

bool RenderThemeWiiU::paintMenuList(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{	
	GraphicsContext* const ctx = i.context;
	
	ctx->save();	
	ctx->clip(rect);
	FloatRect frect = rect;
    ctx->setFillColor(defaultButtonBGColor, ColorSpaceDeviceRGB);
	ctx->fillRect(frect);//, Color(0xc0, 0xc0, 0xc0, 0xff));
		
	//draw border
    ctx->setStrokeStyle(SolidStroke);
	ctx->setStrokeColor(defaultBorderColor, ColorSpaceDeviceRGB);	    
	ctx->strokeRect(rect, 1);

	if(o && o->isMenuList()){
		paintMenuListButton(o, i, rect);
	}
	
	ctx->restore();
	
	return false;
}

#if ENABLE(VIDEO)
// Attempt to retrieve a HTMLMediaElement from a Node. Returns 0 if one cannot be found.
static HTMLMediaElement* mediaElementParent(Node* node)
{
    if (!node)
        return 0;
    Node* mediaNode = node->shadowAncestorNode();
    if (!mediaNode || !mediaNode->isElementNode() || !static_cast<Element*>(mediaNode)->isMediaElement())
        return 0;

    return static_cast<HTMLMediaElement*>(mediaNode);
}
#endif

bool RenderThemeWiiU::paintSliderTrack(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	  	 IntRect bounds = rect;
	    // modify the placement of the slider track so that it properly lines up with the bounding box
	    // that the thumb will move over

	    if (o->style()->appearance() ==  SliderHorizontalPart) {
	        bounds.setHeight(trackWidth);
	        bounds.setY(rect.y() + rect.height() / 2 - trackWidth / 2);
	    } else if (o->style()->appearance() == SliderVerticalPart) {
	        bounds.setWidth(trackWidth);
	        bounds.setX((rect.x() + rect.width() / 2 - trackWidth / 2) );
	        //bounds.setY((rect.y() - vertical_adjust ));
	    }

	    IntPoint left = IntPoint(bounds.x() + trackWidth, (bounds.y() + bounds.maxY()) / 2);
	    i.context->save();
	    i.context->setStrokeColor(defaultButtonBGColor, ColorSpaceDeviceRGB);
	    i.context->setFillColor(defaultButtonBGColor, ColorSpaceDeviceRGB);
	    i.context->fillRect(bounds);
	#if ENABLE(VIDEO)
	    HTMLMediaElement *mediaElement = mediaElementParent(o->node());
	    if (mediaElement) {
	        i.context->setStrokeColor(Color(0, 0xff, 0), ColorSpaceDeviceRGB);
	        IntPoint right = IntPoint(left.x() + mediaElement->percentLoaded() * trackWidth, left.y());
	        i.context->drawLine(left, right);
	        left = right;
	    }
	#endif
	    i.context->setStrokeColor(defaultBorderColor, ColorSpaceDeviceRGB);
	    i.context->setFillColor(defaultBorderColor, ColorSpaceDeviceRGB);
	    if (o->style()->appearance() ==  SliderHorizontalPart) {
	    	bounds.setHeight(1);
	    	bounds.setY(bounds.y() + trackWidth / 2);
	    	i.context->fillRect(bounds);
	        //i.context->drawLine(left, IntPoint(rect.maxX() - 2, left.y()));
	    } else {
	        // draw the internal vertical line inside the slider-track to provide some definition
	        //int width_padding = 5;
	        //int internal_padding = 3;

	        bounds.setWidth(1);
	        //bounds.setHeight(rect.width()- width_padding);
	        bounds.setX(bounds.x() + trackWidth / 2);
	        //bounds.setY(bounds.y()+ internal_padding);
	        i.context->fillRect(bounds);
	    }
	    i.context->restore();
	    return false;

}


bool RenderThemeWiiU::paintSliderThumb(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{

    bool rc = RenderTheme::paintSliderThumb(o, i, rect);
    i.context->save();
    i.context->setStrokeColor(defaultBorderColor, ColorSpaceDeviceRGB);
    i.context->setFillColor(defaultButtonBGColor, ColorSpaceDeviceRGB);
    i.context->fillRect(rect);
    i.context->strokeRect(rect, 1);
#if ENABLE(VIDEO)
    HTMLMediaElement* mediaElement = mediaElementParent(o->node());
    if (mediaElement) {
        float pt = (mediaElement->currentTime() - mediaElement->startTime()) / mediaElement->duration();
        FloatRect intRect = rect;
        intRect.setX(intRect.x() + intRect.width() * pt - 2);
        intRect.setWidth(5);
        i.context->fillRect(intRect);
    }
#endif
    i.context->restore();
    return false;
}


void RenderThemeWiiU::adjustTextFieldStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    style->resetBorder();
    style->resetPadding();
    style->setHeight(Length(Auto));
}

bool RenderThemeWiiU::paintTextField(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	GraphicsContext* const ctx = i.context;
	
	ctx->save();
	
	ctx->clip(rect);
	
    //draw border
    ctx->setStrokeColor(defaultBorderColor, ColorSpaceDeviceRGB);
    
    ctx->strokeRect(rect, 1);
    
	ctx->restore();
	
	return false;
}


void RenderThemeWiiU::adjustTextAreaStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    adjustTextFieldStyle(selector, style, e);
}

bool RenderThemeWiiU::paintTextArea(RenderObject* o, const PaintInfo& i, const IntRect& r)
{
    return paintTextField(o, i, r);
}

void RenderThemeWiiU::adjustSearchFieldResultsButtonStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    adjustSearchFieldCancelButtonStyle(selector, style, e);
}

bool RenderThemeWiiU::paintSearchFieldResultsButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{

	notImplemented();
	return false;
}

void RenderThemeWiiU::adjustSearchFieldResultsDecorationStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    style->resetBorder();
    style->resetPadding();

    // FIXME: This should not be hard-coded.
    IntSize size = IntSize(14, 14);
    style->setWidth(Length(size.width(), Fixed));
    style->setHeight(Length(size.height(), Fixed));
}

bool RenderThemeWiiU::paintSearchFieldResultsDecoration(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	notImplemented();
	return false;
}

void RenderThemeWiiU::adjustSearchFieldCancelButtonStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    style->resetBorder();
    style->resetPadding();

    // FIXME: This should not be hard-coded.
    IntSize size = IntSize(14, 14);
    style->setWidth(Length(size.width(), Fixed));
    style->setHeight(Length(size.height(), Fixed));
}

bool RenderThemeWiiU::paintSearchFieldCancelButton(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
	notImplemented();
	return false;
}

void RenderThemeWiiU::adjustSearchFieldStyle(CSSStyleSelector* selector, RenderStyle* style, Element* e) const
{
    adjustTextFieldStyle(selector, style, e);
}

const int sliderThumbWidth = 7;
const int sliderThumbHeight = 15;

void RenderThemeWiiU::adjustSliderThumbSize(RenderStyle* style) const
{
    if (style->appearance() == SliderThumbVerticalPart) {
        style->setWidth(Length(sliderThumbHeight, Fixed));
        style->setHeight(Length(sliderThumbWidth, Fixed));
    } else if (style->appearance() == SliderThumbHorizontalPart) {
        style->setWidth(Length(sliderThumbWidth, Fixed));
        style->setHeight(Length(sliderThumbHeight, Fixed));
    }
}


bool RenderThemeWiiU::paintSearchField(RenderObject* o, const PaintInfo& i, const IntRect& rect)
{
    return paintTextField(o, i, rect);
}

void RenderThemeWiiU::systemFont(int, FontDescription&) const
{
    // If you remove this notImplemented(), replace it with an comment that explains why.
    notImplemented();
}


String RenderThemeWiiU::extraDefaultStyleSheet()
{
	String extraStyles;
	
	return extraStyles;
}

#if ENABLE(PROGRESS_TAG)
void RenderThemeWiiU::adjustProgressBarStyle(CSSStyleSelector*, RenderStyle* style, Element*) const
{
    style->setBoxShadow(nullptr);
}

// These values have been copied from RenderThemeChromiumSkia.cpp
static const int progressActivityBlocks = 5;
static const int progressAnimationFrames = 10;
static const double progressAnimationInterval = 0.125;
double RenderThemeWiiU::animationRepeatIntervalForProgressBar(RenderProgress*) const
{
    return progressAnimationInterval;
}

double RenderThemeWiiU::animationDurationForProgressBar(RenderProgress*) const
{
    return progressAnimationInterval * progressAnimationFrames * 2; // "2" for back and forth;
}

IntRect RenderThemeWiiU::calculateProgressRect(RenderObject* renderObject, const IntRect& fullBarRect)
{
    IntRect progressRect(fullBarRect);
    RenderProgress* renderProgress = toRenderProgress(renderObject);
    if (renderProgress->isDeterminate()) {
        int progressWidth = progressRect.width() * renderProgress->position();
        if (renderObject->style()->direction() == RTL)
            progressRect.setX(progressRect.x() + progressRect.width() - progressWidth);
        progressRect.setWidth(progressWidth);
        return progressRect;
    }

    double animationProgress = renderProgress->animationProgress();

    // Never let the progress rect shrink smaller than 2 pixels.
    int newWidth = max(2, progressRect.width() / progressActivityBlocks);
    int movableWidth = progressRect.width() - newWidth;
    progressRect.setWidth(newWidth);

    // We want the first 0.5 units of the animation progress to represent the
    // forward motion and the second 0.5 units to represent the backward motion,
    // thus we multiply by two here to get the full sweep of the progress bar with
    // each direction.
    if (animationProgress < 0.5)
        progressRect.setX(progressRect.x() + (animationProgress * 2 * movableWidth));
    else
        progressRect.setX(progressRect.x() + ((1.0 - animationProgress) * 2 * movableWidth));
    return progressRect;
}

bool RenderThemeWiiU::paintProgressBar(RenderObject* o, const PaintInfo& i, const IntRect& rect) {
	IntRect progressRect = calculateProgressRect(o, rect);
	GraphicsContext* const ctx = i.context;


	ctx->save();

	ctx->clip(rect);

	FloatRect frect = rect;

    //fill background
    ctx->setFillColor(defaultButtonBGColor, ColorSpaceDeviceRGB);
    ctx->fillRect(frect);//, defaultButtonBGColor);

    Color progressFill(0, 0, 255, 100);

    FloatRect prect = progressRect;
    ctx->setFillColor(progressFill, ColorSpaceDeviceRGB);
    ctx->fillRect(prect);

    //draw border
    ctx->setStrokeColor(defaultBorderColor, ColorSpaceDeviceRGB);
    ctx->strokeRect(rect, 1);

    ctx->restore();
}
#endif

#if ENABLE(VIDEO)

/* CRANK: DLM
   What we get asked to draw here is very dependent upon what is going on the in MediaPlayer
   instance.  For example, if there is no valid duration, we will not get asked to draw the MediaSlider or thumb.
*/

bool RenderThemeWiiU::paintMediaFullscreenButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
    return true;
}

bool RenderThemeWiiU::paintMediaPlayButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;
    FloatRect frect(rect);
    
    HTMLMediaElement *mediaElement = mediaElementParent(o->node());
    if (mediaElement) {
        if (mediaElement->paused()) {
            if (!imgPlayButton) {
              	 imgPlayButton = loadRenderThemeImageFromMemory(playButtonPng, sizeof(playButtonPng));
            }

            ctx->drawImage(imgPlayButton.get(), ColorSpaceDeviceRGB, frect); 
        } else {
            if (!imgPauseButton) {
                imgPauseButton = loadRenderThemeImageFromMemory(pauseButtonPng, sizeof(pauseButtonPng));
            } 

            ctx->drawImage(imgPauseButton.get(), ColorSpaceDeviceRGB, frect); 
        }
    }

    return false;
}

bool RenderThemeWiiU::paintMediaMuteButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;

    if (!imgMuteButton) {
        imgMuteButton = loadRenderThemeImageFromMemory(muteButtonPng, sizeof(muteButtonPng));
    }

    FloatRect frect(rect);
    ctx->drawImage(imgMuteButton.get(), ColorSpaceDeviceRGB, frect); 

    return false;
}

bool RenderThemeWiiU::paintMediaSeekBackButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;

    return true;
}

bool RenderThemeWiiU::paintMediaSeekForwardButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

/*
    CRANK: DLM This code is modified from the RenderThemeEfl.cpp version.
*/

bool RenderThemeWiiU::paintMediaSliderTrack(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const context = i.context; 

    const int mediaSliderHeight = 16;

    IntSize sliderRect(3, 3);
    //ctx->fillRoundedRect(rect, thumbRect, thumbRect, thumbRect, thumbRect, Color(186, 0, 0), ColorSpaceDeviceRGB);

    context->fillRoundedRect(IntRect(rect.x(), rect.y() + (rect.height() - mediaSliderHeight) / 2,
                                        rect.width(), mediaSliderHeight),
                             sliderRect, sliderRect, sliderRect, sliderRect,
                             Color(186, 186, 186), ColorSpaceDeviceRGB);

    RenderStyle* style = o->style();
    HTMLMediaElement* mediaElement = toParentMediaElement(o);

    if (!mediaElement)
        return false;

    // Draw the buffered ranges. This code is highly inspired from
    // Chrome for the gradient code.
    float mediaDuration = mediaElement->duration();
    RefPtr<TimeRanges> timeRanges = mediaElement->buffered();
    IntRect trackRect = rect;
    int totalWidth = trackRect.width();

    trackRect.inflate(-style->borderLeftWidth());
    context->save();
    context->setStrokeStyle(NoStroke);

    for (unsigned index = 0; index < timeRanges->length(); ++index) {
        ExceptionCode ignoredException;
        float start = timeRanges->start(index, ignoredException);
        float end = timeRanges->end(index, ignoredException);
        int width = ((end - start) * totalWidth) / mediaDuration;
        IntRect rangeRect;
        if (!index) {
            rangeRect = trackRect;
            rangeRect.setWidth(width);
        } else {
            rangeRect.setLocation(IntPoint(trackRect.x() + start / mediaDuration* totalWidth, trackRect.y()));
            rangeRect.setSize(IntSize(width, trackRect.height()));
        }

        // Don't bother drawing empty range.
        if (rangeRect.isEmpty())
            continue;

        IntPoint sliderTopLeft = rangeRect.location();
        IntPoint sliderTopRight = sliderTopLeft;
        sliderTopRight.move(0, rangeRect.height());

        context->fillRect(FloatRect(rect), Color(145, 145, 145), ColorSpaceDeviceRGB);
    }
    context->restore();
    return true;
}

bool RenderThemeWiiU::paintMediaSliderThumb(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;

    IntSize thumbRect(3, 3);
    ctx->fillRoundedRect(rect, thumbRect, thumbRect, thumbRect, thumbRect, Color(255, 255, 255), ColorSpaceDeviceRGB);
    return false;
}

/*
    CRANK: DLM if we return true from the paintMedia* functions, we will get a default
           volume slider implementation.  
*/
bool RenderThemeWiiU::paintMediaVolumeSliderContainer(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;
    return true;
}

bool RenderThemeWiiU::paintMediaVolumeSliderTrack(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;
    return true;
}

bool RenderThemeWiiU::paintMediaVolumeSliderThumb(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;
    return true;
}

bool RenderThemeWiiU::paintMediaRewindButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

bool RenderThemeWiiU::paintMediaReturnToRealtimeButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

bool RenderThemeWiiU::paintMediaToggleClosedCaptionsButton(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

bool RenderThemeWiiU::paintMediaControlsBackground(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

bool RenderThemeWiiU::paintMediaCurrentTime(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context; 
    return true;
}

bool RenderThemeWiiU::paintMediaTimeRemaining(RenderObject *o, const PaintInfo &i, const IntRect &rect)
{
	GraphicsContext *const ctx = i.context;
    return true;
}
#endif

}
