/*
 * Copyright (C) 2008 Kenneth Rohde Christiansen
 *           (C) 2008 Afonso Rabelo Costa Jr.
 *           (C) 2009-2010 ProFUSION embedded systems
 *           (C) 2009-2010 Samsung Electronics
 *           (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "SharedTimer.h"
#include "WiiUEventManager.h"

#include <wtf/Assertions.h>
#include <wtf/CurrentTime.h>
#include <wtf/MainThread.h>

#if PLATFORM(WIIU)
	#include <PlatformOS.h>
 	#include <PlatformDefines.h>
	using namespace LibPlatform;
#elif OS(WINDOWS)
	#include <windows.h>
#endif

namespace WebCore {

#if PLATFORM(WIIU)
static bool _bInitAlarm = false;
static OS::Alarm alarm;
#elif OS(WINDOWS)
HANDLE timerQueue = 0, timer = 0;
#endif
static void (*sharedTimerFiredFunction)();

/*
 * This sets the function that should be called when a timer fires.  This
 * generally does not change and is usually set to a function that will process
 * an internal webkit timer queue.
 */
void setSharedTimerFiredFunction(void (*func)())
{
    sharedTimerFiredFunction = func;
}

/*
 * WiiU callback that gets called when our alarm fires.
 */
#if PLATFORM(WIIU)
static void timeout_cb(void* alarm, void* context)
#elif OS(WINDOWS)
static void NTAPI timeout_cb(PVOID data, BOOLEAN)
#endif
{
    if (sharedTimerFiredFunction)
	{
		WiiUBaseEventData *pEvent = new WiiUTimerFired((void*)sharedTimerFiredFunction);
		if (pEvent)
			WiiUEventManager::instance()->putEvent(pEvent);
	}
}

/*
 * Cancels the wiiu alarm that was set previously.
 */
void stopSharedTimer()
{
#if PLATFORM(WIIU)
	if (_bInitAlarm)
		OS::CancelAlarm(&alarm);
#elif OS(WINDOWS)
	if(timer)
		DeleteTimerQueueTimer(timerQueue, timer, 0);
	timer = 0;
#endif
}

/*
 * Sets the time until the next timer should be fired.  If the time is zero or less
 * then the timer fires immediately.
 */
void setSharedTimerFireInterval(double interval)
{
#if PLATFORM(WIIU)
	// Create the OS alarm if we didn't do so, yet. If we have one: make sure it's off.
	if (_bInitAlarm)
	{
		OS::CreateAlarm(&alarm);
		_bInitAlarm = true;
	}
	else
		stopSharedTimer();
	
	// Do we have any valid interval?
	if (interval > 0.0)
	{
		// Yes, we actually have some period to wait. See if we can express it in an OSTick value (32bit!) - we clamp otherwise, the next best thing we can do...
		interval *= 1000.0;

		const double kMaxInterval = (double(0xffffffff) * 1000.0) / SYSTEM_TIMER_CLOCK;
		double timeInMs;
		if (interval > kMaxInterval)
			timeInMs = kMaxInterval;
		else
			timeInMs = unsigned(interval);
			
		OS::Tick ticks = OS::MillisecondsToTicks(timeInMs);
			
		// After all this conversion is the time still valid for a wait?
		// (precision might have yielded us a "zero wait" after all)
		if (ticks)
		{
			// Start the timer & exit...
			OS::SetAlarm(&alarm, ticks, timeout_cb);
			return;
		}
	}
	
	// Ok, we had no real time to wait at all. Just add the event directly...
	WiiUBaseEventData *pEvent = new WiiUTimerFired((void*)sharedTimerFiredFunction);
	if (pEvent)
		WiiUEventManager::instance()->putEvent(pEvent);
#elif OS(WINDOWS)
	static int init_alarm = 0;
	long long timeInMs = interval < 0.0 ? 0 : static_cast<long long>(interval * 1000);

	// If we have not created a system alarm, we should do so at this point.
	// Otherwise, we are going to change the time of the next firing, so
	// cancel the previous alarm.
	if (init_alarm == 0) {
		timerQueue = CreateTimerQueue();
		init_alarm = 1;
	} else {
		stopSharedTimer();
	}

	// If the time is 0 or negative, put the event in the queu now.  Otherwise
	// calculate the ticks for the wiiu alarm.
	if (timeInMs <=0) {
		WiiUBaseEventData *pEvent = new WiiUTimerFired((void*)sharedTimerFiredFunction);
		if (pEvent)
			WiiUEventManager::instance()->putEvent(pEvent);
	} else {
		if(timer)
			DeleteTimerQueueTimer(timerQueue, timer, 0);
		CreateTimerQueueTimer(&timer, timerQueue, timeout_cb, NULL, (timeInMs <= 0) ? 1 : timeInMs, 0, WT_EXECUTEINTIMERTHREAD | WT_EXECUTEONLYONCE);
	}
#endif
}


} // namespace WebCore

