/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "WebViewWiiU.h"
#include "NotImplemented.h"
#include "Document.h"
#include "DragData.h"
#include "EventHandler.h"
#include "EditorClientWiiU.h"
#include "FocusController.h"
#include "FrameLoaderTypes.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "FrameView.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "GraphicsLayer.h"
#include "DragClientWiiU.h"
#include "InspectorClientWiiU.h"
#include "DeviceMotionClientWiiU.h"
#include "DeviceOrientationClientWiiU.h"
#include "RenderObject.h"
#include "FloatRect.h"
#include "HashMap.h"
#include "IntRect.h"
#include "IntSize.h"
#include "IntPoint.h"
#include "LayerWiiU.h"
#include "Node.h"
#include "Page.h"
#include "InspectorConsoleInstrumentation.h"
#include "InspectorController.h" 
#include "PlatformKeyboardEvent.h"
#include "PlatformMouseEvent.h"
#include "network/curl/ResourceHandleManager.h"
#include "ScrollView.h"
#include "Settings.h"
#include "SystemTime.h"
#include "DragController.h"
#include <WiiUEventManager.h>
#include <ContextMenuClientWiiU.h>
#include <ChromeClientWiiU.h>
#include <FrameLoaderClientWiiU.h>
#include <TextStream.h>
#include <stdlib.h>
#include <string.h>
#include <runtime/InitializeThreading.h>
#include <wtf/Threading.h>
#include <JSBase.h>
#include <JSStringRef.h>
#include <JSContextRef.h>
#include <JSValueRef.h>
#include <JSGlobalObjectFunctions.h>
#include <APICast.h>

#include "MemoryCache.h"
#include "PageCache.h"

#include <JSSystemEventDispatcher.h>
#include <wk_appinterface.h>

#include <PlatformOS.h>
#include <PlatformMemory.h>
using namespace LibPlatform;

#if PLATFORM(WIIU)
#include <thread_ctrl.h>
#include <shim.h>
#endif
#include <fcntl.h>
#include <png.h>
#include <unistd.h>

//#include "JSExtensionController.h"

#if ENABLE(DATABASE)
#include "DatabaseTracker.h"
#endif

#if ENABLE(OFFLINE_WEB_APPLICATIONS)
#include "ApplicationCacheStorage.h"
#endif

#include "GCController.h"
#include "MemoryCache.h"
#include "PageCache.h"
#include "FontCache.h"
#include "CrossOriginPreflightResultCache.h"
 
#include "WiiUDataObject.h"

#include <wtf/wiiu/StatisticsWiiU.h>

using namespace WebCore;

// --------------------------------------------------------------------------------------------------

class WebViewPrivate {
public:
    WebViewPrivate(WebView *view);
    virtual ~WebViewPrivate();

	//TODO: Make this globally accessable at some point
	JSGlobalContextRef globalContext();
	JSGlobalContextRef storedGlobalContext;
	JSDOMWindow *globalObject;
	
    WebCore::Page *page;
    WebCore::Settings *pageSettings;
    WebCore::Frame *mainFrame;

	std::string requiredRPLList;

    WebView *const webView;
    void *m_surface;
    GraphicsContext *ctx;

    PopupMenuWiiU *m_popup;
    IntRect m_popupRect;
	//JSExtensionController *jsxc;

    bool animationsScheduled;

    GraphicsLayer *rootLayer;
    bool rootLayerNeedsDisplay;

    //Crank: DLM keep track of custom cursor hot spot
    WebCore::IntPoint m_cursorHotspot;
};

// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------

WebViewPrivate::WebViewPrivate(WebView *view) : 
	storedGlobalContext(NULL),
	webView(view),
	ctx(0),
	m_popup(0),
   rootLayer(0) {
	
}

// --------------------------------------------------------------------------------------------------

WebViewPrivate::~WebViewPrivate() {
	mainFrame->loader()->stopAllLoaders();
	
	JSC::globalFuncUnloadExtensions(toJS(globalContext()));
	delete page;
	delete ctx;
	//delete jsxc;
}

// --------------------------------------------------------------------------------------------------

JSGlobalContextRef WebView::getJSGlobalContext()
{
	return m_priv->globalContext();
}

// --------------------------------------------------------------------------------------------------

WebView* WebView::GetCurrentWebView(JSContextRef context)
{
	WebView *pView = (WebView *)WebViewTV();	// if none of the following are true, default to the TV

	if (((WebView *)WebViewTV())->getJSGlobalContext() == context)
		pView = (WebView *)WebViewTV();
	else if (((WebView *)WebViewGamePad(0))->getJSGlobalContext() == context)
		pView = (WebView *)WebViewGamePad(0);
	else if(NumGamePadWebViews() > 1) //Added to confirm WebViewGamePad(1) exists
	{
		if (((WebView *)WebViewGamePad(1))->getJSGlobalContext() == context)
			pView = (WebView *)WebViewGamePad(1);
}

	return pView;
}

// --------------------------------------------------------------------------------------------------

bool WebView::authenticate(const char *url, char **user, char **pass)
{
	bool RetVal = false;

	if (m_callbacks && m_callbacks->authenticate) {
		RetVal = m_callbacks->authenticate(url, user, pass, m_userData);
	}

	return RetVal;
}

// --------------------------------------------------------------------------------------------------

JSGlobalContextRef WebViewPrivate::globalContext()
{
    Frame* coreFrame = mainFrame;

    return toGlobalRef(coreFrame->script()->globalObject(mainThreadNormalWorld())->globalExec());
}


// --------------------------------------------------------------------------------------------------
/**
	Paint the web view
*/
void WebView::paint()
{
    FrameView *fview = m_priv->mainFrame->view();
    if (fview == NULL)
    	return;

#if PLATFORM(WIIU)
	// Are redraws enabled?
	if (!redrawEnable())
		return;		// -> No!
#endif

#if !PLATFORM(WIIU)
	//From non-wiiu, this function gets called on every v-sync.
	//We'll skip the rest if there's nothing to do.
	if (!redrawPending())
		return;

	//Make sure we're painting as the correct window
	m_pNewRenderer->SetActiveRenderWindow(temp_windowType);
#endif

	// Build a new graphics context if we don't have one already...
    if(m_priv->ctx == NULL)
	{
		// Create the platform specific context...
		GraphicsContextPlatformPrivate *ctx_priv = new GraphicsContextPlatformPrivate; 
		if (ctx_priv == NULL)
			return;
		
		#if !PLATFORM(WIIU)
		//TODO -- ROB: OpenGL Stuff
		//ctx_priv->m_RenderParams = m_RenderParams;
		ctx_priv->setRenderer(m_pNewRenderer);
		ctx_priv->setWindowType(temp_windowType);
		ctx_priv->SetMipmapGeneration(true);
		#endif
		
		// ...and with that create the platform independent one...
		ctx_priv->setSize(m_width, m_height);
		m_priv->ctx = new GraphicsContext(ctx_priv);
		if (m_priv->ctx == NULL)
		{
			delete ctx_priv;
			return;
		}
    } 

    if (!redrawPending())
	{
		#ifndef NDEBUG
			OS::Report("Paint called but no redraw needed\n");
		#endif
    } 

    //call into WebKit's layout engine
    fview->updateLayoutAndStyleIfNeededRecursive();

    // The area we need to work on...
    IntRect contentRect(m_dirtyRect.x, m_dirtyRect.y, m_dirtyRect.width, m_dirtyRect.height);

    // Unite popup and dirty rectangles for the popup (if any)...
//CRANK: DLM check if this is necessary
    if (m_priv->m_popup && !m_priv->m_popupRect.isEmpty())
        contentRect.unite(m_priv->m_popupRect);

    m_priv->ctx->platformContext()->setDirtyRect(contentRect);

    // Call webkit's paint code...
    fview->paint(m_priv->ctx, contentRect);

    // paint popup if there's anything
    if (m_priv->m_popup && !m_priv->m_popupRect.isEmpty()) {
        m_priv->m_popup->paint(m_priv->ctx);
    }
#if ENABLE(INSPECTOR)
    if (m_priv->mainFrame->page()->inspectorController()->highlightedNode()) {
        m_priv->ctx->save();
        m_priv->mainFrame->page()->inspectorController()->drawHighlight(*m_priv->ctx);
        m_priv->ctx->restore();
    }
#endif 

#if USE(ACCELERATED_COMPOSITING)
    // actually render anything in our compositing layers
    if (m_priv->rootLayerNeedsDisplay && m_priv->rootLayer) { 

        LayerWiiU *root = m_priv->rootLayer->platformLayer();
        ASSERT(root);

        root->paintContentsIfDirty();          
        root->paint(m_priv->ctx, false, 1.0);
        m_priv->rootLayerNeedsDisplay = false;
    }
#endif 

	// Call applications "paint" function
    PaintCall();
    
	// This redraw request has been handled
    m_redrawPending = false;
	if(m_loadComplete) {
		STATS_WIIU_INC_PAINT_COUNT();
	}
}

// --------------------------------------------------------------------------------------------------

void WebView::loadExtensionRPLList()
{
	if (m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld()) == m_priv->globalObject) {

		return;
	}
	
	m_priv->globalObject = m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld());

	std::string rplList = m_priv->requiredRPLList;
	
	//remove all spaces
	unsigned int spaceIndex = rplList.find(" ");
	while(spaceIndex != std::string::npos)
	{
		rplList.replace(spaceIndex, 1, "");
		spaceIndex = rplList.find(" ");
	}


	unsigned int prevIndex = 0;
	unsigned int index = rplList.find(",");
	std::string subString;
	//Parse rpl names out from comma seperated list	
	while(index != std::string::npos)
	{
		subString = rplList.substr(prevIndex, index);

		//skip blank comma spaces
		if(subString != "")
		{
			//make a function that calls require('nameOfExtension') and call it
			subString = "require('" + subString + "');";
			JSStringRef fnCode = JSStringCreateWithUTF8CString(subString.data());
			JSObjectRef fnObject = JSObjectMakeFunction(m_priv->globalContext(), NULL, 0, NULL, fnCode, NULL, 1, NULL);
			JSObjectCallAsFunction(m_priv->globalContext(), fnObject, NULL, 0, NULL, NULL);
			JSStringRelease(fnCode);
		}
		rplList = rplList.substr(index + 1, rplList.length() - index);
		index = rplList.find(",");
	}

	if(rplList != "")
	{
		//make a function that calls require('nameOfExtension') and call it
		rplList = "require('" + rplList + "');";

		JSStringRef fnCode = JSStringCreateWithUTF8CString(rplList.data());
		JSObjectRef fnObject = JSObjectMakeFunction(m_priv->globalContext(), NULL, 0, NULL, fnCode, NULL, 1, NULL);
		JSObjectCallAsFunction(m_priv->globalContext(), fnObject, NULL, 0, NULL, NULL);
		JSStringRelease(fnCode);
	}
	
}

// --------------------------------------------------------------------------------------------------

void WebView::blur()
{
	//removes focus from whatever element has it
	WebCore::FocusController * const focusController = getPage()->focusController();
	WebCore::Frame * const frame = focusController->focusedOrMainFrame();
	focusController->setFocusedNode(NULL, frame); 
}

// --------------------------------------------------------------------------------------------------

void WebView::progress(int percent)
{

#if ENABLE(INSPECTOR)
    InspectorInstrumentation::addMessageToConsole(m_priv->page, OtherMessageSource, LogMessageType, LogMessageLevel, String::format("WebView Progress %d%%", percent), 1, String());
#endif
    if (m_callbacks && m_callbacks->progress) {
        #if PLATFORM(WIIU)
        m_callbacks->progress(percent, m_userData);
        #else
        m_callbacks->progress(percent, this);
        #endif
    }

	//HACK!!!
	//The extension loading still needs moved to a better place
	/*if (m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld()) != m_priv->globalObject) {

		loadExtensionRPLList();
		
		m_priv->globalObject = m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld());
	}*/
}

// --------------------------------------------------------------------------------------------------

void WebView::status(int type, const char *data)
{
#if ENABLE (INSPECTOR)
    String statusmsg;
    switch(type) { // from WebViewWiiU.h
    case WEBVIEW_LOAD_STARTED:
        statusmsg = "WEBVIEW_LOAD_STARTED: ";
        break;
    case WEBVIEW_LOAD_COMPLETE:
        statusmsg = "WEBVIEW_LOAD_COMPLETE: ";
        break;
    case WEBVIEW_STATUS:
        statusmsg = "WEBVIEW_STATUS: ";
        break;
    case WEBVIEW_TITLE_CHANGED:
        statusmsg = "WEBVIEW_TITLE_CHANGED: ";
        break;
    default:
        statusmsg = "UNKNOWN";
        break;
    }
    statusmsg += data;

    InspectorInstrumentation::addMessageToConsole(m_priv->page, OtherMessageSource, LogMessageType, LogMessageLevel, statusmsg, 1, String());
#endif
    if (m_callbacks && m_callbacks->status) {
        m_callbacks->status((WebViewState_t)type, data, m_userData);
    }
	
	if (type == WEBVIEW_LOAD_COMPLETE ) {
		m_loadComplete = true; 
	}

	//HACK!!!
	//The extension loading still needs moved to a better place
	/*if (m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld()) != m_priv->globalObject) {

		loadExtensionRPLList();
		
		m_priv->globalObject = m_priv->mainFrame->script()->globalObject(mainThreadNormalWorld());
	}*/
}

// --------------------------------------------------------------------------------------------------

void WebView::error(const char *url, const char *error)
{
#if ENABLE(INSPECTOR)
    InspectorInstrumentation::addMessageToConsole(m_priv->page, OtherMessageSource, LogMessageType, ErrorMessageLevel, String::format("Error, %s, on URL: %s", error, url), 1, String());
#endif
    if (m_callbacks && m_callbacks->error) {
        m_callbacks->error(url, error, m_userData);
    }
}

// --------------------------------------------------------------------------------------------------

WebCore::Page * WebView::open_window(const char *url)
{
	if (m_callbacks && m_callbacks->open_window)
	{
		WebView *view = (WebView *) m_callbacks->open_window(url, m_userData);
		if (view) {
			return view->getPage();
		}
	}

	return NULL;
}

// --------------------------------------------------------------------------------------------------

void WebView::console_message(const char *message, int line_number, const char *sourceID)
{
	if (m_callbacks && m_callbacks->console_message)
	{
		return m_callbacks->console_message(message, line_number, sourceID, m_userData);
	}

	OS::Report("CONSOLE MESSAGE: %s line=%d source=%s\n", message, line_number, sourceID);
}

// --------------------------------------------------------------------------------------------------

void WebView::js_alert(const char * msg)
{
#if ENABLE (INSPECTOR)
    InspectorInstrumentation::addMessageToConsole(m_priv->page, JSMessageSource, LogMessageType, LogMessageLevel, String::format("JS Alert %s", msg), 1, String());
#endif
	if (m_callbacks && m_callbacks->js_alert)
	{
		return m_callbacks->js_alert(msg, m_userData);
	}

	OS::Report("JS:ALERT: %s\n", msg);
}

// --------------------------------------------------------------------------------------------------

bool WebView::js_confirm(const char * msg)
{
#if ENABLE (INSPECTOR)
    InspectorInstrumentation::addMessageToConsole(m_priv->page, JSMessageSource, LogMessageType, LogMessageLevel, String::format("JS Confirm %s", msg), 1, String());
#endif
	if (m_callbacks && m_callbacks->js_confirm)
	{
		return m_callbacks->js_confirm(msg, m_userData);
	}

	OS::Report("JS:CONFIRM: %s\n", msg);
	return false;
}

// --------------------------------------------------------------------------------------------------

bool WebView::js_prompt(const char *message, const char *default_result, char **result)
{

#if ENABLE (INSPECTOR)
    InspectorInstrumentation::addMessageToConsole(m_priv->page, JSMessageSource, LogMessageType, LogMessageLevel, String::format("JS Prompt %s", message), 1, String());
#endif

	if (m_callbacks && m_callbacks->js_prompt)
	{
		return m_callbacks->js_prompt(message, default_result, result, m_userData);
	}

	OS::Report("JS:PROMPT: %s\n", message);
	*result = NULL;
	return false;
}

// --------------------------------------------------------------------------------------------------

static void logError(const char *msg)
{
    OS::Report("%s\n", msg);
}

// --------------------------------------------------------------------------------------------------

void WebView::initSettings(WebViewSettings_t *settings)
{
    m_priv->pageSettings->setLoadsImagesAutomatically(settings->LoadsImagesAutomatically);
    m_priv->pageSettings->setDefaultFixedFontSize(settings->DefaultFixedFontSize);
    m_priv->pageSettings->setDefaultFontSize(settings->DefaultFontSize);
    m_priv->pageSettings->setSerifFontFamily(settings->SerifFontFamily);
    m_priv->pageSettings->setFixedFontFamily(settings->FixedFontFamily);
    m_priv->pageSettings->setSansSerifFontFamily(settings->SansSerifFontFamily);
    m_priv->pageSettings->setStandardFontFamily(settings->StandardFontFamily);
    m_priv->pageSettings->setJavaScriptEnabled(settings->JavaScriptEnabled);
    m_priv->pageSettings->setJavaScriptCanOpenWindowsAutomatically(settings->JavaScriptCanOpenWindowsAutomatically);
    m_priv->pageSettings->setPluginsEnabled(false);
    m_priv->pageSettings->setLocalStorageEnabled(settings->LocalStorageEnabled);
    m_priv->pageSettings->setOfflineWebApplicationCacheEnabled(settings->OfflineWebApplicationCacheEnabled);
    m_priv->pageSettings->setUsesEncodingDetector(true);
    m_priv->pageSettings->setLocalStorageDatabasePath(String(settings->LocalStorageDatabasePath));
    m_priv->pageSettings->setSessionStorageQuota(settings->SessionStorageQuota);
	m_priv->pageSettings->setDeveloperExtrasEnabled(settings->developerExtras);

	m_priv->requiredRPLList = settings->requiredRPLList;
	//CRANK-ME Setting Hixie to false, as wss:// not working properly with.
	m_priv->pageSettings->setUseHixie76WebSocketProtocol(false);

	// setup page cache as indicated
	// FIXME: THIS IS A GLOBAL SETTING -=NOT=- PER VIEW - we might want to mode this!
    WebCore::pageCache()->setCapacity(settings->pageCacheSize);

	// tell this page to either use or not use the page cache depending on its state
    m_priv->pageSettings->setUsesPageCache(settings->pageCacheSize == 0 ? false : true);
}

// --------------------------------------------------------------------------------------------------

WebView::WebView(int w, int h, WebViewCallbacks_t *cbs, WebViewPaintCB_t paintCB,
		WebViewSettings_t *settings,
        void *surface, bool noSetup, void *userdata, WebView *inspectorview)
    :
#if !PLATFORM(WIIU)
      m_pNewRenderer(0)
    , temp_windowType((WiiUWindowType)(GAMEPAD_UNDEFINED_ID))
    , m_x(0)
#else
      m_x(0)
#endif
    , m_y(0)
    , m_width(w)
    , m_height(h)
    , m_visible(false)
    , m_redrawPending(false)
    , m_dont_quit(true)
    , m_userData(userdata)
    , m_renderToTV(false)
    , m_renderToGamePad(false)
    , m_callbacks(cbs)
    , m_paintCB(paintCB)
    , m_settings(settings)
    , m_displayType(NO_DISPLAY_TYPE)
    , m_doingDragAndDrop(false)
    , m_eventState(WEBVIEW_EVENT_UNINIT)
{   
    m_priv = new WebViewPrivate(this);
	
#if PLATFORM(WIIU)
	// Setup redraw enable/disable according to BG mode state
    m_bRedrawEnable = !shim_IsBackground();

	WTF::String savePath = String("/vol/save/account/webkit");
	if(!fileExists(savePath)){
		makeAllDirectories(savePath);
	}
#else
	m_bRedrawEnable = true;
#endif

    WebCore::InitializeLoggingChannelsIfNecessary();
    ScriptController::initializeThreading();
    WTF::initializeMainThread();

    m_priv->m_surface = surface;

    WebCore::Page::PageClients pageClients;
    pageClients.chromeClient = static_cast<WebCore::ChromeClient*>(new WebCore::ChromeClientWiiU(this));
#if ENABLE(CONTEXT_MENUS)
    pageClients.contextMenuClient = static_cast<WebCore::ContextMenuClient*>(new WebCore::ContextMenuClientWiiU());
#endif
    pageClients.editorClient = static_cast<WebCore::EditorClient*>(new WebCore::EditorClientWiiU(this));

    pageClients.dragClient = static_cast<WebCore::DragClient*>(new WebCore::DragClientWiiU(this));
#if ENABLE(INSPECTOR)
    pageClients.inspectorClient = static_cast<WebCore::InspectorClient*>(new WebCore::InspectorClientWiiU(this, inspectorview));
#endif

#if ENABLE(DEVICE_ORIENTATION)
    pageClients.deviceMotionClient = static_cast<WebCore::DeviceMotionClient*>(new WebCore::DeviceMotionClientWiiU);
    pageClients.deviceOrientationClient = static_cast<WebCore::DeviceOrientationClient*>(new WebCore::DeviceOrientationClientWiiU);
#endif

    m_priv->page = new WebCore::Page(pageClients);
    if (!m_priv->page) {
        logError("Could not create WebKit Page");
        return;
    }

    m_priv->pageSettings = m_priv->page->settings();
    if (!m_priv->pageSettings) {
        logError("Could not get page settings.");
        return;
    }

    if (m_settings) {
		initSettings(m_settings);
		// FIXME: THIS IS A GLOBAL SETTING -=NOT=- PER VIEW - we might want to mode this!
		setObjectCacheCapacities(m_settings->minDeadCapacity, m_settings->maxDeadCapacity, m_settings->capacity);
	}

	#if !PLATFORM(WIIU)
	std::string webkit_savedir = GetSaveDir();
	webkit_savedir.append("webkit");
	#endif

#if ENABLE(DATABASE)
	#if PLATFORM(WIIU)
	WTF::String databasesDirectory = WebCore::pathByAppendingComponent(String("/vol/save/account/webkit"), "Databases");
	#else
	WTF::String databasesDirectory = WebCore::pathByAppendingComponent(String(webkit_savedir.c_str()), "Databases");
	#endif
	WebCore::DatabaseTracker::tracker().setDatabaseDirectoryPath(databasesDirectory);
#endif

	if (!noSetup)
	{
		#if PLATFORM(WIIU)
		WTF::String path = String("/vol/save/account/webkit");
		#else
		WTF::String path = String(webkit_savedir.c_str());
		#endif
		if (!path.isNull()) {
			WebCore::cacheStorage().setCacheDirectory(path);
		}
	}

    WebCore::FrameLoaderClientWiiU *flc = new WebCore::FrameLoaderClientWiiU(this);
    m_priv->mainFrame = Frame::create(m_priv->page, 0, flc).get();
    flc->setWebFrame(m_priv->mainFrame);
	flc->setCustomUserAgent("Mozilla/5.0 (Linux; U; Android 2.3; en-us) AppleWebKit/999+ (KHTML, like Gecko) Safari/999.9");
    m_priv->mainFrame->init();
	
	//m_priv->page->setMainFrame(m_priv->mainFrame); 

    if (!m_priv->mainFrame) {
        logError("Could not create main frame.");
        return;
    }

    FocusController *focusController = m_priv->page->focusController();
    focusController->setFocusedFrame(m_priv->mainFrame);
    focusController->setFocused(true);
    focusController->setActive(true);
	#if PLATFORM(WIIU)
    WebCore::ResourceHandleManager::sharedInstance()->setCookieJarFileName("/vol/save/account/webkit/cookies");
	#else
    std::string cookiedir = webkit_savedir;
    #if OS(WINDOWS)
    cookiedir.append("\\cookies");
    #else
    cookiedir.append("/cookies");
    #endif
    WebCore::ResourceHandleManager::sharedInstance()->setCookieJarFileName(cookiedir.c_str());
	#endif
}

// --------------------------------------------------------------------------------------------------
//CRANK-ME added to start close Web inspector
void WebView::enableWebInspector(bool enable)
{
#if ENABLE(INSPECTOR)
    if (enable == true) {

        m_priv->page->inspectorController()->show();
    }
    else
    {
        m_priv->page->inspectorController()->close();
    }
#endif
}
// --------------------------------------------------------------------------------------------------
void WebView::setVisible(bool b)
{
    m_visible = b;

    //repaint();
}

// --------------------------------------------------------------------------------------------------

WebView::~WebView()
{
    delete m_priv;
}

// --------------------------------------------------------------------------------------------------

void WebView::goForward()
{
    m_priv->page->goForward();
}

// --------------------------------------------------------------------------------------------------

void WebView::goBack()
{
    m_priv->page->goBack();
}

// --------------------------------------------------------------------------------------------------

void WebView::reload()
{
	if(!m_loadComplete) return;
    m_priv->mainFrame->loader()->reload();
	m_loadComplete = false;
	//Reset the event state to "uninitialized" because we are starting page loading over again
	SetEventState(WEBVIEW_EVENT_UNINIT);
	STATS_WIIU_CLEAR_PAINT_COUNT();
}

// --------------------------------------------------------------------------------------------------

void WebView::stop()
{
    m_priv->mainFrame->loader()->stop();
}

// --------------------------------------------------------------------------------------------------

void WebView::load(const char *url)
{
    KURL kurl(ParsedURLString, url);
    m_priv->page->mainFrame()->loader()->load(ResourceRequest(kurl), true);
	m_loadComplete = false;
	//Reset the event state to "uninitialized" because we are starting page loading over again
	SetEventState(WEBVIEW_EVENT_UNINIT);
	STATS_WIIU_CLEAR_PAINT_COUNT();
}

/*!
	Returns a \e CString that contains the URL that the WebView is currently displaying.  
*/
CString WebView::getURLCharacters()
{
	Document *doc = m_priv->mainFrame->document();
	KURL url = doc->url();
	return url.string().utf8();
}

// --------------------------------------------------------------------------------------------------

static int wiiu_png_write(png_structp png, png_bytep data,
    png_size_t length)
{
    PlatformFileHandle *fh = (PlatformFileHandle *)png_get_io_ptr(png);
    return writeToFile(*fh, (const char *)data, (int)length);
}

// --------------------------------------------------------------------------------------------------

static int wiiu_png_flush(png_structp png)
{
    PlatformFileHandle *fh = (PlatformFileHandle *)png_get_io_ptr(png);
    int fd = *fh;
#if OS(WINDOWS)
    return _commit(fd);
#else
    return fsync(fd);
#endif
}

// --------------------------------------------------------------------------------------------------

void WebView::SetUserAgent(const char* agent)
{
	FrameLoaderClient *client = m_priv->page->mainFrame()->loader()->client();
	FrameLoaderClientWiiU *c = dynamic_cast<FrameLoaderClientWiiU*>(client);
	if (c)
		c->setCustomUserAgent(agent);
}

String WebView::GetUserAgent()
{
	String strUA = "";
	FrameLoaderClient *client = m_priv->page->mainFrame()->loader()->client();
	FrameLoaderClientWiiU *c = dynamic_cast<FrameLoaderClientWiiU*>(client);
	if (c)
	{
		const KURL k;	// k is not used in userAgent, but we need a dummy
		strUA = c->userAgent(k);
	}

	return strUA;
}

void WebView::SetCustomHeader(const char* header)
{
	FrameLoaderClient *client = m_priv->page->mainFrame()->loader()->client();
	FrameLoaderClientWiiU *c = dynamic_cast<FrameLoaderClientWiiU*>(client);
	if (c)
		c->setCustomHeader(header);
}

std::vector<String> WebView::GetCustomHeaders()
{
	std::vector<String> customHeaders;

	FrameLoaderClient *client = m_priv->page->mainFrame()->loader()->client();
	FrameLoaderClientWiiU *c = dynamic_cast<FrameLoaderClientWiiU*>(client);
	if (c)
		customHeaders = c->customHeaders();

	return customHeaders;
}

#if ENABLE(REQUEST_ANIMATION_FRAME)

// --------------------------------------------------------------------------------------------------

/* CRANK: DLM: This is called from the ChromeClient as a notification that a requestAnimationFrame callback has
   been requested */
void WebView::scheduleAnimation()
{
    m_priv->animationsScheduled = true;
}

// --------------------------------------------------------------------------------------------------

/* CRANK: DLM: This task is posted to the ScriptExecutionContext so that the requestAnimationFrame
   callback(s) are called at a "safe" time.  Calling serviceScriptedAnimations
   directly leads to instabilitites, leading to a crash after 20 - 30 frames. */
struct ScriptedAnimationsTask : public ScriptExecutionContext::Task {

    DOMTimeStamp ts;
    Document *doc;

    virtual void performTask(ScriptExecutionContext *)
    { 
    	doc->serviceScriptedAnimations(ts);
    } 
};

void WebView::runScheduledAnimations()
{
    if (m_priv->animationsScheduled && m_priv->mainFrame && m_priv->mainFrame->domWindow()) {
        m_priv->animationsScheduled = false;

        /* create and post an asynchronous task to the script controller */
        WTF::PassOwnPtr<ScriptedAnimationsTask> task = adoptPtr(new ScriptedAnimationsTask);
        task->ts = convertSecondsToDOMTimeStamp(WTF::jsCurrentTime());
        task->doc = m_priv->mainFrame->document();
        m_priv->mainFrame->domWindow()->scriptExecutionContext()->postTask(task);
    } 
}
#endif 

// -------------------------------------------------------------------------------------------------- 
void WebView::setCursor(const Cursor *cursor)
{
    if (!m_callbacks || !m_callbacks->set_cursor) return;

    Cursor::Type type = cursor->type();

    if (type == Cursor::Custom) {
        Image *image = cursor->image(); 
        void *texture;
        if (!image || image->isNull()) {
            texture = NULL;
        } else { 
            texture = image->nativeImageForCurrentFrame();
        }

        m_callbacks->set_cursor((WebViewCursorType)type, texture);

    } else {
        m_callbacks->set_cursor((WebViewCursorType)type, 0);
    }

    //keep track of hotspot and apply as offset to mouse events
    m_priv->m_cursorHotspot = cursor->hotSpot();
}

// --------------------------------------------------------------------------------------------------
void WebView::saveScreenshot(const char *filename)
{
#if PLATFORM(WIIU)
	int width, height, pitch;
	unsigned char *pData = (unsigned char*)m_priv->ctx->platformContext()->getContextDataAsRGBA8(width, height, pitch);
	
    PlatformFileHandle *fh = new PlatformFileHandle;
    const char *filename_checked = filename;
    *fh = openFile(filename_checked, OpenForWrite);
    if (*fh == invalidPlatformFileHandle)
	{
        delete fh;
        OS::Report("error could not save screenshot to: %s\n", filename);
        return;
    }

    png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    png_infop info = png_create_info_struct(png);

    //set up error handler
    if (setjmp(png_jmpbuf(png)))
	{
        png_destroy_write_struct(&png, &info);
        closeFile(*fh);
        delete fh;
        OS::Report("error could not write png.");
    }
	else
	{
		png_set_write_fn(png, (void *)fh, (png_rw_ptr)wiiu_png_write, (png_flush_ptr)wiiu_png_flush);
	
		png_set_IHDR(png, info, m_width, m_height, 8, PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	
		png_bytepp rows = (png_bytepp)LibPlatform::Memory::AllocFromDefaultHeap(m_height*sizeof(png_bytep));
		ASSERT(rows);
		for (int i = 0; i < m_height; ++i)
			rows[i] = (png_bytep)(pData + i*4*pitch);
	
		png_write_info(png, info);
		png_write_image(png, rows);
		png_write_end(png, 0);
		closeFile(*fh);
		delete fh;

		LibPlatform::Memory::FreeToDefaultHeap(rows);
	}
	
	// note: we can this free right now, as the CPU is synced vs. the CPU in the code above
    LibPlatform::Memory::FreeToDefaultHeap(pData);
#else

	GraphicsContextPlatformPrivate * priv = m_priv->ctx->platformContext();
	//WiiURenderer * Renderer = priv->getRenderer();

	GAMEPADERROR Result;
	unsigned int ImageWidth = priv->getRenderContext()->m_FBOInfo.m_BufferDimensions.x;
	unsigned int ImageHeight = priv->getRenderContext()->m_FBOInfo.m_BufferDimensions.y;
	unsigned int DataSize = ImageWidth * ImageHeight * 4;
	unsigned char * ImageData = new unsigned char[DataSize];

	//Get the OpenGL image data
	Result = GetOpenGLTextureData(ImageData, priv->getRenderContext()->m_FBOInfo.m_ColorBufferID, ImageWidth, ImageHeight, true);
	if (Result)
	{
		delete [] ImageData;
		OS::Report("error could not save screenshot to: %s\n", filename);
		return;
	}


	PlatformFileHandle *fh = new PlatformFileHandle;
	const char *filename_checked = filename;
	*fh = openFile(filename_checked, OpenForWrite);
	if (*fh == invalidPlatformFileHandle)
	{
		delete fh;
		OS::Report("error could not save screenshot to: %s\n", filename);
		return;
	}

	png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	png_infop info = png_create_info_struct(png);

	//set up error handler
	if (setjmp(png_jmpbuf(png)))
	{
		png_destroy_write_struct(&png, &info);
		closeFile(*fh);
		delete fh;
		OS::Report("error could not write png.");
	}
	else
	{
		png_set_write_fn(png, (void *)fh, (png_rw_ptr)wiiu_png_write, (png_flush_ptr)wiiu_png_flush);

		png_set_IHDR(png, info, ImageWidth, ImageHeight, 8, PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

		png_bytepp rows = (png_bytepp)new unsigned char[ImageHeight * sizeof(png_bytep)];//MEMAllocFromDefaultHeap(m_height*sizeof(png_bytep));
		ASSERT(rows);

		for (int i = 0; i < m_height; ++i)
			rows[i] = (png_bytep)(ImageData + i * 4 * ImageWidth);

		png_write_info(png, info);
		png_write_image(png, rows);
		png_write_end(png, 0);
		closeFile(*fh);
		delete fh;

		delete [] rows;
	}
	delete [] ImageData;
#endif
}

// --------------------------------------------------------------------------------------------------

void WebView::setRedrawPending()
{
    m_redrawPending = true;
}

// --------------------------------------------------------------------------------------------------

bool WebView::redrawPending()
{
    return m_redrawPending;
}

// --------------------------------------------------------------------------------------------------

void WebView::setDirtyRect(const PlatformIntRect &dirtyRect)
{
    m_dirtyRect = dirtyRect;
}

// --------------------------------------------------------------------------------------------------

void WebView::getDirtyRect(PlatformIntRect &dirtyRect)
{
    dirtyRect = m_dirtyRect;
}

// --------------------------------------------------------------------------------------------------

void WebView::sendEvent(WiiUBaseEventData* data)
{
	WiiUEventManager *cemInst = WiiUEventManager::instance();
	cemInst->putEvent(data);
}

// --------------------------------------------------------------------------------------------------

void WebView::dumpInnerText()
{
	Frame* coreFrame = m_priv->mainFrame;
	FrameView* view = coreFrame->view();
	if (view && view->layoutPending())
		view->layout();

	Element* documentElement = coreFrame->document()->documentElement();
	if (!documentElement) {
		OS::Report("NO DOCUMENT ELEMENT");
		return;
	}

	const char *data =  documentElement->innerText().utf8().data();
	OS::Report("INNER TEXT=%s\n", data);
}

// --------------------------------------------------------------------------------------------------

Page *WebView::getPage()
{
	return m_priv->page;
}

// --------------------------------------------------------------------------------------------------

bool WebView::dontQuit()
{
	return m_dont_quit;
}

// --------------------------------------------------------------------------------------------------

void WebView::redraw()
{
	if (redrawPending())
		paint();
}

// --------------------------------------------------------------------------------------------------

void WebView::onMousePressed(const WebCore::PlatformMouseEvent & ev)
{
	WebCore::PlatformMouseEvent	tmpEv(ev);

    //apply cursor hot spot offset
    tmpEv.setPosition(tmpEv.x() + m_priv->m_cursorHotspot.x(), tmpEv.y() + m_priv->m_cursorHotspot.y());

    if (m_priv->mainFrame && m_priv->m_popup && !m_priv->m_popupRect.isEmpty()) {
        int x = tmpEv.x() + m_priv->mainFrame->view()->scrollXForFixedPosition();
        int y = tmpEv.y() + m_priv->mainFrame->view()->scrollYForFixedPosition();
            m_priv->m_popup->processEvent(x, y);
    } else { 
    	m_priv->mainFrame->eventHandler()->handleMousePressEvent(tmpEv);
    }
} 

// --------------------------------------------------------------------------------------------------

void WebView::onMouseReleased(const WebCore::PlatformMouseEvent & ev)
{
	WebCore::PlatformMouseEvent	tmpEv(ev);

	//apply cursor hot spot offset
	tmpEv.setPosition(tmpEv.x() + m_priv->m_cursorHotspot.x(), tmpEv.y() + m_priv->m_cursorHotspot.y());
	tmpEv.setClickCount(1);

	m_priv->mainFrame->eventHandler()->handleMouseReleaseEvent(tmpEv);
    if (m_doingDragAndDrop) {
		//RJD Crank:  After the mouse release has been handled, we need to emit a perform drag event if we are dragging and dropping.
    	IntPoint position(tmpEv.x(), tmpEv.y());
    	DragData dragData((DragDataRef) m_currentDragDataObject.get(), position,  position, static_cast<DragOperation>(1));
    	m_priv->page->dragController()->setIsHandlingDrag(true);
    	m_priv->page->dragController()->performDrag(&dragData);
        m_priv->page->dragController()->dragEnded();
        m_doingDragAndDrop = false;

    }
}

// --------------------------------------------------------------------------------------------------

void WebView::onMouseMoved(const WebCore::PlatformMouseEvent & ev)
{
    if (m_doingDragAndDrop) {
		//RJD Crank:  If we are dragging and dropping, update the drag data only and do not emit the mouse move event. 
    	IntPoint position(ev.x(), ev.y());
    	DragData dragData((DragDataRef) m_currentDragDataObject.get(), position,  position, static_cast<DragOperation>(1));
    	m_priv->page->dragController()->setIsHandlingDrag(true);
    	m_priv->page->dragController()->dragUpdated(&dragData);
    	return;
    }

	WebCore::PlatformMouseEvent	tmpEv(ev);

    //apply cursor hot spot offset
    tmpEv.setPosition(tmpEv.x() + m_priv->m_cursorHotspot.x(), tmpEv.y() + m_priv->m_cursorHotspot.y());

    m_priv->mainFrame->eventHandler()->handleMouseMoveEvent(tmpEv);

}

// --------------------------------------------------------------------------------------------------

void WebView::onKeyDown(const WebCore::PlatformKeyboardEvent & ev)
{
	m_priv->mainFrame->eventHandler()->keyEvent(ev);
}

// --------------------------------------------------------------------------------------------------

void WebView::onKeyUp(const WebCore::PlatformKeyboardEvent & ev)
{
	m_priv->mainFrame->eventHandler()->keyEvent(ev);
}

// --------------------------------------------------------------------------------------------------

void WebView::scrollLeft()
{
	IntSize point(-20, 0); 
	m_priv->mainFrame->view()->scrollBy(point);
}

// --------------------------------------------------------------------------------------------------

void WebView::scrollRight()
{
	IntSize point(20, 0); 
	m_priv->mainFrame->view()->scrollBy(point);
}

// --------------------------------------------------------------------------------------------------

void WebView::scrollUp()
{
	IntSize point(0, -20); 
	m_priv->mainFrame->view()->scrollBy(point);
}

// --------------------------------------------------------------------------------------------------

void WebView::scrollDown()
{
	IntSize point(0, 20); 
	m_priv->mainFrame->view()->scrollBy(point);
}

// --------------------------------------------------------------------------------------------------

void WebView::focusLost()
{
	FocusController* focusController = m_page->focusController();
	Frame* frame = focusController->focusedOrMainFrame();

	focusController->setFocused(false);
	m_page->focusController()->setActive(false);

	// If we are pan-scrolling when we lose focus, stop the pan scrolling.
	frame->eventHandler()->stopAutoscrollTimer();
}

// --------------------------------------------------------------------------------------------------

void WebView::focusGained()
{
	FocusController* focusController = m_page->focusController();
	focusController->setFocused(true);
	focusController->setActive(true);
}

// --------------------------------------------------------------------------------------------------
void WebView::startDragging(void * image, RefPtr<WiiUDataObject> dataObject )
{
    m_doingDragAndDrop = true;
    Cursor cursor = grabCursor();
    m_currentDragDataObject = dataObject;
    setCursor(&cursor);
}

// --------------------------------------------------------------------------------------------------

void WebView::setPopup(PopupMenuWiiU *popup)
{
    m_priv->m_popup = popup;
}

// --------------------------------------------------------------------------------------------------

void WebView::setPopupRect(const IntRect & rect)
{
    m_priv->m_popupRect = rect;
}

// --------------------------------------------------------------------------------------------------

void WebView::SetRenderToTV(bool render_to_tv) 
{
	m_renderToTV = render_to_tv;
	m_gamePadIndex = -1;
}

// --------------------------------------------------------------------------------------------------

void WebView::SetRenderToGamePad(bool render_to_gamepad, int index)
{
	m_renderToGamePad = render_to_gamepad;
	m_gamePadIndex = index;
}

// --------------------------------------------------------------------------------------------------

void WebView::setProxyInfo(const char *host, unsigned long port, WebKitProxyType type, const char *username, const char *password)
{
	WebCore::ResourceHandleManager::ProxyType webcoreType;
	switch (type)
	{
	default:
	case WEBKITPROXY_HTTP:
		webcoreType = WebCore::ResourceHandleManager::HTTP; break;
	case WEBKITPROXY_SOCKS4:
		webcoreType = WebCore::ResourceHandleManager::Socks4; break;
	case WEBKITPROXY_SOCKS4A:
		webcoreType = WebCore::ResourceHandleManager::Socks4A; break;
	case WEBKITPROXY_SOCKS5:
		webcoreType = WebCore::ResourceHandleManager::Socks5; break;
	case WEBKITPROXY_SOCKS5HOSTNAME:
		webcoreType = WebCore::ResourceHandleManager::Socks5Hostname; break;
	};

	WebCore::ResourceHandleManager::sharedInstance()->setProxyInfo(String(host), port, webcoreType, String(username), String(password));
}

// --------------------------------------------------------------------------------------------------

void WebView::setCertificate(const char *cert) {
	WebCore::ResourceHandleManager::sharedInstance()->setCertificate(CString(cert));
}

// --------------------------------------------------------------------------------------------------

bool WebView::GetRenderToTV(void) const
{
	return m_renderToTV;
}

// --------------------------------------------------------------------------------------------------

bool WebView::GetRenderToGamePad(void) const
{
	return m_renderToGamePad;
}

// --------------------------------------------------------------------------------------------------

int WebView::GetGamePadIndex(void) const
{
    return m_gamePadIndex;
}

// --------------------------------------------------------------------------------------------------
WebView::DisplayType WebView::GetDisplayType() const
{
	return m_displayType;
}

// --------------------------------------------------------------------------------------------------
void WebView::SetDisplayType(DisplayType type)
{
    m_displayType = type;
}

// --------------------------------------------------------------------------------------------------

WebViewEventState_t WebView::GetEventState()
{
	return m_eventState;
}

// --------------------------------------------------------------------------------------------------

// This function sets the event state of the WebView and, if applicable, dispatches an event the extensions can listen for
void WebView::SetEventState(WebViewEventState_t eventState)
{
	//If the state is being upgraded (uninit->init->ready->load) or reset to uninit
	if(eventState == (m_eventState+1) || (!eventState && m_eventState))
	{
		WiiUSystemChannel channel = CHANNEL_0;
		//If there are any GamePad WebViews open and this is one of them
		if(NumGamePadWebViews() && (this == WebViewGamePad(0)))
		{
			//Switch the channel to the GamePad window
			channel = CHANNEL_1;
		}
		std::string eventName = "";
		switch(eventState)
		{
			//uninitialized means that the window is starting the loading process from scratch
		case WEBVIEW_EVENT_UNINIT:
			eventName = "uninitialized";
			break;
			//initialized means that the global object has been reset and the extensions have been loaded from config.xml
		case WEBVIEW_EVENT_INIT:
			eventName = "initialized";
			break;
			//ready means that the document is ready and the user can start to edit it, same as DOMContentLoaded event
		case WEBVIEW_EVENT_READY:
			eventName = "ready";
			break;
			//load means that the document is fully loaded, same as onload event
		case WEBVIEW_EVENT_LOAD:
			eventName = "load";
			break;
		}
		JSSystemEventDispatcher::GetInstance()->DispatchDisplayEvent(eventName, channel);

		m_eventState = eventState;
	}
}

// --------------------------------------------------------------------------------------------------

#if !PLATFORM(WIIU)
void WebView::_ForcePaintCall()
{
	//If this webview hasn't had a context created for it yet,
	//we can't force it to paint anything.
	if (!m_priv || !m_priv->ctx)
	{
		//printf("_ForcePaintCall() failed. m_priv: 0x%X; ctx: 0x%X\n", m_priv, m_priv ? m_priv->ctx : 0);
		return;
	}

	m_priv->ctx->platformContext()->getRenderer()->SetActiveRenderWindow(m_priv->ctx->platformContext()->getWindowType());
	m_paintCB((void*)m_priv->ctx->platformContext()->getRenderContext()->m_FBOInfo.m_ColorBufferID, (WebViewHandle_t)this, m_userData);
}
#endif

// --------------------------------------------------------------------------------------------------
void WebView::setRootGraphicsLayer(WebCore::GraphicsLayer *layer)
{
    //this is set to zero when the compositing layers are removed
    m_priv->rootLayer = layer; 
}
// --------------------------------------------------------------------------------------------------
void WebView::setRootLayerNeedsDisplay()
{
    //this flag is picked up in the paint call and causes the render layers
    //to be rendered.
    m_priv->rootLayerNeedsDisplay = true;

}
// --------------------------------------------------------------------------------------------------
/**
	Issue the paint call to user code after the internal texture has been updated
*/
void WebView::PaintCall(void)
{
	// If this view is visible, hand out the texture to the user's render code...
	if(m_visible)
	{
		#if PLATFORM(WIIU)
		m_paintCB(m_priv->ctx->platformContext()->prepareForPaint(), (WebViewHandle_t)this, m_userData);
		#else
		m_priv->ctx->platformContext()->prepareForPaint();
		m_priv->ctx->platformContext()->getRenderer()->SetActiveRenderWindow(m_priv->ctx->platformContext()->getWindowType());
		m_paintCB((void*)m_priv->ctx->platformContext()->getRenderContext()->m_FBOInfo.m_ColorBufferID, (WebViewHandle_t)this, m_userData);
		#endif
	}
}

// --------------------------------------------------------------------------------------------------
/**
*/
void WebView::sendBGNotification(void)
{
	if( m_priv->page )
		m_priv->page->setVisibilityState(PageVisibilityStateHidden, false);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void WebView::sendFGNotification(void)
{
	if( m_priv->page )
		m_priv->page->setVisibilityState(PageVisibilityStateVisible, false);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void WebView::sendShutdownNotification(void)
{
	if(m_priv->mainFrame)
		m_priv->mainFrame->loader()->shouldClose();
	if( m_priv->page )
		m_priv->page->setVisibilityState(PageVisibilityStateShutdown, false);
}

// --------------------------------------------------------------------------------------------------
// - WebViewManager ---------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------

WebViewManager::WebViewManager(void)
	: m_TVwebView(NULL)
    , m_inspectorView(NULL)
	, m_bPanicExit(true)
{
	pthread_mutex_init(&m_GamePadWebViewsMutex, NULL);
	m_restartWebViews = false; 
}

// --------------------------------------------------------------------------------------------------

WebViewManager::~WebViewManager(void)
{
	if (!m_bPanicExit)
	{
		DestroyWebViews(true);
		pthread_mutex_destroy(&m_GamePadWebViewsMutex);
	}
}

// --------------------------------------------------------------------------------------------------

void WebViewManager::SetPanicExit(void)
{
	m_bPanicExit = true;
}

// --------------------------------------------------------------------------------------------------

void WebViewManager::CreateWebViews(WebViewCallbacks_t *callbacks, WebViewSettings_t *settings, unsigned int gamepad_web_view_count, WebViewInspectorLocation_t location)
{
	static WebViewSize_t gamePadSize = { 854, 480};
	static WebViewSize_t splitTvSize = {settings->tvSize.width, settings->tvSize.height >> 1};

	if (location == WEBVIEW_INSPECTOR_SPLIT_TV) {
	    m_inspectorView = (WebView*)WebViewCreate(&splitTvSize, NULL, callbacks, settings, true, NULL, NULL);
	    m_inspectorView->SetDisplayType(WebView::TV_LOWER_DISPLAY_TYPE);
	    settings->tvSize.height = splitTvSize.height;
	} else if (location == WEBVIEW_INSPECTOR_GAMEPAD0) {
	    m_inspectorView = (WebView*)WebViewCreate(&gamePadSize, NULL, callbacks, settings, true, NULL, NULL);
	    m_inspectorView->SetDisplayType(WebView::GAMEPAD_DISPLAY_TYPE);
	}
	
	
	m_TVwebView = (WebView*)WebViewCreate(&settings->tvSize, NULL, callbacks, settings, false, NULL, m_inspectorView);
	if (location == WEBVIEW_INSPECTOR_SPLIT_TV)
	{
	    m_TVwebView->SetDisplayType(WebView::TV_UPPER_DISPLAY_TYPE);
	}
	else
	{
	    m_TVwebView->SetDisplayType(WebView::TV_DISPLAY_TYPE);
	}
	m_TVwebView->SetRenderToTV(true);

	for(unsigned int i = 0; i < gamepad_web_view_count; ++i) 
	{
	    if (( i == 0) && (location == WEBVIEW_INSPECTOR_GAMEPAD0)) {
	        continue;
	    }

	    WebView *gamepad_web_view = (WebView*)WebViewCreate(&gamePadSize, NULL, callbacks, settings, true, NULL, NULL);
	    gamepad_web_view->SetDisplayType(WebView::GAMEPAD_DISPLAY_TYPE);
	    gamepad_web_view->SetRenderToGamePad(true, i);

		pthread_mutex_lock(&m_GamePadWebViewsMutex);
	    m_GamePadWebViews.push_back(gamepad_web_view);
		pthread_mutex_unlock(&m_GamePadWebViewsMutex);
	}
	
	m_settings  = settings;
	m_callbacks = callbacks;
	m_location = location; 
}

// --------------------------------------------------------------------------------------------------

void WebViewManager::DestroyWebViews(bool bOnWebkitThread) 
{
	if (m_inspectorView)
	{
	    WebViewDestroy((WebViewHandle_t)m_inspectorView, bOnWebkitThread);
	    m_inspectorView = NULL;
	}

	if(m_TVwebView)
	{
		WebViewDestroy((WebViewHandle_t)m_TVwebView, bOnWebkitThread);
		m_TVwebView = NULL;
	}

	pthread_mutex_lock(&m_GamePadWebViewsMutex);
	std::vector<WebView *> gamePadWebViewsCopy = m_GamePadWebViews;
	m_GamePadWebViews.clear();
	pthread_mutex_unlock(&m_GamePadWebViewsMutex);
	for(std::vector<WebView *>::iterator i = gamePadWebViewsCopy.begin(); i != gamePadWebViewsCopy.end(); ++i)
	{
		if(*i)
			WebViewDestroy((WebViewHandle_t)*i, bOnWebkitThread);
	}
}

// --------------------------------------------------------------------------------------------------

WebView *WebViewManager::GetTVWebView(void)
{
	return m_TVwebView;
}
// --------------------------------------------------------------------------------------------------

WebView *WebViewManager::GetInspectorWebView(void)
{
    return m_inspectorView;
}

// --------------------------------------------------------------------------------------------------

WebView *WebViewManager::GetGamePadWebView(unsigned int index)
{
	pthread_mutex_lock(&m_GamePadWebViewsMutex);
	WebView *ret = m_GamePadWebViews[index];
	pthread_mutex_unlock(&m_GamePadWebViewsMutex);
	return ret;
}

// --------------------------------------------------------------------------------------------------

WebView::DisplayType WebViewManager::GetDisplayType(WebView *web_view)
{
    return web_view->GetDisplayType();


	// Identify which WebView was passed in.
   // if(web_view == m_TVwebView)
	//	return TV_DISPLAY_TYPE;

	//for(unsigned int i = 0; i < m_GamePadWebViews.size(); ++i)
	//{
	//	if(web_view == m_GamePadWebViews[i])
	//		return GAMEPAD_DISPLAY_TYPE;
	//}

	// If we get here then web_view isn't tied to any of the displays.
	//return NO_DISPLAY_TYPE;
}

// --------------------------------------------------------------------------------------------------

int WebViewManager::GetNumWebViews(void)
{
	int numviews = 0;
	if(m_TVwebView) ++numviews;
	if(m_inspectorView) ++numviews;
	numviews += GetNumGamePadWebViews();
	return numviews;
}

// --------------------------------------------------------------------------------------------------

int WebViewManager::GetNumGamePadWebViews(void)
{
	int numviews = 0;
	pthread_mutex_lock(&m_GamePadWebViewsMutex);
	for(unsigned int i = 0; i < m_GamePadWebViews.size(); ++i)
		if(m_GamePadWebViews[i]) ++numviews;
	pthread_mutex_unlock(&m_GamePadWebViewsMutex);
	return numviews;
}

/*!
	Returns a pointer to the \e WebViewCallbacks_t structure used to 
*/
WebViewCallbacks_t *WebViewManager::getCallbacks()
{
	return m_callbacks;
} 

/*!
	Returns a pointer to the \e WebViewSettings_t structure used to initially create the web views. 
*/
WebViewSettings_t *WebViewManager::getSettings()
{
	return m_settings; 
}
 
/*!
	Returns the current \e WebViewInspectorLocataion_t in use by the framework.  
*/
WebViewInspectorLocation_t WebViewManager::getLocation()
{
	return m_location; 
} 

/*!
	Sets the restart webviews flag to the  \a bool value that is passed in. 
*/
void WebViewManager::setRestartWebViews(bool restart)
{
	m_restartWebViews = restart; 
}

/*!
	Returns \e true if the current WebViews should be torn down and restarted. 
*/
bool WebViewManager::restartWebViews()
{
	return m_restartWebViews; 
}

/*!
 	Cause the heap manager to force a garbage collection now.  It is recommended to use
 	collectGarbageSoon.
 */
void WebView::collectGarbageNow()
{
	gcController().garbageCollectNow();
}

/*!
 	Cause the heap manager to force a garbage collection sometime in the near future.  This is preferable to
 	collectGarbageNow.
 */
void WebView::collectGarbageSoon()
{
	gcController().garbageCollectSoon();
}

/*!
     Frees up as much memory as possible by cleaning all memory caches such
     as page, object and font cache.
 */
void WebView::clearCaches()
{
    // Turn the cache off and on.  Disabling the object cache will remove all
    // resources from the cache.  They may still live on if they are referenced
    // by some Web page though.
    if (!WebCore::memoryCache()->disabled()) {
        WebCore::memoryCache()->setDisabled(true);
        WebCore::memoryCache()->setDisabled(false);
    }

    int pageCapacity = WebCore::pageCache()->capacity();
    // Setting size to 0, makes all pages be released.
    WebCore::pageCache()->setCapacity(0);
    WebCore::pageCache()->releaseAutoreleasedPagesNow();
    WebCore::pageCache()->setCapacity(pageCapacity);

    // Invalidating the font cache and freeing all inactive font data.
    WebCore::fontCache()->invalidate();

    // Empty the Cross-Origin Preflight cache
    WebCore::CrossOriginPreflightResultCache::shared().empty();

}

/*!
     Sets the maximum number of pages to hold in the memory page cache to \a pages.

     The Page Cache allows for a nicer user experience when navigating forth or back
     to pages in the forward/back history, by pausing and resuming up to \a pages.

     For more information about the feature, please refer to:

     http://webkit.org/blog/427/webkit-page-cache-i-the-basics/
*/

void WebView::setMaximumPagesInCache(long pages)
{
    WebCore::pageCache()->setCapacity(pages);
}

/*!
     Returns the maximum number of web pages that are kept in the memory cache.
*/

int WebView::maximumPagesInCache()
{
    return WebCore::pageCache()->capacity();
}

/*!
    Specifies the capacities for the memory cache for dead objects such as
    stylesheets or scripts.

    The \a cacheMinDeadCapacity specifies the \e minimum number of bytes that
    dead objects should consume when the cache is under pressure.

    \a cacheMaxDead is the \e maximum number of bytes that dead objects should
    consume when the cache is \bold not under pressure.

    \a totalCapacity specifies the \e maximum number of bytes that the cache
    should consume \bold overall.

    The cache is enabled by default. Calling setObjectCacheCapacities(0, 0, 0)
    will disable the cache. Calling it with one non-zero enables it again.
*/
void WebView::setObjectCacheCapacities(long cacheMinDeadCapacity, long cacheMaxDead, long totalCapacity)
{
    bool disableCache = !cacheMinDeadCapacity && !cacheMaxDead && !totalCapacity;
    WebCore::memoryCache()->setDisabled(disableCache);

    WebCore::memoryCache()->setCapacities(cacheMinDeadCapacity, cacheMaxDead, totalCapacity);
}

void WebView::clearStatistics() {
	STATS_WIIU_CLEAR_ALL();
}

void WebView::dumpStatistics() {
	unsigned int pc = StatisticsWiiU::instance()->getPaintCount(); 
	unsigned int ic = StatisticsWiiU::instance()->getImageCount(); 
	unsigned int fcc = StatisticsWiiU::instance()->getFontCacheCount(); 
	unsigned int tlc = StatisticsWiiU::instance()->getTransparencyLayerCount(); 
	unsigned int ibc = StatisticsWiiU::instance()->getImageBufferCount(); 
	
	OS::Report("Paint count = %d\n", pc); 
	OS::Report("Image count = %d\n", ic); 
	OS::Report("Font Cache count = %d\n", fcc);
	OS::Report("Transparency Layer count = %d\n", tlc);
	OS::Report("Image Buffer count = %d\n", ibc); 
}	

/*!
    Specifies the file that should be used to store cookies in

    The \a filename specifies the \e name of the file that should be used to
    store the cookies in.
*/

void WebView::setCookieJarFilename(char *filename) {
	WebCore::ResourceHandleManager::sharedInstance()->setCookieJarFileName(filename);
}

/*!
 	Clear all of the current session cookies that are stored.
 */
void WebView::clearSessionCookies() {
	WebCore::ResourceHandleManager::sharedInstance()->clearSessionCookies();
}

/*!
 	Clear all of the current cookies that are stored.  This will remove the cookie file as well.
 */
void WebView::clearAllCookies() {
	WebCore::ResourceHandleManager::sharedInstance()->clearAllCookies();
}

/*!
 	Save all of the current persistent cookies to disk.
 */
void WebView::flushCookiesToDisk() {
	WebCore::ResourceHandleManager::sharedInstance()->flushCookiesToDisk();
}

StatisticsWiiU *getStatistics() {
	return StatisticsWiiU::instance(); 
}


