/*
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef WEB_VIEW_QNX_H
#define WEB_VIEW_QNX_H

#include <wtf/Platform.h> // PLATFORM(...)
#include <wtf/Assertions.h> // ASSERT
#include <wtf/RefPtr.h>
#include <wtf/text/WTFString.h>

#include <wk_appinterface.h>

#if !PLATFORM(WIIU)
#include "WiiURenderParams.h"
#include "renderer/WiiURenderer.h"
#endif

#include <vector>

#if defined(WIN32) && !PLATFORM(WIIU_DEV)
	#if WEBKIT_IMPLEMENTATION
		#define WEBKIT_API __declspec(dllexport)
	#else//WEBKIT_IMPLEMENTATION
		#define WEBKIT_API __declspec(dllimport)
	#endif//WEBKIT_IMPLEMENTATION
#else//defined(WIN32)
	#define WEBKIT_API __attribute__((visibility("default")))
#endif//defined(WIN32)

using namespace WTF; 

namespace WebCore {
    class Cursor;
	class Page;
	class Frame;
    class GraphicsLayer;
	class WiiUEventManager;
	class PlatformMouseEvent;
	class PlatformKeyboardEvent;
	class WiiUDataObject;
	class IntPoint;
	class IntRect;
	class PopupMenuWiiU;
}

namespace WTF {
class StatisticsWiiU;
}

class WebViewPrivate;

struct PlatformIntRect {
    int x, y, width, height;
};

typedef const struct OpaqueJSContext* JSContextRef;
typedef struct OpaqueJSContext* JSGlobalContextRef;

class WebView;

struct WebViewCallbacks {
    void (*status)(int type, const char *data);
    void (*error)(const char *url, const char *error);
    void (*progress)(int percent);
    void (*paint)(void *texture, WebView *webview);
    WebCore::Page * (*open_window)(const char *url);
};

class WebView {

public:
    enum DisplayType {TV_DISPLAY_TYPE, GAMEPAD_DISPLAY_TYPE, TV_UPPER_DISPLAY_TYPE, TV_LOWER_DISPLAY_TYPE, NO_DISPLAY_TYPE};

    WEBKIT_API WebView(int w, int h,
					   WebViewCallbacks_t *cbs, WebViewPaintCB_t paintCB, WebViewSettings_t *settings,
					   void *surface, bool noSetup, void *userData, WebView* inspectorview);
    virtual ~WebView();

	// callbacks
	void status(int type, const char *data);
	void error(const char *url, const char *error);
	void progress(int percent);
	WebCore::Page *open_window(const char *url);
	bool authenticate(const char *url, char **user, char **pass);

    WEBKIT_API void paint();
	void initSettings(WebViewSettings_t *settings);

    void console_message(const char *message, int line_number, const char *sourceID);
    void js_alert(const char * msg);
    bool js_confirm(const char * msg);
    bool js_prompt(const char *message, const char *default_result, char **result);

	// main API
    WEBKIT_API void goForward();
    WEBKIT_API void goBack();
    WEBKIT_API void reload();
    WEBKIT_API void stop();
    WEBKIT_API void load(const char * url);

    WEBKIT_API void onMousePressed(const WebCore::PlatformMouseEvent & ev);
    WEBKIT_API void onMouseReleased(const WebCore::PlatformMouseEvent & ev);
    WEBKIT_API void onMouseMoved(const WebCore::PlatformMouseEvent & ev);

    WEBKIT_API void onKeyDown(const WebCore::PlatformKeyboardEvent & ev);
    WEBKIT_API void onKeyUp(const WebCore::PlatformKeyboardEvent & ev);

    WEBKIT_API void scrollLeft();
    WEBKIT_API void scrollRight();
    WEBKIT_API void scrollUp();
    WEBKIT_API void scrollDown();

    WEBKIT_API void focusLost();
    WEBKIT_API void focusGained();

    WEBKIT_API void redraw();

    WEBKIT_API void saveScreenshot(const char *filename);

	WEBKIT_API void SetUserAgent(const char* agent);
	WEBKIT_API String GetUserAgent();
	WEBKIT_API void SetCustomHeader(const char* header);
	WEBKIT_API std::vector<String> GetCustomHeaders();

#if ENABLE(REQUEST_ANIMATION_FRAME)
    /**
        Allows ChromeClient to notify the WebView that requestAnimationFrame has been called. 
    */
    WEBKIT_API void scheduleAnimation();

    /**
        Allows browser to notify the WebView that animation frame callbacks should be run.
    */
    WEBKIT_API void runScheduledAnimations();
#endif

    // CRANK-ME: Drag and Drop Support
	void startDragging(void * image, RefPtr<WebCore::WiiUDataObject> dataObj);


    void setCursor(const WebCore::Cursor *cursor);

    int x() const { return m_x; }
    int y() const { return m_y; }
    int width() const { return m_width; }
    int height() const { return m_height; }
    
    bool isVisible() const { return m_visible; }    
    void setVisible(bool b);

    void setRedrawPending();
    bool redrawPending();
    void setDirtyRect(const PlatformIntRect &dirtyRect);
    void getDirtyRect(PlatformIntRect &dirtyRect);
	void dumpInnerText();

	static void sendEvent(WiiUBaseEventData* data);
	void enableWebInspector(bool enable);

	WebCore::Page *getPage();
	bool dontQuit();

	//APW: Adding access to global context
	JSGlobalContextRef getJSGlobalContext();
	static WebView *GetCurrentWebView(JSContextRef context);

    void setPopup(WebCore::PopupMenuWiiU *popup);
    void setPopupRect(const WebCore::IntRect & rect);
	void SetRenderToTV(bool render_to_tv);
	void SetRenderToGamePad(bool render_to_gamepad, int index);
	bool GetRenderToTV(void) const;
	bool GetRenderToGamePad(void) const;
	int GetGamePadIndex(void) const;

    void collectGarbageNow();
    void collectGarbageSoon();
    void clearCaches();
    void setMaximumPagesInCache(long pages);
    int maximumPagesInCache();
    void setObjectCacheCapacities(long cacheMinDeadCapacity, long cacheMaxDead, long totalCapacity);
	
    void setCookieJarFilename(char *filename);
    void clearSessionCookies();
    void clearAllCookies();
    void flushCookiesToDisk();
	
	// RJD Crank: Clear the statistics class of all current statistics.  
	void clearStatistics(); 
	
	// RJD Crank: Print out of all the current statistic values.  
	void dumpStatistics(); 
	
	// RJD Crank: Grabs a pointer the statistics instance. 
	StatisticsWiiU *getStatistics(); 

	static void setProxyInfo(const char *host, unsigned long port, WebKitProxyType type, const char *username, const char *password);
	static void setCertificate(const char *cert);
	DisplayType GetDisplayType(void) const;
	void        SetDisplayType(DisplayType type);

	WebViewEventState_t GetEventState();
	void SetEventState(WebViewEventState_t eventState);

	#if !PLATFORM(WIIU)
	void _ForcePaintCall();
	//TODO -- ROB: OpenGL Stuff
	//tWiiURenderParams m_RenderParams;
	WiiURenderer * m_pNewRenderer;
	WiiUWindowType temp_windowType;
	#endif

    void setRootGraphicsLayer(WebCore::GraphicsLayer *);
    void setRootLayerNeedsDisplay();

	void loadExtensionRPLList();

	void blur();
	CString getURLCharacters();
	void sendBGNotification(); 
	void sendFGNotification(); 
	void sendShutdownNotification(); 

	bool redrawEnable() const				{ return m_bRedrawEnable; }
	void setRedrawEnable(bool bEnable)		{ m_bRedrawEnable = bEnable; }

private:
	int						m_x;
	int						m_y;
	int						m_width;
	int						m_height;
	
	bool					m_visible;
	bool                    m_dragActivated;

    bool					m_redrawPending;
    bool					m_dont_quit;
    PlatformIntRect			m_dirtyRect;
	
    void *					m_userData;

	WebViewPrivate *		m_priv;
	WebCore::Frame* 		m_frame;
	WebCore::Page*			m_page;

	bool 					m_renderToTV;
	bool					m_renderToGamePad;
	bool					m_loadComplete; 
	int						m_gamePadIndex;

	WebViewCallbacks_t *	m_callbacks;
	WebViewPaintCB_t	  (*m_paintCB);
	WebViewSettings_t *		m_settings;
	DisplayType             m_displayType;


	WebViewEventState_t		m_eventState;

	bool                    m_doingDragAndDrop;

	RefPtr<WebCore::WiiUDataObject>	m_currentDragDataObject;

	bool					m_bRedrawEnable;

    inline void PaintCall(void);
};

class WebViewManager {
public:


	WebViewManager(void);
	~WebViewManager(void);

	void CreateWebViews(WebViewCallbacks_t *callbacks, WebViewSettings_t *settings, unsigned int gamepad_web_view_count, WebViewInspectorLocation_t location);
	void DestroyWebViews(bool bOnWebkitThread);

	WebView *GetTVWebView(void);
	WebView *GetInspectorWebView(void);
	WebView *GetGamePadWebView(unsigned int index);

	WebView::DisplayType GetDisplayType(WebView *web_view);

	int GetNumWebViews(void);
	int GetNumGamePadWebViews(void);
	
	WebViewCallbacks_t *getCallbacks(); 
	WebViewSettings_t *getSettings(); 
	WebViewInspectorLocation_t getLocation(); 
	
	void setRestartWebViews(bool restart); 
	bool restartWebViews();
	
	void SetPanicExit(void);

private:
	WebView *				m_TVwebView;
	std::vector<WebView *>	m_GamePadWebViews;
	pthread_mutex_t			m_GamePadWebViewsMutex;
	WebView *               m_inspectorView;
	WebViewCallbacks_t *    m_callbacks;
	WebViewSettings_t *		m_settings;
	WebViewInspectorLocation_t 	m_location;
	bool						m_restartWebViews;
	bool					m_bPanicExit;

};

#endif
