/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if OS(WINDOWS)
    #include <windows.h>
    #include <stdio.h>
#else
    #include <sys/types.h>
    #include <sys/socket.h>
    #if OS(UNIX)
        #include <netinet/in.h>
    #endif
#endif

#if PLATFORM(WIIU)
    #include <PlatformOS.h>
    #include <PlatformMemory.h>
    #include <PlatformDefines.h>
    using namespace LibPlatform;
#endif

#include "WiiUEventManager.h"
#include "TcpServerWiiU.h"
#include "wk_appinterface.h"
#include "network/wiiu/WebSocketStreamHandle.h"
#include "network/wiiu/WebSocketStreamHandlePrivate.h"
#include "network/SocketStreamHandleClient.h"
#include "WebInspectorServer.h"

#if PLATFORM(WIIU)
#define SLEEPMILLISECONDS(x) OS::SleepMilliseconds(x)

#elif OS(WINDOWS)
#define SLEEPMILLISECONDS Sleep

#define socketclose closesocket
#define SOSocket socket
#define SOBind   bind
#define SOListen listen
#define SOAccept accept
#define SOClose  closesocket

#elif OS(UNIX)
#define SLEEPMILLISECONDS(x) usleep(1000*x)

#define socketclose close
#define SOSocket socket
#define SOBind   bind
#define SOListen listen
#define SOAccept accept
#define SOClose  close
#endif

namespace WebCore {

#define STACK_SIZE 8 * 1024 * 1204

int TcpServerWiiU::m_state = STATE_INIT;
pthread_mutex_t TcpServerWiiU::m_stateMutex;

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
void TcpServerWiiU::hio_status_callback(DebugIO::Status status, void *context) {
    TcpServerWiiU *tcpServer = (TcpServerWiiU*)context;
    if(status == DEBUGIO_STATUS_NEW_CONNECTION) {
        OS::Report("[INSPECTOR] New connection\n");
        pthread_mutex_lock(&tcpServer->m_stateMutex);
        int state = tcpServer->m_state;
        pthread_mutex_unlock(&tcpServer->m_stateMutex);
        if (state == STATE_LISTENING) {
            WiiUWebInspectorServerAcceptEvent *event = new WiiUWebInspectorServerAcceptEvent((int)tcpServer->m_hioHandle);
            if (event) {
                event->isHio = true;
                WebKitSendEvent(event);
                WebSocketStreamHandlePrivate::m_hioConnectedMutex.lock();
                WebSocketStreamHandlePrivate::m_hioConnected = true;
                WebSocketStreamHandlePrivate::m_hioConnectedMutex.unlock();
            }
        }
    } else if(status == DEBUGIO_STATUS_NO_CONNECTIONS) {
        OS::Report("[INSPECTOR] No connections\n");
        WebSocketStreamHandlePrivate::m_hioConnectedMutex.lock();
        WebSocketStreamHandlePrivate::m_hioConnected = false;
        WebSocketStreamHandlePrivate::m_hioConnectedMutex.unlock();
    } else if(status == DEBUGIO_STATUS_OK) {
        OS::Report("[INSPECTOR] ok\n");
    } else {
        OS::Report("[TCP Server] Host IO error: %d\n", status);
    }
}
#endif

TcpServerWiiU::TcpServerWiiU()
#if PLATFORM(WIIU) || OS(UNIX)
    : m_serverFd(-1)
#elif OS(WINDOWS)
    : m_serverFd(INVALID_SOCKET)
#endif
{
    m_serverFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    m_state = STATE_INIT;

    pthread_mutex_init(&m_stateMutex, NULL);

#if PLATFORM(WIIU)
    unsigned char *ThreadStack = (unsigned char *) Memory::AllocFromDefaultHeap(sizeof(unsigned char) * STACK_SIZE);
#endif

//Make sure this Thread is running on the same core as webkit.
#if PLATFORM(WIIU)
    pthread_attr_t  attr;
    sched_param_t   sparam;
    pthread_attr_init(&attr);
    sparam.sched_priority = PTHREAD_WIIU_PRIO_DEFAULT;

/*
	Why on core 2?

	Unfortunately here the abstraction around process switching etc. breaks a GINORMOUS leak:
	We cannot have the worker thread run on core 0 or 1 as that would force us to move this
	thread around during process switches.
	THIS we cannot do (reliably) as this thread is almost 100% of its time inside the select()
	call. As this is considered "IO", this would block any attempt to move it to another core.

	Simple solution: use core 2 (which should be good enough for this debug thread anyway)
*/
    sparam.sched_affinity = PTHREAD_WIIU_AFFINITY_CORE2;

    pthread_attr_setschedparam(&attr, &sparam);
    pthread_create(&m_thread, &attr, TcpServerWiiU::ClientLoop, &m_serverFd);
#else
    pthread_create(&m_thread, NULL, TcpServerWiiU::ClientLoop, &m_serverFd);
#endif

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
    m_hioHandle = DEBUGIO_HANDLE_INVALID;
#endif
}

TcpServerWiiU::~TcpServerWiiU()
{
    WiiUClose();

    void *ret;
    pthread_join(m_thread, &ret);
}

bool TcpServerWiiU::WiiUListen(unsigned long bindAddress, unsigned short port)
{
    struct sockaddr_in serverAddr;

    memset(&serverAddr, 0, sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port  = htons(port);
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    int nextState = STATE_LISTENING;
    int reuseAddr = 1;
    setsockopt(m_serverFd, SOL_SOCKET, SO_REUSEADDR, (void*)&reuseAddr, sizeof(reuseAddr));

    if (bind(m_serverFd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) < 0) {
        OS::Report("Web Server Bind failed");
        #if !(PLATFORM(WIIU))
        return false;
        #else
        nextState = STATE_BINDING;
        #endif
    }

    if (nextState != STATE_BINDING && listen(m_serverFd, 10) < 0) {
        OS::Report("Web Server Listen Failed");
        #if !(PLATFORM(WIIU)) || !defined(ENABLE_HIO_INSPECTOR)
        return false;
        #endif
    }

    #if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
    // Fixme: Needs less magic numbers and strings
    m_hioHandle = DebugIO::Open("wiiu_inspect", hio_status_callback, this, DEBUGIO_CHANNEL_OPTION_READ_WRITE, 0);
    //if (m_hioHandle == DEBUGIO_STATUS_INVALID_CHANNEL) OS::Report("[TCP Server] Host IO open invalid channel\n");
    if (m_hioHandle < 0) {
        OS::Report("[TCP Server] Host IO channel could not be opened\n");
    }
    #endif

    pthread_mutex_lock(&m_stateMutex);
    m_state = nextState;
    pthread_mutex_unlock(&m_stateMutex);
    return true;
}

void TcpServerWiiU::WiiUClose()
{
#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
	DebugIO::Close(m_hioHandle);
	m_hioHandle = DEBUGIO_HANDLE_INVALID;
#endif
    socketclose(m_serverFd);
    pthread_mutex_lock(&m_stateMutex);
    m_state = STATE_CLOSE;
    pthread_mutex_unlock(&m_stateMutex);
}

#if !PLATFORM(WIIU)
int global_inspector_server_done = 0;
#endif

void *TcpServerWiiU::ClientLoop(void * data)
{
#if PLATFORM(WIIU) || OS(UNIX)
    int *serverFd = (int*) data;
#elif OS(WINDOWS)
    SOCKET *serverFd = (SOCKET*)data;
#endif

    //CRANK-ME Need to protect the m_state since it called from call backs too.
    #if !PLATFORM(WIIU)
    // Fixme: 
    //  This thread will not exit until STATE_CLOSE.
    //  STATE_CLOSE will not happen until the destructor is called.
    //  This instance is statically initialized.
    //  On windows, the destructor will not be called until all the threads exits.
    while (!global_inspector_server_done) {
    #else
    for (;;) {
    #endif
        pthread_mutex_lock(&m_stateMutex);
        int state = m_state;
        pthread_mutex_unlock(&m_stateMutex);

        switch (state) {
        case STATE_INIT:
        {
            SLEEPMILLISECONDS(2000);
            break;
        }
        case STATE_LISTENING:
        {
            int connectionFd = -1;
            struct timeval tv;
            fd_set rfds;

            FD_ZERO(&rfds);
            FD_SET(*serverFd, &rfds);

            tv.tv_sec = 1;
            tv.tv_usec = 0;

            int n = select(*serverFd + 1, &rfds, NULL, NULL, &tv);
            if (n > 0)
                connectionFd = accept(*serverFd, NULL, NULL);

            if (connectionFd > 0) {
                //Send Events
                WiiUWebInspectorServerAcceptEvent *event = new WiiUWebInspectorServerAcceptEvent((int)connectionFd);
                if (event)
                    WebKitSendEvent(event);
                else
                    socketclose(connectionFd);
				}
            #if (PLATFORM(WIIU)) && defined(ENABLE_HIO_INSPECTOR)
            if (n < 0) SLEEPMILLISECONDS(2000);
            #endif
            break;
        }
        case STATE_CLOSE:
        {
            OS::Report("TcpServerWiiU, shutting down\n");
			//SOClose(m_serverFd);
            pthread_mutex_lock(&m_stateMutex);
            m_state = STATE_CLOSE;
            pthread_mutex_unlock(&m_stateMutex);
            return NULL;
        }
        case STATE_BINDING:
        {
            if (WebCore::WebInspectorServer::HasInstance()) {
                struct sockaddr_in serverAddr;
                memset(&serverAddr, 0, sizeof(serverAddr));
                serverAddr.sin_family = AF_INET;
                serverAddr.sin_port  = htons(WebCore::WebInspectorServer::shared().getPort());
                serverAddr.sin_addr.s_addr = INADDR_ANY;

                if (bind(*serverFd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) >= 0) {
                    if (listen(*serverFd, 10) >= 0) {
                        pthread_mutex_lock(&m_stateMutex);
                        m_state = STATE_LISTENING;
                        pthread_mutex_unlock(&m_stateMutex);
                        continue;
                    }
                }
            }
            SLEEPMILLISECONDS(2000);
            break;
        }
        default:
        {
            SLEEPMILLISECONDS(2000);
            break;
        }
        }
    }
    return NULL;
}

void
TcpServerWiiU::HandleAccept(int fd)
{
	#if ENABLE(INSPECTOR_SERVER)
   WebCore::WebInspectorServer::shared().didAcceptConnection(WebCore::WebSocketStreamHandle::create((int)fd, (WebCore::SocketStreamHandleClient*)NULL));
	#endif
}

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
void 
TcpServerWiiU::HandleAcceptHIO(int hioHandle)
{
	#if ENABLE(INSPECTOR_SERVER)
    WebCore::WebInspectorServer::shared().didAcceptConnection(WebCore::WebSocketStreamHandle::create((int)hioHandle, (WebCore::SocketStreamHandleClient*)NULL, true));
	#endif
}
#endif
}
