/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef JSEVENTSYSTEM_H
#define JSEVENTSYSTEM_H

// STL includes
#include <string>
#include <map>
#include <vector>

#include <WebKit/wiiu/Api/wk_event.h>

/**
@brief Event dispatched from JSEventDispatchers to all its listeners.
*/
class JSEvent : public WiiUBaseEventData
{
public:
	/**
	@brief Creates an event of the given type and passes along any data as a void pointer.
	@param event_type Unique name for the event.
	@param data The data stored here is unique to the event.
		Listeners should know what kind of structure it is by the event type.
	*/
	JSEvent(WiiUEventType_t event_id, std::string event_type, void *data = NULL, std::string event_string = "");

	/**
	@brief Destroy event.
	*/
	virtual ~JSEvent(void);

	/**
	@brief String representation of the event. Just its name.
	*/
	virtual std::string ToString(void);

	/**
	@brief Event data.
	*/
	void *GetData(void){return data;}

	/**
	@brief Event type for use with event queue
	*/
	virtual WiiUEventType_t	type() const{ return event_id; }

private:
	std::string event_type;
	std::string event_string; // If the event type isn't unique, this is used to identify the event
	void *data;
	WiiUEventType_t event_id;
};

/**
@brief Any class the extends this can register itself with JSEventDispatcher instances and can receive event notifications.
	All events are sent to the child classes OnEvent function.
*/
class JSEventListener {
public:
	/**
	@brief Create listener and asign it the given name.
	*/
	JSEventListener(std::string listener_name);

	/**
	@brief Destroy listener.
	*/
	virtual ~JSEventListener(void);

	/**
	@brief This function must be implemented by all derived classes.
		If the derived class registers to listen for a specific event with a JSEventDispatcher it will recieve events of that type through this function.
	@param js_event Dispatched event.
	*/
	virtual void OnEvent(JSEvent js_event) = 0;

	/**
	@brief String representation of the Listener.
	@return Unique name of listener.
	*/
	virtual std::string ToString(void);



private:
	std::string listener_name;
	
};

class JSEventDispatcher {
public:
	/**
	@brief Creates an event dispatcher with the given string as its id.
		All dispatcher names should probably be unique, however I dont think it breaks anything if they are not...
	*/
	JSEventDispatcher(std::string dispatcher_name);
	virtual ~JSEventDispatcher(void);

	/**
	@brief Tells this dispatcher that the given listener wants to be notified about any events that match the given event type.
	@param js_listener Event listener to be attached.
	@param event_type String name of the event type to listen to. All events should have unique names.
	@return False if the given listener is already listening to the given event type.
	*/
	virtual bool AddEventListener(JSEventListener *js_listener, std::string event_type);

	/**
	@brief No longer notify the given listener about the given events.
	@param js_listener JSListener that no longer should recieve notifications about the given event.
	@param event_type The name of the event to be ignored.
	*/
	virtual void RemoveEventListener(JSEventListener *js_listener, std::string event_type);

	/**
	@brief Remove all event notifications from the given listener.
	@param js_listener JSListener that no longer should recieve notifications.
	*/
	virtual void RemoveAllListeners(JSEventListener *js_listener);

	/**
	@brief Dispatch the given event to all listeners registered to it.
	@param js_event Event to be dispatched.
	*/
	virtual void DispatchEvent(JSEvent js_event);

	/**
	@brief Dispatch the given event to all listeners registered to it by the given event_string.
	@param js_event Event to be dispatched.
	@param event_string Unique string used to find the listeners of the given js_event.
	*/
	virtual void DispatchEvent(JSEvent js_event, std::string event_string);

	/**
	@brief Does this dispatcher have any listeners on the given event type?
	@return The number of listeners registered to the given event.
	*/
	virtual unsigned int HasEventListener(std::string event_type);

	/**
	@brief Is the given listener currently listening to the given event?
	@param js_listener 
	@param event_type 
	@return True if the given listener is currently listening for events of the given event_type
	*/
	virtual bool HasEventListener(JSEventListener *js_listener, std::string event_type);

	/**
	@brief String representation of the dispatcher.
		Displays the name of this dispatcher, all events being listened to, and all listeners for each event.
	*/
	std::string ToString(void);

private:
	std::string dispatcher_name; // Hopefully unique name identifying the dispatcher.
	std::map<std::string, std::vector<JSEventListener *> *> event_listener_map; // Maps event names to lists of listeners.
};

#endif//JSEVENTSYSTEM_H
