/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <PlatformOS.h>
#include <JSContextRefPrivate.h>
#include <JSSystemEventDispatcher.h>
#include <JSExtensionData.h>
#include "JSExtensionUtils.h"

namespace JSExtUtils
{
	void replaceSubstring(std::string & str, std::string target, std::string replacement)
	{
		size_t index = str.find(target, index); 
		while (index != std::string::npos) { 
			 
			 /* Make the replacement. */ 
			 str.replace(index, target.length(), replacement); 
 
			 /* Advance index forward so the next iteration doesn't pick it up as well. */ 
			 index += replacement.length(); 
			 index = str.find(target, index); 
		} 
	}

	bool checkCallbackScope(JSContextRef callbackContext, JSValueRef callbackScope, JSClassRef expectedClass, JSValueRef* exception)
	{
		if(!JSValueIsObjectOfClass(callbackContext, callbackScope, expectedClass))
		{
			if(exception)
			{
				JSStringRef errorString;
				char infoBuffer[2048];
				JSStringRef infoJSStr = JSContextCreateBacktrace(callbackContext, 2);
				JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
				JSStringRelease(infoJSStr);

				char errorBuffer[2048];
				sprintf(errorBuffer, "%s : Cannot access, scope is invalid\n", infoBuffer);
				LibPlatform::OS::Report(errorBuffer);
				sprintf(errorBuffer, "%s : Cannot access, scope is invalid    ", infoBuffer);
				errorString = JSStringCreateWithUTF8CString(errorBuffer);
				*exception = JSValueMakeString(callbackContext, errorString);
				JSStringRelease(errorString);
			}
			return false;
		}
		return true;
	}

	std::string getJSTypeAsString(JSType type)
	{
		std::string retVal = "undefined";
		switch(type)
		{
		case kJSTypeNull:
			retVal = "null";
			break;
		case kJSTypeBoolean:
			retVal = "Boolean";
			break;
		case kJSTypeNumber:
			retVal = "Number";
			break;
		case kJSTypeString:
			retVal = "String";
			break;
		case kJSTypeObject:
			retVal = "Object";
			break;

		}
		return retVal;
	}

	

	bool checkArgumentList(size_t expectedArgumentCount, JSType expectedTypes[], JSContextRef context,  size_t providedArgumentCount, const JSValueRef arguments[], JSObjectRef thisObject, JSValueRef* exception, size_t numDefaultArguments)
	{
		JSStringRef errorString;
		if(thisObject && !JSObjectGetPrivate(thisObject))
		{
			if(exception)
			{
				char errorBuffer[2048];
				sprintf(errorBuffer, "Invalid scope for callback function.");
				errorString = JSStringCreateWithUTF8CString(errorBuffer);
				*exception = JSValueMakeString(context, errorString);
				JSStringRelease(errorString);
			}
			return false;
		}
	
		//confirm that the number of arguments doesn't exceed the maximum and are at least the minimum
		if(expectedArgumentCount >= providedArgumentCount)
		{
			if(providedArgumentCount >= (expectedArgumentCount - numDefaultArguments))
			{
				JSType type;
				//check each argument against the type provided
				for(int i = 0; i < providedArgumentCount; ++i)
				{
					type = JSValueGetType(context, arguments[i]);
					//if the types of the parameters don't match
					if(type != expectedTypes[i])
					{
						if(exception)
						{
							std::string providedType = getJSTypeAsString(type);
							std::string expectedType = getJSTypeAsString(expectedTypes[i]);
							char infoBuffer[2048];
							JSStringRef infoJSStr = JSContextCreateBacktrace(context, 2);
							JSStringGetUTF8CString(infoJSStr, infoBuffer, 2048);
							JSStringRelease(infoJSStr);
							char errorBuffer[2048];
							sprintf(errorBuffer, "%s \nParameter %u is of type %s, not type %s as expected\n", infoBuffer, i + 1, providedType.c_str(), expectedType.c_str());
							LibPlatform::OS::Report(errorBuffer);
							sprintf(errorBuffer, "%s :Parameter %u is of type %s, not type %s as expected    ", infoBuffer, i + 1, providedType.c_str(), expectedType.c_str());
							char funcbuffer[2048];
							sprintf(funcbuffer, "throw new Error('%s');", errorBuffer);
							std::string funcString = funcbuffer;
							errorString = JSStringCreateWithUTF8CString(errorBuffer);
							*exception = JSValueMakeString(context, errorString);
							JSStringRelease(errorString);
						}

						return false;
					}
				}
				return true;
			}
			else
			{
				if(exception)
				{
					char errorBuffer[2048];
					sprintf(errorBuffer, "The number of provided parameters (%u) was less than the minimum number (%u)\n", providedArgumentCount, (expectedArgumentCount - numDefaultArguments));
					LibPlatform::OS::Report(errorBuffer);
					sprintf(errorBuffer, "The number of provided parameters (%u) was less than the minimum number (%u)    ", providedArgumentCount, (expectedArgumentCount - numDefaultArguments));
					char funcbuffer[2048];
					sprintf(funcbuffer, "throw new Error('%s');", errorBuffer);
					std::string funcString = funcbuffer;
					replaceSubstring(funcString, "\n", ", ");
					errorString = JSStringCreateWithUTF8CString(errorBuffer);
					*exception = JSValueMakeString(context, errorString);
					JSStringRelease(errorString);
				}
			}
			return false;
		}


		//if an exception pointer was passed
		if(exception)
		{
			char errorBuffer[2048];
			sprintf(errorBuffer, "The number of provided parameters (%u) exceeded the expected number (%u)\n", providedArgumentCount, expectedArgumentCount);
			LibPlatform::OS::Report(errorBuffer);
			sprintf(errorBuffer, "The number of provided parameters (%u) exceeded the expected number (%u)    ", providedArgumentCount, expectedArgumentCount);
			char funcbuffer[2048];
			sprintf(funcbuffer, "throw new Error('%s');", errorBuffer);
			std::string funcString = funcbuffer;
			replaceSubstring(funcString, "\n", ", ");
			errorString = JSStringCreateWithUTF8CString(errorBuffer);
			*exception = JSValueMakeString(context, errorString);
			JSStringRelease(errorString);
		}
	
		return false;
	}


	static JSClassDefinition classDef = {
		0,							/* version */
		0,							/* attributes */
		"JSWiiUSystemExtension",	/* name */
		0,							/* parent class */
		0,							/* static values */
		0,							/* static functions */
		0,							/* initialize callback */
		0,							/* finalize callback */
		0,							/* hasProperty callback */
		0,							/* getProperty callback */
		0,							/* setProperty callback */
		0,							/* deleteProperty callback */
		0,							/* getPropertyNames callback */
		0,							/* callAsFunction callback */
		0,							/* callAsConstructor callback */
		0,							/* hasInstance callback */
		0							/* convertToType callback */
	};

	//HACKY!!!! FIX ME LATER!!!!!
	//#include "JSDispatcher.cpp"
	//#include "JSWiiUSystem.cpp"
	//#include "JSWiiUDisplayManager.cpp"
	//#include "JSTVDisplay.cpp"
	//#include "JSGamePadDisplay.cpp"
	//#include "JSProfanityFilter.cpp"






	static JSClassRef theClass = 0;

	static void packageFinalize(JSObjectRef object)
	{
		/* Finalizes*/
	}

	static JSClassDefinition packageClassDef = {
		0,             /* version */
		0,             /* attributes */

		"Package",        /* name */
		0,             /* parent class */

		0,             /* static values */
		0,			   /* static functions */

		0,             /* initialize callback */
		packageFinalize,             /* finalize callback */
		0,             /* hasProperty callback */
		0,             /* getProperty callback */
		0,             /* setProperty callback */
		0,             /* deleteProperty callback */
		0,             /* getPropertyNames callback */
		0,             /* callAsFunction callback */
		0,             /* callAsConstructor callback */
		0,             /* hasInstance callback */
		0              /* convertToType callback */
	};



	JSObjectRef getPackageObject(JSContextRef ctx, JSObjectRef base, std::string packageName)
	{
		JSClassRef tempClass = JSClassCreate(&packageClassDef);
		unsigned int prevIndex = 0;
		unsigned int index = packageName.find(".");
		std::string subString;
		JSObjectRef prevPackage = base;
		JSObjectRef package;
		JSStringRef jsPackageStr;
		JSType typeCheck;
		JSValueRef packageVal;
		while(index != std::string::npos)
		{
			subString = packageName.substr(prevIndex, index);
			jsPackageStr = JSStringCreateWithUTF8CString(subString.data());

			packageVal = JSObjectGetProperty(ctx, prevPackage, jsPackageStr, NULL);
			typeCheck = JSValueGetType(ctx, packageVal);
			//if the subPackage doesn't exist
			if(typeCheck == kJSTypeUndefined)
			{
				//create it
				package = JSObjectMake(ctx, tempClass, NULL);
				JSObjectSetProperty(ctx, prevPackage, jsPackageStr, package, kJSPropertyAttributeNone, NULL);
			}
			else
			{
				package = JSValueToObject(ctx, packageVal, NULL);
			}

			JSStringRelease(jsPackageStr);
			prevPackage = package;
			packageName = packageName.substr(index + 1, packageName.length() - index);
			index = packageName.find(".");
		}

		subString = packageName;
		jsPackageStr = JSStringCreateWithUTF8CString(subString.data());

		packageVal = JSObjectGetProperty(ctx, prevPackage, jsPackageStr, NULL);
		typeCheck = JSValueGetType(ctx, packageVal);
		//if the subPackage doesn't exist
		if(typeCheck == kJSTypeUndefined)
		{
			//create it
			package = JSObjectMake(ctx, tempClass, NULL);
			JSObjectSetProperty(ctx, prevPackage, jsPackageStr, package, kJSPropertyAttributeNone, NULL);
		}
		else
		{
			package = JSValueToObject(ctx, packageVal, NULL);
		}

		return package;
	}

}

JSCustomEventDispatcher::JSCustomEventDispatcher() : JSEventDispatcher("SystemDispatcher") {
	isDispatching = false;
}

JSCustomEventDispatcher::~JSCustomEventDispatcher(){
}

bool JSCustomEventDispatcher::AddEventListener(JSEventListener *js_listener, std::string event_type)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		add_list.push_back(ListenerEventPair(js_listener, event_type));
		return true;
	}
	else
	{
		return JSEventDispatcher::AddEventListener(js_listener, event_type);
	}
}

void JSCustomEventDispatcher::RemoveEventListener(JSEventListener *js_listener, std::string event_type)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		remove_list.push_back(ListenerEventPair(js_listener, event_type));
	}
	else
	{
		JSEventDispatcher::RemoveEventListener(js_listener, event_type);
	}
}

void JSCustomEventDispatcher::RemoveAllListeners(JSEventListener *js_listener)
{
	if(isDispatching)
	{
		//If this function is called during a callback, the action must be queued up to prevent the event_listener_map from getting changed
		clear_list.push_back(js_listener);
	}
	else
	{
		JSEventDispatcher::RemoveAllListeners(js_listener);
	}
}

void JSCustomEventDispatcher::DispatchEvent(JSEvent js_event)
{
	//This flag tells the other functions to wait until we have finished processing all the listener callbacks
	isDispatching = true;
	JSEventDispatcher::DispatchEvent(js_event);
	isDispatching = false;
	processQueuedRequests();
}

void JSCustomEventDispatcher::DispatchEvent(JSEvent js_event, std::string event_string)
{
	//This flag tells the other functions to wait until we have finished processing all the listener callbacks
	isDispatching = true;
	JSEventDispatcher::DispatchEvent(js_event, event_string);
	isDispatching = false;
	processQueuedRequests();
}

void JSCustomEventDispatcher::processQueuedRequests()
{
	//Now we perform all the add/remove actions that were queued up during event callbacks
	typedef std::vector<JSEventListener *>::iterator L;
	for(L l = this->clear_list.begin(); l != this->clear_list.end(); ++l)
	{
		JSEventDispatcher::RemoveAllListeners(*l);
	}
	this->clear_list.clear();

	typedef std::vector<ListenerEventPair>::iterator K;
	for(K k = this->remove_list.begin(); k != this->remove_list.end(); ++k)
	{
		JSEventDispatcher::RemoveEventListener(k->js_listener, k->event_type);
	}
	this->remove_list.clear();

	for(K k = this->add_list.begin(); k != this->add_list.end(); ++k) {
		JSEventDispatcher::AddEventListener(k->js_listener, k->event_type);
	}
	this->add_list.clear();
}