/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "JSSystemEventDispatcher.h"
#include <wtf/Platform.h> // PLATFORM(...)

#include "wk_appinterface.h"
#include "WebViewWiiU.h"
#include <algorithm>
#ifdef _WIIU_
#include <shim.h>
#endif
#include <PlatformAccounts.h>
#include <PlatformError.h>
#include <PlatformInput.h>
#include <PlatformDefines.h>
using namespace LibPlatform;

#define ENUM_TO_STRING(e) # e
#define CHAR_0_VALUE 48

#define REMOTE_URCC_BUTTON_MASK_NO_HOME (REMOTE_URCC_BUTTON_MASK - REMOTE_BUTTON_HOME)
#define GAMEPAD_BUTTON_MASK_NO_HOME (GAMEPAD_BUTTON_MASK - GAMEPAD_BUTTON_HOME - GAMEPAD_BUTTON_HOME)
#define REMOTE_CL_URCC_BUTTON_MASK_NO_HOME (REMOTE_URCC_BUTTON_MASK - REMOTE_CL_BUTTON_HOME)

static std::vector<bool> wii_remote_connected;
static std::vector<bool> wii_remote_disconnected;
static std::vector<bool> wii_remote_extension_connected;
static std::vector<bool> wii_remote_extension_disconnected;
static std::vector<unsigned> wii_remote_battery_levels;
static std::vector<bool> wii_remote_mpls_connected;
static std::vector<bool> wii_remote_mpls_disconnected;
static std::vector<bool> display_remote_connected;
static std::vector<bool> display_remote_disconnected;
static std::vector<unsigned> gamepad_battery_levels;
static std::vector<WebViewEventState_t> webview_event_state;

namespace {
	unsigned buffer = 0;
	unsigned remote_reads_behind = 0;
	unsigned gamepad_reads_behind = 0;

	struct DelayedDispatch {
		std::string event_name;
		std::string event_string;
		void *data;
	};

	std::vector<DelayedDispatch> delayed_dispatches;
}



void DispatchFunctorEvent(std::string event_name, std::string event_string, void *data) {
	JSEvent js_event(WIIU_EMPTY_EVENT, event_name, (void *)data); 
	JSSystemEventDispatcher::GetInstance()->DispatchEvent(js_event, event_string);
}

void QueueDispatchFunctorEvent(std::string event_name, std::string event_string, void *data) {
	DelayedDispatch dd = {event_name, event_string, data};
	delayed_dispatches.push_back(dd);
}

void DispatchRemoteTrigger(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasRemoteTriggered(channel, ::buffer);

		if(trigger_data < 0xffff && trigger_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = trigger_data;

			DispatchFunctorEvent(
				"press",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					BUTTON,
					"press"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchRemoteRelease(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned release_data = input->WasRemoteReleased(channel, ::buffer);

		if(release_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = release_data;
			
			DispatchFunctorEvent(
				"release",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					BUTTON,
					"release"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchRemoteTrigOrRel(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasRemoteTriggered(channel, ::buffer);
		unsigned release_data = input->WasRemoteReleased(channel, ::buffer);

		// HACK: Fix it!
		if(trigger_data > 0xffff || release_data > 0xffff) return;

		if(trigger_data < 0xffff && (trigger_data | release_data) & (REMOTE_BUTTON_LEFT | REMOTE_BUTTON_UP | REMOTE_BUTTON_RIGHT | REMOTE_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsRemotePressed(channel) & (REMOTE_BUTTON_LEFT | REMOTE_BUTTON_UP | REMOTE_BUTTON_RIGHT | REMOTE_BUTTON_DOWN);

			switch(press_data) {
			case 0x0001:	key_data.x = -1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_LEFT
			case 0x0008:	key_data.x =  0; key_data.y =  1; break; // GAMEPADBASE_BUTTON_UP
			case 0x0002:	key_data.x =  1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_RIGHT
			case 0x0004:	key_data.x =  0; key_data.y = -1; break; // GAMEPADBASE_BUTTON_DOWN

			case 0x0009:	key_data.x = -1; key_data.y =  1; break; // left up
			case 0x000a:	key_data.x =  1; key_data.y =  1; break; // up right
			case 0x0006:	key_data.x =  1; key_data.y = -1; break; // right down
			case 0x0005:	key_data.x = -1; key_data.y = -1; break; // down left
			default:		key_data.x =  0; key_data.y =  0; break;
			};

			DispatchFunctorEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchRemoteAcceleration(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 dir_data = input->GetRemoteAccelerationDirection(channel, ::buffer);
		float mag_data = input->GetRemoteAccelerationMagnitude(channel, ::buffer);
		float change_data = input->GetRemoteAccelerationChange(channel, ::buffer);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data
		);
	}
}

void DispatchRemoteGyro(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"update"
			), 
			(void *)&gyro_data
		);
	}
}

void DispatchRemoteGyroCalibration(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	//OS::Report("input->GetMotionPlusCalibrationStatus %f \n", input->GetMotionPlusCalibrationStatus(channel));

	if(::buffer < remote_reads_behind && input->GetMotionPlusCalibrationStatus(channel) == 0.0f) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		DispatchFunctorEvent(
			"calibrationSuccess",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationSuccess"
			), 
			(void *)&gyro_data
		);
	}
}

void DispatchRemoteGyroCalibrationFailed(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	//OS::Report("input->GetMotionPlusCalibrationStatus %f \n", input->GetMotionPlusCalibrationStatus(channel));

	if(::buffer < remote_reads_behind && input->GetMotionPlusCalibrationFailed(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		DispatchFunctorEvent(
			"calibrationFail",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationFail"
			), 
			(void *)&gyro_data
		);
	}
}

void DispatchRemotePointer(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer < remote_reads_behind && input->CheckRemotePointerPosition(channel, ::buffer)) {
		
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
	
		WebView *pView = (WebView*)WebViewTV();
		if (pView) {
			float2 position_data = input->GetRemotePointerPosition(channel, pView->width(), pView->height(), ::buffer);
			float distance_data = input->GetRemoteDistance(channel, ::buffer);
		
			position_data.x += pView->width() / 2.0f;
			position_data.y += pView->height() / 2.0f;

			WiiUMoveData move_data;
			move_data.channel = channel;
			move_data.x = position_data.x;
			move_data.y = position_data.y;
			move_data.z = distance_data;

			DispatchFunctorEvent(
				"update",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DPD,
					"update"
				), 
				(void *)&move_data
			);
		}
	}
}

void DispatchRemoteConnect(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_connected.size() <= channel) {
		wii_remote_connected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasRemoteController(channel)) {

		if(!wii_remote_connected[channel]) {
			wii_remote_connected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;
			
			DispatchFunctorEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_connected[channel] = false;
	}
}

void DispatchRemoteDisconnect(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_disconnected.size() <= channel) {
		wii_remote_disconnected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasRemoteController(channel)) {
		if(!wii_remote_disconnected[channel]) {
			wii_remote_disconnected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerDisconnected"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_disconnected[channel] = false;
	}
}

void DispatchRemoteExtensionAdded(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_extension_connected.size() <= channel) {
		wii_remote_extension_connected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasClassicController(channel) || input->HasNunChuckController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		if(!wii_remote_extension_connected[channel] && input->IsExtensionReady(channel)) {
			wii_remote_extension_connected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;
			
			DispatchFunctorEvent(
				"controllerAdded",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerAdded"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_extension_connected[channel] = false;
	}
}

void DispatchRemoteExtensionRemoved(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_extension_disconnected.size() <= channel) {
		wii_remote_extension_disconnected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasClassicController(channel) && !input->HasNunChuckController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		if(!wii_remote_extension_disconnected[channel] && input->IsExtensionReady(channel)) {
			wii_remote_extension_disconnected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"controllerRemoved",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerRemoved"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_extension_disconnected[channel] = false;
	}
}

void DispatchRemoteBatteryLevelChanged(signed channel) {
	
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	// Expand the array to however many remotes we have
	while(wii_remote_battery_levels.size() <= channel) {
		wii_remote_battery_levels.push_back(input->GetRemoteBatteryLevel(channel));
	}

	if(!input->IsRemoteEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->GetRemoteBatteryLevel(channel) != wii_remote_battery_levels[channel]) {
			wii_remote_battery_levels[channel] = input->GetRemoteBatteryLevel(channel);
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"batteryLevelChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"batteryLevelChange"
				), 
				(void *)&scalar_data
			);
	}
}

void DispatchRemoteMplsAttached(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_mpls_connected.size() <= channel) {
		wii_remote_mpls_connected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(!input->IsRemoteEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->RemoteHasMotionPlusInserted(channel) && input->IsExtensionReady(channel)) {
		if(!wii_remote_mpls_connected[channel]) {
			wii_remote_mpls_connected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"mplsAdded",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"mplsAdded"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_mpls_connected[channel] = false;
	}
}

void DispatchRemoteMplsRemoved(signed channel) {
	// Expand the array to however many remotes we have
	while(wii_remote_mpls_disconnected.size() <= channel) {
		wii_remote_mpls_disconnected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(!input->IsRemoteEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(!input->RemoteHasMotionPlusInserted(channel) && input->IsExtensionReady(channel)) {
		if(!wii_remote_mpls_disconnected[channel]) {
			wii_remote_mpls_disconnected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"mplsRemoved",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"mplsRemoved"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		wii_remote_mpls_disconnected[channel] = false;
	}
}

void DispatchClassicTrigger(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasClassicTriggered(channel, ::buffer);
	
		if(trigger_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = trigger_data;
			
			DispatchFunctorEvent(
				"press",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					BUTTON,
					"press"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchClassicRelease(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		unsigned release_data = input->WasClassicReleased(channel, ::buffer);

		 if(release_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = release_data;
			
			DispatchFunctorEvent(
				"release",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					BUTTON,
					"release"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchClassicTrigOrRel(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->WasClassicTriggered(channel, ::buffer);
		unsigned release_data = input->WasClassicReleased(channel, ::buffer);

		// HACK: Fix it!
		if(trigger_data > 0xffff || release_data > 0xffff) return;

		if(trigger_data < 0xffff && (trigger_data | release_data) & (REMOTE_CL_BUTTON_LEFT | REMOTE_CL_BUTTON_UP | REMOTE_CL_BUTTON_RIGHT | REMOTE_CL_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsClassicPressed(channel) & (REMOTE_CL_BUTTON_LEFT | REMOTE_CL_BUTTON_UP | REMOTE_CL_BUTTON_RIGHT | REMOTE_CL_BUTTON_DOWN);

			switch(press_data) {
			case 0x0002:	key_data.x = -1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_LEFT
			case 0x0001:	key_data.x =  0; key_data.y =  1; break; // GAMEPADBASE_BUTTON_UP
			case 0x8000:	key_data.x =  1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_RIGHT
			case 0x4000:	key_data.x =  0; key_data.y = -1; break; // GAMEPADBASE_BUTTON_DOWN

			case 0x0003:	key_data.x = -1; key_data.y =  1; break; // left up
			case 0x8001:	key_data.x =  1; key_data.y =  1; break; // up right
			case 0xC000:	key_data.x =  1; key_data.y = -1; break; // right down
			case 0x4002:	key_data.x = -1; key_data.y = -1; break; // down left
			default:		key_data.x =  0; key_data.y =  0; break;
			};

			DispatchFunctorEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchClassicLStick(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		float2 lstick_data = input->GetClassicLStick(channel, ::buffer);

		if(lstick_data.x || lstick_data.y) {
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = lstick_data.x;
			vector2_data.y = lstick_data.y;
			
			DispatchFunctorEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&lstick_data
			);
		}
	}
}

void DispatchClassicRStick(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		float2 rstick_data = input->GetClassicRStick(channel, ::buffer);

		if(rstick_data.x || rstick_data.y) {
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = rstick_data.x;
			vector2_data.y = rstick_data.y;
			
			DispatchFunctorEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					RSTICK,
					"move"
				), 
				(void *)&vector2_data
			);
		}
	}
}

void DispatchNunChuckStick(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		float2 stick_data = input->GetNunChuckStick(channel, ::buffer);

		if(stick_data.x || stick_data.y) {
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = stick_data.x;
			vector2_data.y = stick_data.y;

			DispatchFunctorEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_FS, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data
			);
		}
	}
}

void DispatchNunChuckAccel(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		float3 dir_data = input->GetNunChuckAccelerationDirection(channel, ::buffer);
		float mag_data = input->GetNunChuckAccelerationMagnitude(channel, ::buffer);
		float change_data = input->GetNunChuckAccelerationChange(channel, ::buffer);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE_FS, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data
		);
	}
}

void DispatchGamePadTrigger(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->WasGamePadTriggered(channel, ::buffer);

		// HACK: Fix it!
		if(!trigger_data || trigger_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;


		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = trigger_data;

		DispatchFunctorEvent(
			"press",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				BUTTON,
				"press"
			), 
			(void *)&key_data
		);

	}
}

void DispatchGamePadRelease(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned release_data = input->WasGamePadReleased(channel, ::buffer);

		// HACK: Fix it!
		if(!release_data || release_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;
		
		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = release_data;
			
		DispatchFunctorEvent(
			"release",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				BUTTON,
				"release"
			), 
			(void *)&key_data
		);
		
	}
}


void DispatchGamePadTrigOrRel(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->WasGamePadTriggered(channel, ::buffer);
		unsigned release_data = input->WasGamePadReleased(channel, ::buffer);

		// HACK: Fix it!
		if(trigger_data > GAMEPAD_BUTTON_MASK_NO_HOME || release_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;

		if((trigger_data | release_data) & (GAMEPAD_BUTTON_LEFT | GAMEPAD_BUTTON_UP | GAMEPAD_BUTTON_RIGHT | GAMEPAD_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsGamePadPressed(channel) & (GAMEPAD_BUTTON_LEFT | GAMEPAD_BUTTON_UP | GAMEPAD_BUTTON_RIGHT | GAMEPAD_BUTTON_DOWN);

			switch(press_data) {
			case 0x0800:	key_data.x = -1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_LEFT
			case 0x0200:	key_data.x =  0; key_data.y =  1; break; // GAMEPADBASE_BUTTON_UP
			case 0x0400:	key_data.x =  1; key_data.y =  0; break; // GAMEPADBASE_BUTTON_RIGHT
			case 0x0100:	key_data.x =  0; key_data.y = -1; break; // GAMEPADBASE_BUTTON_DOWN

			case 0x0a00:	key_data.x = -1; key_data.y =  1; break; // left up
			case 0x0600:	key_data.x =  1; key_data.y =  1; break; // up right
			case 0x0500:	key_data.x =  1; key_data.y = -1; break; // right down
			case 0x0900:	key_data.x = -1; key_data.y = -1; break; // down left
			default:		key_data.x =  0; key_data.y =  0; break;
			};

			DispatchFunctorEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data
			);
		}
	}
}

void DispatchGamePadLStick(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 lstick_data = input->GetGamePadLStick(channel);

		if(lstick_data.x || lstick_data.y) {
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = lstick_data.x;
			vector2_data.y = lstick_data.y;
			
			DispatchFunctorEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data
			);
		}
	}
}

void DispatchGamePadRStick(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 rstick_data = input->GetGamePadRStick(channel);

		if(rstick_data.x || rstick_data.y) {
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = rstick_data.x;
			vector2_data.y = rstick_data.y;

			DispatchFunctorEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					RSTICK,
					"move"
				), 
				(void *)&vector2_data
			);
		}
	}
}

void DispatchGamePadGyro(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float3 vel_data = input->GetGamePadGyro(channel);
		float3 angle_data = input->GetGamePadAngle(channel);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				GYRO,
				"update"
			), 
			(void *)&gyro_data
		);
	}
}

void DispatchGamePadPointer(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel) && input->CheckGamePadTouchPosition(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 touch_position = input->GetGamePadTouchPosition(channel);
		WiiUVector2Data vector2_data;
		vector2_data.channel = channel;
		vector2_data.x = touch_position.x;
		vector2_data.y = touch_position.y;
		
		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				TOUCH,
				"update"
			), 
			(void *)&vector2_data
		);
	}
}

void DispatchGamePadAcceleration(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float3 dir_data = input->GetGamePadAccelerationDirection(channel);
		float mag_data = input->GetGamePadAccelerationMagnitude(channel);
		float change_data = input->GetGamePadAccelerationChange(channel);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		DispatchFunctorEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data
		);
	}
}

void DispatchGamePadBatteryLevelChanged(signed channel) {
	
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	// Expand the array to however many remotes we have
	while(gamepad_battery_levels.size() <= channel) {
		gamepad_battery_levels.push_back(input->GetGamePadBatteryLevel(channel));
	}

	if(!input->IsGamePadEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->GetGamePadBatteryLevel(channel) != gamepad_battery_levels[channel]) {
			gamepad_battery_levels[channel] = input->GetGamePadBatteryLevel(channel);
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"batteryLevelChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"batteryLevelChange"
				), 
				(void *)&scalar_data
			);
	}
}


void DispatchGamePadConnect(signed channel) {
	// Expand the array to however many remotes we have
	while(display_remote_connected.size() <= channel) {
		display_remote_connected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasGamePadController(channel)) {
		if(!display_remote_connected[channel]) {
			display_remote_connected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;

			DispatchFunctorEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data
			);
		} 
	} else {
		display_remote_connected[channel] = false;
	}
}

void DispatchGamePadDisconnect(signed channel) {
	// Expand the array to however many remotes we have
	while(display_remote_disconnected.size() <= channel) {
		display_remote_disconnected.push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasGamePadController(channel)) {
		if(!display_remote_disconnected[channel]) {
			display_remote_disconnected[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			DispatchFunctorEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT, 
					"controllerDisconnected"
				), 
				(void *)&scalar_data
			);
		}
	} else {
		display_remote_disconnected[channel] = false;
	}
}

void DispatchCAMReady(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel) && input->CameraReady(channel)) {
		DispatchFunctorEvent(
			"ready",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"ready"
			), 
			0
		);
	}
}

void DispatchCAMStreamOpen(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel) && input->CameraStreamOpened(channel)) {
		DispatchFunctorEvent(
			"open",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"open"
			), 
			0
		);
	}
}

void DispatchCAMStreamUpdate(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel) && input->UpdateCameraStream(channel)) {
		//DispatchFunctorEvent(ENUM_TO_STRING(WiiUSystemEvent::CAM_STREAM_UPDATE), channel, 0);
	}
}

void DispatchCAMStreamClose(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->CameraStreamClosed(channel)) {
		DispatchFunctorEvent(
			"close",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"close"
			), 
			0
		);
	}
}

void DispatchCAMExit(signed channel) {
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->CameraExited(channel)) {
		DispatchFunctorEvent(
			"exit",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				DEFAULT,
				"exit"
			), 
			0
		);
	}
}

void DispatchDisplayLoad(signed channel) {
	return;
}

void DispatchDisplayInitialized(signed channel) {
	return;
}

void DispatchDisplayUninitialized(signed channel) {
	return;
}

void DispatchDisplayReady(signed channel) {
	return;
}

void DispatchErrorWindowClose(signed) {
	
}

void DispatchErrors(signed) {
	LibPlatform::Error *error = LibPlatform::Error::GetInstance();

	while(error->HasReports()) {
		LibPlatform::Error::ReportInfo report_info = error->NextReport();

		WiiUErrorData error_data;
		error_data.code = report_info.code;
		error_data.description = report_info.description.c_str();

		// Extract the first three digits for the error type
		error_data.type = ((unsigned)report_info.type - ((unsigned)report_info.type % 10000)) / 10000;

		DispatchFunctorEvent(
			"error",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				SYSTEM, 
				"error"
			), 
			(void *)&error_data 
		);
	}
}


namespace {
	JSSystemEventDispatcher *instance = 0;
};

JSSystemEventDispatcher::JSSystemEventDispatcher(void) : JSEventDispatcher("SystemDispatcher") {
	// Create space for the ARRAY ARRAY ARRAY MAP OF DOOM!
	event_strings.resize(DEVICE_COUNT);
	for(int d = 0; d < DEVICE_COUNT; ++d) {
		event_strings[d].resize(CHANNEL_COUNT);
		for(int c = 0; c < CHANNEL_COUNT; ++c) {
			event_strings[d][c].resize(CONTROL_COUNT);
		}
	}

	// REMOTE unique event strings
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("REMOTE_CONNECT0", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("REMOTE_DISCONNECT0", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["controllerAdded"] =	KeyDispatcherPair("REMOTE_EXTADDED0", DispatchRemoteExtensionAdded);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["controllerRemoved"] =	KeyDispatcherPair("REMOTE_EXTREMOVED0", DispatchRemoteExtensionRemoved);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("REMOTE_BATTERYCHANGED0", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["mplsAdded"] =	KeyDispatcherPair("REMOTE_MPLSADDED0", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_0][DEFAULT]["mplsRemoved"] =	KeyDispatcherPair("REMOTE_MPLSREMOVED0", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_0][BUTTON]["press"] =					KeyDispatcherPair("REMOTE_TRIGGERED0", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_0][BUTTON]["release"] =					KeyDispatcherPair("REMOTE_RELEASED0", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_0][DPD]["update"] =						KeyDispatcherPair("REMOTE_MOVED0", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_0][ACCELEROMETER]["update"] =			KeyDispatcherPair("REMOTE_ACCEL0", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_0][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL0", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_0][GYRO]["update"] =						KeyDispatcherPair("REMOTE_GYRO0", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_0][GYRO]["calibrationSuccess"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATION0", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_0][GYRO]["calibrationFail"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL0", DispatchRemoteGyroCalibrationFailed);
	
																					 
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("REMOTE_CONNECT1", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("REMOTE_DISCONNECT1", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["controllerAdded"] =	KeyDispatcherPair("REMOTE_EXTADDED1", DispatchRemoteExtensionAdded);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["controllerRemoved"] =	KeyDispatcherPair("REMOTE_EXTREMOVED1", DispatchRemoteExtensionRemoved);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("REMOTE_BATTERYCHANGED1", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["mplsAdded"] =	KeyDispatcherPair("REMOTE_MPLSADDED1", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_1][DEFAULT]["mplsRemoved"] =	KeyDispatcherPair("REMOTE_MPLSREMOVED1", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_1][BUTTON]["press"] =					KeyDispatcherPair("REMOTE_TRIGGERED1", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_1][BUTTON]["release"] =					KeyDispatcherPair("REMOTE_RELEASED1", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_1][DPD]["update"] =						KeyDispatcherPair("REMOTE_MOVED1", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_1][ACCELEROMETER]["update"] =			KeyDispatcherPair("REMOTE_ACCEL1", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_1][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL1", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_1][GYRO]["update"] =						KeyDispatcherPair("REMOTE_GYRO1", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_1][GYRO]["calibrationSuccess"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATION1", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_1][GYRO]["calibrationFail"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL1", DispatchRemoteGyroCalibrationFailed);
																					 
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("REMOTE_CONNECT2", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("REMOTE_DISCONNECT2", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["controllerAdded"] =	KeyDispatcherPair("REMOTE_EXTADDED2", DispatchRemoteExtensionAdded);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["controllerRemoved"] =	KeyDispatcherPair("REMOTE_EXTREMOVED2", DispatchRemoteExtensionRemoved);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("REMOTE_BATTERYCHANGED2", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["mplsAdded"] =	KeyDispatcherPair("REMOTE_MPLSADDED2", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_2][DEFAULT]["mplsRemoved"] =	KeyDispatcherPair("REMOTE_MPLSREMOVED2", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_2][BUTTON]["press"] =					KeyDispatcherPair("REMOTE_TRIGGERED2", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_2][BUTTON]["release"] =					KeyDispatcherPair("REMOTE_RELEASED2", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_2][DPD]["update"] =						KeyDispatcherPair("REMOTE_MOVED2", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_2][ACCELEROMETER]["update"] =			KeyDispatcherPair("REMOTE_ACCEL2", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_2][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL2", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_2][GYRO]["update"] =						KeyDispatcherPair("REMOTE_GYRO2", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_2][GYRO]["calibrationSuccess"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATION2", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_2][GYRO]["calibrationFail"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL2", DispatchRemoteGyroCalibrationFailed);
																					 
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("REMOTE_CONNECT3", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("REMOTE_DISCONNECT3", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["controllerAdded"] =	KeyDispatcherPair("REMOTE_EXTADDED3", DispatchRemoteExtensionAdded);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["controllerRemoved"] =	KeyDispatcherPair("REMOTE_EXTREMOVED3", DispatchRemoteExtensionRemoved);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("REMOTE_BATTERYCHANGED3", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["mplsAdded"] =	KeyDispatcherPair("REMOTE_MPLSADDED3", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_3][DEFAULT]["mplsRemoved"] =	KeyDispatcherPair("REMOTE_MPLSREMOVED3", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_3][BUTTON]["press"] =					KeyDispatcherPair("REMOTE_TRIGGERED3", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_3][BUTTON]["release"] =					KeyDispatcherPair("REMOTE_RELEASED3", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_3][DPD]["update"] =						KeyDispatcherPair("REMOTE_MOVED3", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_3][ACCELEROMETER]["update"] =			KeyDispatcherPair("REMOTE_ACCEL3", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_3][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL3", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_3][GYRO]["update"] =						KeyDispatcherPair("REMOTE_GYRO3", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_3][GYRO]["calibrationSuccess"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATION3", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_3][GYRO]["calibrationFail"] =		KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL3", DispatchRemoteGyroCalibrationFailed);

	// Classic controller strings
	event_strings[REMOTE_CL][CHANNEL_0][BUTTON]["press"] =				KeyDispatcherPair("REMOTE_CL_TRIGGERED0", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_0][BUTTON]["release"] =				KeyDispatcherPair("REMOTE_CL_RELEASED0", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_0][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_LSTICK0", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_0][RSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_RSTICK0", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_0][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL0", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_1][BUTTON]["press"] =				KeyDispatcherPair("REMOTE_CL_TRIGGERED1", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_1][BUTTON]["release"] =				KeyDispatcherPair("REMOTE_CL_RELEASED1", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_1][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_LSTICK1", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_1][RSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_RSTICK1", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_1][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL1", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_2][BUTTON]["press"] =				KeyDispatcherPair("REMOTE_CL_TRIGGERED2", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_2][BUTTON]["release"] =				KeyDispatcherPair("REMOTE_CL_RELEASED2", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_2][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_LSTICK2", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_2][RSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_RSTICK2", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_2][DIRECTION]["directionChange"] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL2", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_3][BUTTON]["press"] =				KeyDispatcherPair("REMOTE_CL_TRIGGERED3", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_3][BUTTON]["release"] =				KeyDispatcherPair("REMOTE_CL_RELEASED3", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_3][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_LSTICK3", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_3][RSTICK]["move"] =					KeyDispatcherPair("REMOTE_CL_RSTICK3", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_3][DIRECTION]["directionChange"] =	KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL3", DispatchClassicTrigOrRel);
																		
	// Nun-chuck controller event strings								
	event_strings[REMOTE_FS][CHANNEL_0][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_FS_STICK0", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_0][ACCELEROMETER]["update"] =		KeyDispatcherPair("REMOTE_FS_ACCEL0", DispatchNunChuckAccel);
																		
	event_strings[REMOTE_FS][CHANNEL_1][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_FS_STICK1", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_1][ACCELEROMETER]["update"] =		KeyDispatcherPair("REMOTE_FS_ACCEL1", DispatchNunChuckAccel);
																		
	event_strings[REMOTE_FS][CHANNEL_2][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_FS_STICK2", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_2][ACCELEROMETER]["update"] =		KeyDispatcherPair("REMOTE_FS_ACCEL2", DispatchNunChuckAccel);
																		
	event_strings[REMOTE_FS][CHANNEL_3][LSTICK]["move"] =					KeyDispatcherPair("REMOTE_FS_STICK3", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_3][ACCELEROMETER]["update"] =		KeyDispatcherPair("REMOTE_FS_ACCEL3", DispatchNunChuckAccel);

	// GAMEPAD unique event strings
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("GAMEPAD_CONNECT0", DispatchGamePadConnect);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("GAMEPAD_DISCONNECT0", DispatchGamePadDisconnect);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("GAMEPAD_BATTERYCHANGED0", DispatchGamePadBatteryLevelChanged);
	event_strings[GAMEPAD][CHANNEL_0][BUTTON]["press"] =					KeyDispatcherPair("GAMEPAD_TRIGGERED0", DispatchGamePadTrigger);
	event_strings[GAMEPAD][CHANNEL_0][BUTTON]["release"] =					KeyDispatcherPair("GAMEPAD_RELEASED0", DispatchGamePadRelease);
	event_strings[GAMEPAD][CHANNEL_0][TOUCH]["update"] =					KeyDispatcherPair("GAMEPAD_MOVED0", DispatchGamePadPointer);
	event_strings[GAMEPAD][CHANNEL_0][ACCELEROMETER]["update"] =			KeyDispatcherPair("GAMEPAD_ACCEL0", DispatchGamePadAcceleration);
	event_strings[GAMEPAD][CHANNEL_0][LSTICK]["move"] =					KeyDispatcherPair("GAMEPAD_LSTICK0", DispatchGamePadLStick);
	event_strings[GAMEPAD][CHANNEL_0][RSTICK]["move"] =					KeyDispatcherPair("GAMEPAD_RSTICK0", DispatchGamePadRStick);
	event_strings[GAMEPAD][CHANNEL_0][DIRECTION]["directionChange"] =		KeyDispatcherPair("GAMEPAD_TRIG_OR_REL0", DispatchGamePadTrigOrRel);
	event_strings[GAMEPAD][CHANNEL_0][GYRO]["update"] =						KeyDispatcherPair("GAMEPAD_GYRO0", DispatchGamePadGyro);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["ready"] =					KeyDispatcherPair("CAM_READY0", DispatchCAMReady);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["open"] =					KeyDispatcherPair("CAM_STREAM_OPEN0", DispatchCAMStreamOpen);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["update"] =					KeyDispatcherPair("CAM_STREAM_UPDATE0", DispatchCAMStreamUpdate);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["close"] =					KeyDispatcherPair("CAM_STREAM_CLOSE0", DispatchCAMStreamClose);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT]["exit"] =					KeyDispatcherPair("CAM_EXIT0", DispatchCAMExit);
																		
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["controllerConnected"] =	KeyDispatcherPair("GAMEPAD_CONNECT1", DispatchGamePadConnect);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["controllerDisconnected"] =	KeyDispatcherPair("GAMEPAD_DISCONNECT1", DispatchGamePadDisconnect);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["batteryLevelChange"] =	KeyDispatcherPair("GAMEPAD_BATTERYCHANGED1", DispatchGamePadBatteryLevelChanged);
	event_strings[GAMEPAD][CHANNEL_1][BUTTON]["press"] =					KeyDispatcherPair("GAMEPAD_TRIGGERED1", DispatchGamePadTrigger);
	event_strings[GAMEPAD][CHANNEL_1][BUTTON]["release"] =					KeyDispatcherPair("GAMEPAD_RELEASED1", DispatchGamePadRelease);
	event_strings[GAMEPAD][CHANNEL_1][TOUCH]["update"] =					KeyDispatcherPair("GAMEPAD_MOVED1", DispatchGamePadPointer);
	event_strings[GAMEPAD][CHANNEL_1][ACCELEROMETER]["update"] =			KeyDispatcherPair("GAMEPAD_ACCEL1", DispatchGamePadAcceleration);
	event_strings[GAMEPAD][CHANNEL_1][LSTICK]["move"] =					KeyDispatcherPair("GAMEPAD_LSTICK1", DispatchGamePadLStick);
	event_strings[GAMEPAD][CHANNEL_1][RSTICK]["move"] =					KeyDispatcherPair("GAMEPAD_RSTICK1", DispatchGamePadRStick);
	event_strings[GAMEPAD][CHANNEL_1][DIRECTION]["directionChange"] =		KeyDispatcherPair("GAMEPAD_TRIG_OR_REL1", DispatchGamePadTrigOrRel);
	event_strings[GAMEPAD][CHANNEL_1][GYRO]["update"] =						KeyDispatcherPair("GAMEPAD_GYRO1", DispatchGamePadGyro);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["ready"] =					KeyDispatcherPair("CAM_READY1", DispatchCAMReady);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["open"] =					KeyDispatcherPair("CAM_STREAM_OPEN1", DispatchCAMStreamOpen);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["update"] =					KeyDispatcherPair("CAM_STREAM_UPDATE1", DispatchCAMStreamUpdate);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["close"] =					KeyDispatcherPair("CAM_STREAM_CLOSE1", DispatchCAMStreamClose);
	event_strings[GAMEPAD][CHANNEL_1][DEFAULT]["exit"] =					KeyDispatcherPair("CAM_EXIT1", DispatchCAMExit);

	// File system strings
	event_strings[IO][CHANNEL_0][DEFAULT]["readComplete"] =				KeyDispatcherPair("IO_READ_COMPLETE", 0);
	event_strings[IO][CHANNEL_0][DEFAULT]["saveComplete"] =				KeyDispatcherPair("IO_SAVE_COMPLETE", 0);
	event_strings[IO][CHANNEL_0][DEFAULT]["error"] =					KeyDispatcherPair("IO_ERROR", 0);

	// Account strings
	event_strings[ACT][CHANNEL_0][DEFAULT]["accountServiceToken"] =		KeyDispatcherPair("ACCOUNT_SERVICE_TOKEN_SUCCESS", 0);
	event_strings[ACT][CHANNEL_0][DEFAULT]["accountServiceTokenError"] =		KeyDispatcherPair("ACCOUNT_SERVICE_TOKEN_FAIL", 0);

	// Error strings
	event_strings[ERROR_WINDOW][CHANNEL_0][DEFAULT]["errorWindowClose"] =	KeyDispatcherPair("ERROR_WINDOW_CLOSE", DispatchErrorWindowClose);

	// System event strings
	event_strings[SYSTEM][CHANNEL_0][DEFAULT]["error"] =					KeyDispatcherPair("SYSTEM_ERROR", DispatchErrors);

	// Display event strings
	//CHANNEL_0 == tv
	event_strings[DISPLAY][CHANNEL_0][DEFAULT]["initialized"] =					KeyDispatcherPair("DISPLAY_INIT0", DispatchDisplayInitialized);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT]["load"] =						KeyDispatcherPair("DISPLAY_LOAD0", DispatchDisplayLoad);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT]["uninitialized"] =				KeyDispatcherPair("DISPLAY_UNINIT0", DispatchDisplayUninitialized);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT]["ready"] =						KeyDispatcherPair("DISPLAY_READY0", DispatchDisplayReady);

	//CHANNEL_1 == GAMEPAD_0
	event_strings[DISPLAY][CHANNEL_1][DEFAULT]["initialized"] =					KeyDispatcherPair("DISPLAY_INIT1", DispatchDisplayInitialized);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT]["load"] =						KeyDispatcherPair("DISPLAY_LOAD1", DispatchDisplayLoad);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT]["uninitialized"] =				KeyDispatcherPair("DISPLAY_UNINIT1", DispatchDisplayUninitialized);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT]["ready"] =						KeyDispatcherPair("DISPLAY_READY1", DispatchDisplayReady);

	//CHANNEL_2 == GAMEPAD_1
	event_strings[DISPLAY][CHANNEL_2][DEFAULT]["initialized"] =					KeyDispatcherPair("DISPLAY_INIT2", DispatchDisplayInitialized);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT]["load"] =						KeyDispatcherPair("DISPLAY_LOAD2", DispatchDisplayLoad);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT]["uninitialized"] =				KeyDispatcherPair("DISPLAY_UNINIT2", DispatchDisplayUninitialized);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT]["ready"] =						KeyDispatcherPair("DISPLAY_READY2", DispatchDisplayReady);
}

JSSystemEventDispatcher::~JSSystemEventDispatcher(void) {
}

JSSystemEventDispatcher *JSSystemEventDispatcher::GetInstance(void) {
	if(!::instance) {
		::instance = new JSSystemEventDispatcher();
	}

	return ::instance;
}

void JSSystemEventDispatcher::DestroyInstance(void) {
	if(::instance) {
		delete ::instance;
		::instance = 0;
	}
}

bool JSSystemEventDispatcher::AddDefaultEventListener(JSEventListener *js_listener, std::string event_type) {
	return JSEventDispatcher::AddEventListener(js_listener, event_type);
}

std::string JSSystemEventDispatcher::GetEventString(WiiUSystemDevice device, WiiUSystemChannel channel, WiiUSystemControl control, std::string event_type) {
	return event_strings[device][channel][control][event_type].key;
}

std::string JSSystemEventDispatcher::GetEventString(WiiUSystemDevice device, std::string event_type) {
	return this->GetEventString(device, CHANNEL_0, DEFAULT, event_type);
}

bool JSSystemEventDispatcher::AddEventListener(
	JSEventListener *js_listener, 
	WiiUSystemDevice device, 
	WiiUSystemChannel channel, 
	WiiUSystemControl control,
	std::string event_type) 
{
	KeyDispatcherPair pair = event_strings[device][channel][control][event_type];
	if(!pair.dispatcher) return false;

	bool success = JSEventDispatcher::AddEventListener(js_listener, pair.key);
	if(!success) return false;

	this->add_list.push_back(DispatchFunctor(pair.dispatcher, channel, pair.key));
	return true;
}

bool JSSystemEventDispatcher::AddEventListener(
	JSEventListener *js_listener,
	WiiUSystemDevice device,
	std::string event_type) 
{
	return AddEventListener(js_listener, device, CHANNEL_0, DEFAULT, event_type);
}

void JSSystemEventDispatcher::RemoveEventListener(JSEventListener *js_listener, std::string event_type) {
	remove_list.push_back(ListenerEventPair(js_listener, event_type));
}

void JSSystemEventDispatcher::RemoveAllListeners(JSEventListener *js_listener)
{
	clear_list.push_back(js_listener);
}

void JSSystemEventDispatcher::DispatchInputEvents(void) {
	// Figure out how far behind the main input thread we are
	Input *input = Input::GetInstance();

	typedef std::vector<JSEventListener *>::iterator L;
	for(L l = this->clear_list.begin(); l != this->clear_list.end(); ++l)
	{
		JSEventDispatcher::RemoveAllListeners(*l);
	}
	this->clear_list.clear();

	// Remove all dead listeners before we dispatch
	typedef std::vector<DispatchFunctor>::iterator I;
	typedef std::vector<ListenerEventPair>::iterator K;
	for(K k = this->remove_list.begin(); k != this->remove_list.end(); ++k)
	{
		JSEventDispatcher::RemoveEventListener(k->js_listener, k->event_type);

		//if no listeners remain
		if(!JSEventDispatcher::HasEventListener(k->event_type))
		{
			//remove the matching functor
			for(I i = this->dispatch_functors.begin(); i != this->dispatch_functors.end(); ++i)
			{
				if (i->event_type == k->event_type)
				{
					this->dispatch_functors.erase(i);
					break;
				}
			}
		}
	}
	this->remove_list.clear();

	// Its possible that a dispatch functor is added while dispatching an event.
	// So that we dont get a concurrent modification issue copy over pointers to the current list of functors
	// and iterate over that instead of the functor list.

	this->dispatch_functor_pointers.clear();
	for(I i = this->add_list.begin(); i != this->add_list.end(); ++i) {
		
		//We have to make sure that the Dispatch Functor isn't added to the list twice
		I dfp;
		for(dfp = this->dispatch_functors.begin(); dfp != this->dispatch_functors.end(); ++dfp) {
			//APW - 10/1/12: JSSystemEventDispatcher needs to check against event_name when adding listeners because some events use the same function 
			if(dfp->function == i->function && dfp->event_type == i->event_type)
				break;
		}
		//if the dispatch functor isn't already in the list, add it
		if(dfp == this->dispatch_functors.end())
		{
			this->dispatch_functors.push_back(*i);
		}
	}
	this->add_list.clear();

	for(I i = this->dispatch_functors.begin(); i != this->dispatch_functors.end(); ++i) {
		this->dispatch_functor_pointers.push_back(&(*i));
	}

	::remote_reads_behind = 0;
	::gamepad_reads_behind = 0;

	//input->LockRemoteCache();
	//input->LockGamePadCache();

	unsigned cached_remote_count0 = input->HasCachedRemoteReads(0);
	unsigned cached_remote_count1 = input->HasCachedRemoteReads(1);
	unsigned cached_remote_count2 = input->HasCachedRemoteReads(2);
	unsigned cached_remote_count3 = input->HasCachedRemoteReads(3);
	unsigned cached_gamepad_count0 = input->HasCachedGamePadReads(0);

	if(input->HasRemoteController(0)) ::remote_reads_behind = std::max(cached_remote_count0, ::remote_reads_behind);
	if(input->HasRemoteController(1)) ::remote_reads_behind = std::max(cached_remote_count1, ::remote_reads_behind);
	if(input->HasRemoteController(2)) ::remote_reads_behind = std::max(cached_remote_count2, ::remote_reads_behind);
	if(input->HasRemoteController(3)) ::remote_reads_behind = std::max(cached_remote_count3, ::remote_reads_behind);
	if(input->HasGamePadController(0)) ::gamepad_reads_behind = std::max(cached_gamepad_count0, ::gamepad_reads_behind);

	// Dispatch events for every input update that we're behind
	for(unsigned i = 0; i <= std::max(::gamepad_reads_behind, ::remote_reads_behind); ++i) {
		// Which buffer to read from
		::buffer = i;

		typedef std::vector<DispatchFunctor *>::iterator J;
		for(J j = this->dispatch_functor_pointers.begin(); j != this->dispatch_functor_pointers.end(); ++j) {
			(*(*j))();
		}
	}

	input->ClearCachedRemoteReads(0, cached_remote_count0);
	input->ClearCachedRemoteReads(1, cached_remote_count1);
	input->ClearCachedRemoteReads(2, cached_remote_count2);
	input->ClearCachedRemoteReads(3, cached_remote_count3);
	input->ClearCachedGamePadReads(0, cached_gamepad_count0);

	//input->UnlockGamePadCache();
	//input->UnlockRemoteCache();

	for(unsigned i = 0; i < ::delayed_dispatches.size(); ++i) {
		JSEvent js_event(WIIU_EMPTY_EVENT, delayed_dispatches[i].event_name, (void *)delayed_dispatches[i].data); 
		JSSystemEventDispatcher::GetInstance()->DispatchEvent(js_event, delayed_dispatches[i].event_string);
	}

	delayed_dispatches.clear();
}

void JSSystemEventDispatcher::DispatchSystemEvent(std::string event_name, std::string event_string, void *data, WiiUEventType_t event_type) {
	JSEvent js_event(event_type, event_name, (void *)data); 
	this->DispatchEvent(js_event, event_string);
}

void JSSystemEventDispatcher::DispatchDisplayEvent(std::string event_string, WiiUSystemChannel channel) {

	QueueDispatchFunctorEvent(
		event_string.c_str(),
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				DISPLAY, 
				channel,
				DEFAULT,
				event_string.c_str()
			), 
			NULL
		);
}

void JSSystemEventDispatcher::DispatchErrorEvent(JSEvent* js_event) {

	QueueDispatchFunctorEvent(
			"errorWindowClose",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				ERROR_WINDOW, 
				CHANNEL_0,
				DEFAULT,
				"errorWindowClose"
			), 
			js_event
		);
}
