/*---------------------------------------------------------------------------*

  Copyright 2010-2011 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/
// -----------------------------------------------------------------------------
//  wksupGfx.h
//
// -----------------------------------------------------------------------------

#ifndef __WKSUP_GFX_H__
#define __WKSUP_GFX_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <wksup/wksupGfxTypes.h>
#include <wksup/wksupGfxShader.h>

/// @addtogroup wksupRender
/// @{

typedef struct _WKSUP_MEMHOOKS {
	void*	(*mMEM1alloc)(u32 bytes, u32 alignment);
	void	(*mMEM1free)(void *pAddr);
	void*	(*mFGalloc)(u32 bytes, u32 alignment);
	void	(*mFGfree)(void *pAddr);
} WKSUP_MEMHOOKS;

/// \brief Default color buffer for wksupGfx (global)
extern GX2ColorBuffer wksupColorBuffer;
extern GX2DepthBuffer wksupDepthBuffer;

/// \brief Default context state ptr for wksupGfx (global)
extern GX2ContextState *wksupContextState;

/// \brief Memory allocator that allocates MEM1 for graphics
///
/// This function calls \ref GX2NotifyMemAlloc which enables
/// various (future) debugging features.
void *wksupGfxAllocMEM1(u32 size, u32 align);

/// \brief Memory allocator that allocates from the foreground bucket for graphics
///
/// This function calls \ref GX2NotifyMemAlloc which enables
/// various (future) debugging features.
void *wksupGfxAllocFG(u32 size, u32 align);

/// \brief Memory allocator that allocates MEM2 for graphics
///
/// This function calls \ref GX2NotifyMemAlloc which enables
/// various (future) debugging features.
void *wksupGfxAllocMEM2(u32 size, u32 align);

/// \brief Free MEM1
///
void wksupGfxFreeMEM1(void * ptr);

/// \brief Free memory from the foreground bucket.
///
void wksupGfxFreeFG(void * ptr);

/// \brief Free MEM2
///
void wksupGfxFreeMEM2(void * ptr);

/// \brief Initializes GX2 & AVM; sets up default render/depth/scan buffers.
///
/// It must be called before calling any other GX2 functions. \n
/// argv can take follow argment.
///
///  -wksup_WIDTH=n       (set width to n) \n
///  -wksup_HEIGHT=n      (set height to n) \n
///  -wksup_CB_FORMAT=str (set color buffer format to 8_8_8_8, 10_10_10_2, 2_10_10_10, 16_16_16_16F, or 32_32_32_32F) \n
///  -wksup_SCAN_FORMAT=str (set scan-out buffer format to 8_8_8_8, 10_10_10_2, or 2_10_10_10) \n
///  -wksup_DB_FORMAT=str (set depth buffer format to 16, 32F, 8_24, X24_8_32F) \n
///  -wksup_AA_MODE=n     (set AA mode to 0/1/2/3) \n
///
/// Set up defaults for all options (1280/720/8_8_8_8/32F/no AA). \n
/// Unknown args should be ignored.  Leaves args unchanged.
///
/// This function will call wksupGfxMem1HeapInit(), which will allocate all available memory from the MEM1
/// frame heap and initialize the expanded heap used by wksupGfxAllocMEM1().
///
/// \param argc number of argments
/// \param argv argment lists
//
void wksupGfxInit(int argc, char *argv[], const WKSUP_MEMHOOKS *pMemHooks, GX2AAMode aa_mode);

/// \brief Releases allocated buffers and shuts down GX2.
///
void wksupGfxShutdown(void);

/// \brief Releases foreground only resources and GX2.
///
void wksupGfxReleaseForeground(void);

/// \brief Re-acquires GX2 resources and prepares for rendering.
///
void wksupGfxAcquiredForeground(void);

/// \brief Get wksup graphics running state
///
/// \retval TRUE if \ref wksupGfxInit() has been called; false otherwise.
BOOL wksupGfxIsRunning(void);

/// Set wksup gfx context to current.
/// Need to call this API before Drawing with wksup gfx context created in wksupGfxInit.
void wksupGfxSetContextState(void);

/// \brief Currently, this function waits for the previously rendered frame to appear.
/// Functionality may change in the future.
///
void wksupGfxBeforeRender(void);

/// \brief Swaps the video (scan) buffers, requests new frame to be displayed.
/// Note that it does not wait for the new frame to appear.
///
void wksupGfxDoneRender(void);

/// \brief Synchronize with video swap in a specific way
///
/// This function is an example of how to specify a precise amount of
/// latency with respect to when video swap occurs.
/// \param depth Number of frames rendering may run ahead of display
/// \param percent Additional fraction of a frame to wait after specified swap
///                (expressed as an integer 0...100 percent of the swap interval time)
void wksupGfxWaitForSwap(u32 depth, u32 percent);

/// \brief A helper function to reset render targets to wksup buffers.
///
inline void wksupGfxSetDefaultRenderTarget(void) {
    GX2SetColorBuffer(&wksupColorBuffer, GX2_RENDER_TARGET_0);
	GX2SetDepthBuffer(&wksupDepthBuffer);
}

/// \brief Setting up special color buffer (e.g) multi-layered frame buffer
///
/// \param colorBuffer Ptr to color buffer structure to initialize.
/// \param dim   Desired dim for color buffer.
/// \param width  Desired width for color buffer.
/// \param height Desired height for color buffer.
/// \param depth  Desired depth for color buffer.
/// \param format Desired surface format for color buffer.
/// \param aa     Desired AA mode for color buffer.
///
inline void wksupGfxInitColorBufferEx(GX2ColorBuffer *colorBuffer,
                                   GX2SurfaceDim dim,
                                   u32 width, u32 height, u32 depth,
                                   GX2SurfaceFormat format, GX2AAMode aa)
{
    GX2_CHECK_ENUM_RANGE(dim, GX2_SURFACE_DIM)
    GX2_CHECK_ENUM_RANGE(format, GX2_SURFACE_FORMAT)
    GX2_CHECK_ENUM_RANGE(aa, GX2_AA_MODE)

    colorBuffer->surface.dim = dim;
    colorBuffer->surface.width = width;
    colorBuffer->surface.height = height;
    colorBuffer->surface.depth = depth;
    colorBuffer->surface.numMips = 1; // 1 means base level only
    colorBuffer->surface.format = format;
    colorBuffer->surface.aa = aa;
    colorBuffer->surface.use = GX2_SURFACE_USE_COLOR_BUFFER_TEXTURE;
    colorBuffer->surface.tileMode = GX2_TILE_MODE_DEFAULT;
    colorBuffer->surface.swizzle  = 0;
    colorBuffer->viewMip = 0;
    colorBuffer->viewFirstSlice = 0;
    colorBuffer->viewNumSlices = depth;
    GX2CalcSurfaceSizeAndAlignment(&colorBuffer->surface);
    GX2InitColorBufferRegs(colorBuffer);
}

/// \brief Setting up special depth buffer (e.g) multi-layered frame buffer
///
/// \param depthBuffer Ptr to depth buffer structure to initialize.
/// \param dim  Desired dim for depth buffer.
/// \param width  Desired width for depth buffer.
/// \param height Desired height for depth buffer.
/// \param depth  Desired depth for depth buffer.
/// \param format Desired surface format for depth buffer.
/// \param aa     Desired AA mode for depth buffer.
///
inline void wksupGfxInitDepthBufferEx(GX2DepthBuffer *depthBuffer,
                                   GX2SurfaceDim dim,
                                   u32 width, u32 height, u32 depth,
                                   GX2SurfaceFormat format, GX2AAMode aa)
{
    GX2_CHECK_ENUM_RANGE(dim, GX2_SURFACE_DIM)
    GX2_CHECK_ENUM_RANGE(format, GX2_SURFACE_FORMAT)
    GX2_CHECK_ENUM_RANGE(aa, GX2_AA_MODE)

    depthBuffer->surface.dim = dim;
    depthBuffer->surface.width = width;
    depthBuffer->surface.height = height;
    depthBuffer->surface.depth = depth;
    depthBuffer->surface.numMips = 1; // 1 means base level only
    depthBuffer->surface.format = format;
    depthBuffer->surface.aa = aa;
    depthBuffer->surface.use = (format==GX2_SURFACE_FORMAT_D_D24_S8_UNORM) ? GX2_SURFACE_USE_DEPTH_BUFFER : GX2_SURFACE_USE_DEPTH_BUFFER_TEXTURE;
    depthBuffer->surface.tileMode = GX2_TILE_MODE_DEFAULT;
    depthBuffer->surface.swizzle  = 0;
    depthBuffer->viewMip = 0;
    depthBuffer->viewFirstSlice = 0;
    depthBuffer->viewNumSlices = depth;
    depthBuffer->clearDepth = 1.0f;
    depthBuffer->clearStencil = 0;
    GX2CalcSurfaceSizeAndAlignment(&depthBuffer->surface);
    GX2InitDepthBufferRegs(depthBuffer);
}

void wksupSetHDCPMode(BOOL bEnabled);
BOOL wksupGetHDCPMode(void);

/// @}  // wksupRender

#ifdef __cplusplus
}
#endif

#endif /// __WKSUP_GFX_H__
